////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Host Control Channel Module Implementation
///
///     This file implements the Host Control Channel (HCC) module. This
///     module implements a messaging system for communication between the
///     firmware and host OS. User applications on the host OS can use ioctl
///     calls to send packets to the HCC module. The packets are then sent to
///     the firmware via the GFUN module.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_hcc.h"
#include "tera_hcc_internal.h"
#include "tera_gfun.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// Value to write to the GFUN mailbox to notify the firmware that we are ready
#define HCC_READY_GFUN_MBOX_VAL  (0xDEADBEEF)

// ---- Local Variables -----------------------------------------------------

//
// This workqueue is used to schedule a kernel thread that periodically polls
// the PCoIP HCC device.
//
static struct workqueue_struct *hcc_wq;

//
// The following pointer is initialized when the driver starts. The pointer is
// referenced by the dpc_entry function.
//
// Note: It would be better to pass the pointer into the function but
// workqueues on some Linux distributions do not support passing in a
// parameter. This feature was removed in version 2.6.20 of the kernel.
//
sTERA_DEV_CBLK *hcc_dev_cblk = NULL;

// ---- Local Function Prototypes -------------------------------------------

static void scp_chan_open_cback(
    UINT8 chan_id,
    void *user
    );

static INT hcc_egress_desc_q_proc(
    sHCC_CBLK *cblk
    );

static void hcc_ingress_desc_q_proc(
    sHCC_CBLK     *cblk,
    sTERA_MEM_BUF *buf,
    UINT32         buf_len
    );

static void dpc_entry(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct work_struct *unused
#else
    void *unused
#endif
    );

// --------------------------------------------------------------------------
//  scp_chan_open_cback
//      SCP channel open callback
//
//  Description:
//      This function is a callback for the SCP channel open event.
//
static void scp_chan_open_cback(
    UINT8 chan_id,
    void *user
    )
{
    sHCC_CBLK *cblk = user;

    ASSERT(user != NULL);

    switch (chan_id)
    {
        case TERA_PCOIP_HCC_APDU_USER_HMI:
        {
            mTERA_LOG_INFO("HMI SCP channel open");

            hcc_hmi_scp_chan_open_proc(cblk);
            break;
        }
        case TERA_PCOIP_HCC_APDU_USER_KMP:
        {
            mTERA_LOG_INFO("KMP SCP channel open");

            break;
        }
        case TERA_PCOIP_HCC_APDU_USER_IMG:
        {
            mTERA_LOG_INFO("IMG SCP channel open");

            break;
        }
        default:
        {
            //
            // We should never be here because we only tell the SCP module to
            // open SCP channels that we know of
            //
            mTERA_LOG_ERR("unsupported SCP channel ID=%u", chan_id);
            break;
        }
    }
} // scp_chan_open_cback()

// --------------------------------------------------------------------------
//  hcc_egress_desc_q_proc
//      Send a packet to the egress descriptor queue
//
//  Description:
//      This function sends an outstanding packet (if exists) to the egress
//      descriptor queue. The packet is pulled from the egress packet
//      queue.
//
static INT hcc_egress_desc_q_proc(
    sHCC_CBLK *cblk
    )
{
    INT        ret = 0;
    INT32      err;
    sTERA_PKT *pkt;

    if ((cblk != NULL) && cblk->initialized)
    {
        // Grab a packet from the head of the queue
        pkt = tera_pkt_q_dequeue(&cblk->egress_pkt_q);
        if (pkt != NULL)
        {
#ifdef GFUN_DBG
            sTERA_MEM_BUF *buf = pkt->buf;
            UINT32 *data_ptr = buf->virt_addr;
            mTERA_LOG_DBG("buf data=%08X %08X %08X %08X %08X %08X %08X %08X",
                          data_ptr[0], data_ptr[1], data_ptr[2], data_ptr[3],
                          data_ptr[4], data_ptr[5], data_ptr[6], data_ptr[7]);
#endif

            // Send the packet off to the firmware
            err = tera_gfun_tx(cblk->dev, pkt->buf, pkt->buf_len, NULL);
            if (err != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send packet over GFUN (ret=%d)", ret);

                if (err == TERA_ERR_FULL)
                {
                    //
                    // If the send failed because the egress descriptor queues
                    // are full, then we should add the packet back to onto the
                    // egress packet queue to try to send it again later.
                    //
                    tera_pkt_q_head_enqueue(&cblk->egress_pkt_q, pkt);
                }
                else
                {
                    //
                    // We're here because the send failed and its not because
                    // the egress descriptor queues are full. This must mean
                    // something bad has happened or we're in the process of
                    // being shutdown. Thus, drop the packet on the floor.
                    //
                    tera_pkt_free(cblk->dev, pkt);
                }
            }
            else
            {
#ifdef GFUN_DBG
                mTERA_LOG_DBG("sent packet to GFUN with len=%d", pkt->buf_len);
#endif

                // The GFUN owns the buffer now, so detach it from the packet
                pkt->buf = NULL;

                // Free the packet
                tera_pkt_free(cblk->dev, pkt);

                ret = 1;
            }
        } // end if
    } // end if

    return ret;
} // hcc_egress_desc_q_proc()

// --------------------------------------------------------------------------
//  hcc_ingress_desc_q_proc
//      Process a packet that came from the ingress descriptor queue
//
//  Description:
//      This function processes a packet on the ingress descriptor queue. The
//      packet is MUXed and sent to the appropriate HCC user to process.
//
static void hcc_ingress_desc_q_proc(
    sHCC_CBLK     *cblk,       // Control block
    sTERA_MEM_BUF *buf,        // Packet buffer
    UINT32         buf_len     // Packet buffer length
    )
{
    UINT8  chan_id;
    UINT8 *apdu = NULL;
    UINT32 apdu_len;

    if (buf == NULL)
    {
        mTERA_LOG_ERR("received 0-byte packet");
    }
    else if ((cblk != NULL) && cblk->initialized)
    {
        tera_scp_pkt_proc(cblk->dev, cblk->scp_handle, buf, buf_len, &chan_id,
                          &apdu, &apdu_len, &cblk->egress_pkt_q);

        // Continue processing if we have data (non-NULL APDU pointer)
        if (apdu != NULL)
        {
#ifdef GFUN_DBG
            mTERA_LOG_DBG("chan %d, len %d, data=%02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X",
                          chan_id, apdu_len,
                          apdu[0], apdu[1], apdu[2], apdu[3],
                          apdu[4], apdu[5], apdu[6], apdu[7],
                          apdu[8], apdu[9], apdu[10], apdu[11],
                          apdu[12], apdu[13], apdu[14], apdu[15]);
#endif
            // The SCP channel ID is the same as the HCC user ID
            switch (chan_id)
            {
                case TERA_PCOIP_HCC_APDU_USER_HMI:
                {
                    hcc_hmi_apdu_proc(cblk, apdu, apdu_len);
                    break;
                }
                case TERA_PCOIP_HCC_APDU_USER_KMP:
                {
                    hcc_kmp_apdu_proc(cblk, apdu, apdu_len);
                    break;
                }
                case TERA_PCOIP_HCC_APDU_USER_IMG:
                {
                    hcc_img_apdu_proc(cblk, apdu, apdu_len);
                    break;
                }
                default:
                {
                    mTERA_LOG_ERR("invalid HCC user (chan_id=%u)", chan_id);
                    break;
                }
            }
        }
    } // end if

    // Free the packet buffer
    tera_mem_buf_free(cblk->dev, buf);
} // hcc_ingress_desc_q_proc()

// --------------------------------------------------------------------------
//  dpc_entry
//      Entry point to the deferred procedure call (DPC) to process packets
//
//  Description:
//      The DPC will process in the following priority:
//          1. ingress descriptor queue packets
//          2. egress packet queue packets
//
static void dpc_entry(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct work_struct *unused
#else
    void *unused
#endif
    )
{
    sHCC_CBLK     *cblk;
    INT32          err;
    INT            processed;
    sTERA_MEM_BUF *buf;
    UINT32         buf_len;

    if(hcc_dev_cblk == NULL)
    {
        mTERA_LOG_CRIT("dpc_entry started before initializing hcc_dev_cblk");
    }
    else
    {
        cblk = &hcc_dev_cblk->hcc_cblk;

        do
        {
            if (!cblk->started)
            {
                //
                // We could still be running when the module is being stopped
                // because the DPC runs on a different thread than the rest of the
                // driver. In that case, we need to refrain from doing any
                // processing.
                //
                break;
            }

            processed = 0;

            // Check if there is room on the egress descriptor queue
            err = tera_gfun_tx_ready(cblk->dev);
            if (err == TERA_SUCCESS)
            {
                // Send a packet, if any, from the egress packet queue
                processed |= hcc_egress_desc_q_proc(cblk);
            }
            else if (err == TERA_ERR_FULL)
            {
                mTERA_LOG_DBG("egress descriptor queue full");
            }
            else
            {
                // Stop processing on all other errors
                break;
            }

            if (tera_gfun_rx(cblk->dev, &buf, &buf_len, NULL) == TERA_SUCCESS)
            {
                // Process the packet
                hcc_ingress_desc_q_proc(cblk, buf, buf_len);

                processed |= 1;
            }

#ifdef GFUN_DBG
            // if any messages were processed dump the GFUN registers
            if (processed)
            {
                tera_gfun_reg_dump(cblk->dev);
            }
#endif
        } while (processed);

        // re-schedule the dpc
        queue_delayed_work(hcc_wq, &cblk->work, msecs_to_jiffies(IRQ_TIMER_PERIOD_MS));
    }
} // dpc_entry()

// --------------------------------------------------------------------------
//  tera_hcc_init_phase1
///     Phase 1 initialization of the Host Control Channel (HCC) module
//
//  Description:
///     This function initializes portions of the HCC module. The device
///     control block has both an HCC and HDC control block. Initially the
///     driver does not know if it is communicating with an HCC or HDC. The
///     phase 1 initialization functions are invoked for both the HCC and HDC
///     control blocks within the device control block. After phase 1
///     initialization completes the code can read a register to determine
///     whether it is communicating with an HCC or HDC. At that point the code
///     should call the appropriate phase 2 initialization function.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_hcc_init_phase1(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32           ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk=&dev_cblk->hcc_cblk;
    UINT8           scp_chan;

    mTERA_LOG_DBG("start");

    // Initialize control block variables
    cblk->dev = dev;
    cblk->id_disp_topo = 0;
    cblk->user_cfg_setting.image_qual_pref = TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_DEFAULT;
    cblk->user_cfg_setting.build_to_lossless_enable = TERA_PCOIP_HCC_APDU_BLD_2_LOSSLESS_DEFAULT;
    cblk->pending_ioctl_flag = 0;

    // Create a pending ioctl queue
    init_waitqueue_head(&cblk->pending_ioctl_q);

    // Create a spinlock that ensures only one request is sent to the FW
    init_MUTEX(&cblk->pending_ioctl_lock);

    // Create a spinlock that protects the HCC data structures
    init_MUTEX(&cblk->data_lock);

    // Initialize the egress packet queue
    ret = tera_pkt_q_init(&cblk->egress_pkt_q);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize egress packet queue (ret=%d)",
                      ret);
    }
    else
    {
        //
        // Create a SCP instance for the HCC with a SCP channel for each of
        // the HCC users. The user ID is equivalent to the SCP channel ID.
        //
        scp_chan = ((1 << TERA_PCOIP_HCC_APDU_USER_HMI) |
                    (1 << TERA_PCOIP_HCC_APDU_USER_KMP) |
                    (1 << TERA_PCOIP_HCC_APDU_USER_IMG));

        ret = tera_scp_create(dev, scp_chan, &cblk->scp_handle,
                              scp_chan_open_cback, cblk);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to create SCP instance (ret=%d)", ret);
        }
    }
    if (ret != TERA_SUCCESS)
    {
        //TODO clean up
    }

    mTERA_LOG_DBG("end (ret=%d)", ret);

    return ret;
} // tera_hcc_init_phase1()

// --------------------------------------------------------------------------
//  tera_hcc_init_phase2
///     Phase 2 initialization of the Host Control Channel (HCC) module
//
//  Description:
///     See the comments that describe the tera_hcc_init_phase1() function.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_hcc_init_phase2(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32           ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk=&dev_cblk->hcc_cblk;

    mTERA_LOG_DBG("start");

    // Initialize control block variables
    hcc_dev_cblk = dev_cblk;

    // Initialize the HMI user
    ret = hcc_hmi_init(cblk);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize HMI user (ret=%d)", ret);
    }
    else
    {
        // Initialize the KMP user
        ret = hcc_kmp_init(cblk);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to initialize KMP user (ret=%d)", ret);
        }
        else
        {
            // Initialize the IMG user
            ret = hcc_img_init(cblk);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to initialize IMG user (ret=%d)", ret);
            }
            else
            {
                // Create a workqueue for the dpc_entry function that polls the device
                hcc_wq = create_singlethread_workqueue("hcc_wq");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
                INIT_DELAYED_WORK(&cblk->work, dpc_entry);
#else
                INIT_WORK(&cblk->work, dpc_entry, NULL);
#endif

                // All systems green
                cblk->initialized = 1;
            }
        }
    }
    if (ret != TERA_SUCCESS)
    {
        //TODO clean up
    }

    mTERA_LOG_DBG("end (ret=%d)", ret);

    return ret;
} // tera_hcc_init_phase2()

// --------------------------------------------------------------------------
//  tera_hcc_cleanup
///     HCC module clean up routine
//
//  Description:
///     This function performs the HCC module clean up routine. The system
///     timer resolution is reset back to its default value, and the module is
///     no longer marked as being initialized.
///
/// \return
///     None
///
void tera_hcc_cleanup(
    struct  pci_dev *dev    ///< Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk=&dev_cblk->hcc_cblk;

    mTERA_LOG_DBG("start");

    if (cblk->initialized)
    {
        cblk->initialized = 0;

        //TODO clean up the hmi, kmp, img users
    }

    mTERA_LOG_DBG("end");
} // tera_hcc_cleanup()

// --------------------------------------------------------------------------
//  tera_hcc_start
///     Start the HCC module
//
//  Description:
///     This function starts the HCC module. All local variable values
///     will be reset. If there were any pending feature notify ioctls, then
///     this function will complete them with a cleared feature set.
///
/// \return
///     TERA_SUCCESS
///     TERA_ERR_INVALID_STATE
///     TERA_ERR_NO_MEMORY
///
INT32 tera_hcc_start(
    struct  pci_dev *dev    ///< Device object
    )
{
    INT32           ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk=&dev_cblk->hcc_cblk;

    mTERA_LOG_DBG("start");

    if (!cblk->initialized)
    {
        mTERA_LOG_ERR("HCC module not initialized");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        if (cblk->started)
        {
            mTERA_LOG_DBG("HCC module already started");

            ret = TERA_SUCCESS;
        }
        else
        {
            //
            // Wait to ensure HDC has finished initializing. On some systems the HCC
            // starts before the HDC when coming out of the standby/sleep state. If
            // this delay is not included the FW will start writing to PCI space before
            // the HDC has finished initializing, which is a requirement. If this
            // happens the host system/workstation will crash. See BZ 6266.
            //
            ssleep(2);

            // Write to the GFUN mailbox to tell the firmware we are ready
            ret = tera_gfun_mbox_write(cblk->dev, TERA_GFUN_MBOX_SLOT_2,
                                       HCC_READY_GFUN_MBOX_VAL);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to write to GFUN mailbox (ret=%d)", ret);
            }
            else
            {
                //
                // Now that the shared components have been successfully started,
                // call the start routine of each of the users
                //

                // Start the HMI user
                ret = hcc_hmi_start(cblk);
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("failed to start HMI user (ret=%d)", ret);
                }
                else
                {
                    // Start the KMP user
                    ret = hcc_kmp_start(cblk);
                    if (ret != TERA_SUCCESS)
                    {
                        mTERA_LOG_ERR("failed to start KMP user (ret=%d)", ret);
                    }
                    else
                    {
                        // Schedule the DPC to run to send messages to Tera host
                        queue_delayed_work(hcc_wq, &cblk->work, msecs_to_jiffies(IRQ_TIMER_PERIOD_MS));
                        mTERA_LOG_INFO("HCC thread sched period = %lu jiffies, ticks/sec = %d, ticks/10 ms = %d",
                                       msecs_to_jiffies(IRQ_TIMER_PERIOD_MS), HZ, HZ/100);

                        // Ready to go!
                        cblk->started = 1;
                    }
                }
            }
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // tera_hcc_start()

// --------------------------------------------------------------------------
//  tera_hcc_stop
///     Stop the HCC module
//
//  Description:
///     This function stops the HDC module. All packets from the egress and
///     ingress descriptor queues, and the local cursor queue are freed.
///
/// \return
///     None
///
void tera_hcc_stop(
    struct  pci_dev *dev    ///< Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk=&dev_cblk->hcc_cblk;
    UINT32          reset_counter;

    mTERA_LOG_DBG("start");

    if (cblk->initialized && cblk->started)
    {
        //
        // Clear the started flag first- even if there is a error while
        // stopping the module, we'll still end up in the stopped state anyway.
        //
        cblk->started = 0;

        //
        // Clear the value in the host ready GFUN mailbox to indicate that the
        // host is going offline
        //
        tera_gfun_mbox_write(cblk->dev, TERA_GFUN_MBOX_SLOT_2, 0);

        //
        // Increment the value in the firmware reset counter GFUN mailbox to
        // tell the firmware to reset the HCC
        //
        if (tera_gfun_mbox_read(cblk->dev, TERA_GFUN_MBOX_SLOT_1,
                                &reset_counter) == TERA_SUCCESS)
        {
            mTERA_LOG_DBG("read reset_counter=%u", reset_counter);

            tera_gfun_mbox_write(cblk->dev, TERA_GFUN_MBOX_SLOT_1,
                                 reset_counter + 1);
        }
        else
        {
            //TODO What to do in the case where the mailbox read/write fails?
        }

        // Stop the dpc_entry thread
        cancel_delayed_work(&cblk->work);
        flush_workqueue(hcc_wq);
        destroy_workqueue(hcc_wq);

        // Stop the HMI, KMP and IMG users (note: nothing to do for IMG)
        hcc_hmi_stop(cblk);
        hcc_kmp_stop(cblk);

        // Delete all packets from the egress packet queue
        tera_pkt_q_purge(dev, &cblk->egress_pkt_q);

        //
        // Reset the SCP - we do this AFTER we stop the users because the
        // users may still need to send packets to the firmware
        //
        tera_scp_reset(cblk->scp_handle);
    } // end if

    mTERA_LOG_DBG("end");
} // tera_hcc_stop()

// --------------------------------------------------------------------------
//  tera_hcc_ioctl_proc
///     Process an ioctl request for the HCC device interface
//
//  Description:
///     This function processes an ioctl request for the HCC device interface.
///     If the ioctl is unknown, then it is completed with STATUS_NOT_SUPPORTED.
///
/// \return
///     None
///
INT32 tera_hcc_ioctl_proc(
    struct  pci_dev *dev,
    unsigned int  ioctl,
    unsigned long arg
    )
{
    INT32           ret;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sHCC_CBLK      *cblk = &dev_cblk->hcc_cblk;

    if ((cblk == NULL) || !cblk->initialized)
    {
        mTERA_LOG_ERR("not initialized");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (!cblk->started)
    {
        mTERA_LOG_ERR("not started");
        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        // Switch based on the user ID encoded into the ioctl
        switch (mTERA_PCOIP_DEV_IOCTL_USER_ID_GET(ioctl))
        {
            case TERA_PCOIP_HCC_APDU_USER_HMI:
            {
                ret = hcc_hmi_ioctl_proc(cblk, arg, ioctl);
                break;
            }
            case TERA_PCOIP_HCC_APDU_USER_KMP:
            {
                ret = hcc_kmp_ioctl_proc(cblk, arg, ioctl);
                break;
            }
            case TERA_PCOIP_HCC_APDU_USER_IMG:
            {
                ret = hcc_img_ioctl_proc(cblk, arg, ioctl);
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported ioctl (0x%08x) (user=%u)", ioctl,
                              mTERA_PCOIP_DEV_IOCTL_USER_ID_GET(ioctl));
                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }

    return ret;
} // tera_hcc_ioctl_proc()

