////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Memory Module Implementation
///
///     This file implements the memory pool module. This module provides APIs
///     to easily allocate DMA'able memory. This module also creates a memory
///     pool of DMA'able memory. The memory pool has an advantage of lowering
///     fragmentation of system memory, and also for decreasing alloc/free
///     instructions.
///
///     DMA'able memory MUST be used when sharing memory between software and
///     hardware. However, DMA'able memory is a scarce resource and should not
///     be used for any purpose other than to talk to hardware.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_mem.h"
#include "tera_mem_internal.h"

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

static sTERA_MEM_DMA *mem_alloc(
    struct pci_dev *dev,
    size_t          len
    );

static INT32 mem_pool_create(
    sTERA_DEV_CBLK *dev_cblk
    );

static void mem_pool_destroy(
    struct pci_dev *dev
    );

// --------------------------------------------------------------------------
//  mem_alloc
///     Allocate DMA'able memory
//
//  Description:
///     This function allocates DMA'able memory. This function can be called
///     once the DMA enabler in the control block is initialized.
//
static sTERA_MEM_DMA *mem_alloc(
    struct pci_dev *dev,
    size_t          len
    )
{
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(dev);
    sTERA_MEM_DMA  *buf = NULL;

    // Allocate memory for a memory structure
    buf = (sTERA_MEM_DMA *)kzalloc(sizeof(sTERA_MEM_DMA), GFP_KERNEL);
    if (buf == NULL)
    {
        mTERA_LOG_ERR("Failed to alloc memory for a memory structure with size=%d",
                      (UINT32)sizeof(sTERA_MEM_DMA));
    }
    else
    {
        buf->virt_addr = pci_alloc_consistent(dev, len, &(buf->phys_addr));
        if (buf->virt_addr == NULL)
        {
            mTERA_LOG_ERR("failed to allocate dma-able buffer from OS");
            return NULL;
        }

        // clear the contents of the buffer
        memset(buf->virt_addr, 0, len);

        // Use the system opaque pointer to record usage
        buf->sys = (void *)1;

        buf->len = len;
        cblk->mem_cblk.mem_alloc_cnt++;

        mTERA_LOG_DBG("end: virt addr 0x%lX, dma_addr: 0x%X_0x%X",
                      (unsigned long)buf->virt_addr,
                      (UINT32)(((UINT64)buf->phys_addr >> 32) & 0xFFFFFFFF),
                      (UINT32)(buf->phys_addr & 0xFFFFFFFF));
    }

    return buf;
} // mem_alloc()

// --------------------------------------------------------------------------
//  mem_pool_create
///     Create the DMA'able memory pool
//
//  Description:
///     This function creates the DMA'able memory pool. The function allocates
///     multiple DMA'able buffers and puts them in a structure maintained by
///     the memory manager.
//
static INT32 mem_pool_create(
    sTERA_DEV_CBLK *dev_cblk           ///< Device control block
    )
{
//    sTERA_POOL_CBLK *cblk = &(dev_cblk->pool_cblk);
    sTERA_MEM_CBLK  *cblk = &dev_cblk->mem_cblk;
    UINT32           i;
    INT32            ret = TERA_SUCCESS;

    mTERA_LOG_DBG("start");

    do
    {
        // Initialize all the chunks
        cblk->pool.size = TERA_NUM_CHUNKS * TERA_CHUNK_SIZE;
        for (i = 0; i < TERA_NUM_CHUNKS; i++)
        {
            // Allocate a common buffer for the pool
            cblk->pool.chunk[i].virt_addr = pci_alloc_consistent(dev_cblk->pdev,
                                                                 TERA_CHUNK_SIZE,
                                                                 &(cblk->pool.chunk[i].phys_addr));
            if (cblk->pool.chunk[i].virt_addr == NULL)
            {
                mTERA_LOG_ERR("failed to allocate buffer %d for pool", i);
                ret = TERA_ERR_FAILURE;
                break;
            }

            // Use the avail flag to record usage
            cblk->pool.chunk[i].avail = 1;
        }
    } while (0);

    return ret;
} // mem_pool_create()

// --------------------------------------------------------------------------
//  mem_pool_destroy
///     Destroy the DMA'able memory pool
//
//  Description:
///     This function destroys the DMA'able memory pool. This function frees
///     the region of DMA'able memory used for the memory pool.
//
static void mem_pool_destroy(
    struct pci_dev *dev
    )
{
    UINT32          i;
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(dev);
    sTERA_MEM_CBLK *mem_cblk = &cblk->mem_cblk;

    mTERA_LOG_DBG("start");

    if (mem_cblk->initialized)
    {
        // Free all the chunks
        for (i = 0; i < TERA_NUM_CHUNKS; i++)
        {
            pci_free_consistent(dev,
                                TERA_CHUNK_SIZE,
                                mem_cblk->pool.chunk[i].virt_addr,
                                mem_cblk->pool.chunk[i].phys_addr);
        }
        mem_cblk->initialized = 0;
    }
} // mem_pool_destroy()

// --------------------------------------------------------------------------
//  tera_mem_init
///     Initialize the memory module
//
//  Description:
///     This function initializes the memory module. This function creates a
///     pool of DMAable memory buffers.
///
/// \return
///     TERA_SUCCESS - success
///     TERA_ERR_FAILURE - error
///
INT32 tera_mem_init(
    struct pci_dev *dev
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sTERA_MEM_CBLK *cblk = &dev_cblk->mem_cblk;

    mTERA_LOG_DBG("start");

    cblk->initialized = 0;

    // Pre-allocate a pool of DMA'able memory
    if (mem_pool_create(dev_cblk) != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to create memory pool");
    }
    else
    {
        cblk->initialized = 1;
    }

    mTERA_LOG_DBG("end");

    return (cblk->initialized) ? TERA_SUCCESS : TERA_ERR_FAILURE;
} // tera_mem_init()

// --------------------------------------------------------------------------
//  tera_mem_exit
///     Clean up the memory module
//
//  Description:
///     This function cleans up the memory module. This function frees all
///     DMA'able memory used by the memory pool.
///
/// \return
///     TERA_SUCCESS
///
INT32 tera_mem_exit(
    struct pci_dev *dev
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);

    mTERA_LOG_DBG("start");

    if (dev_cblk->mem_cblk.initialized)
    {
        mem_pool_destroy(dev);

        mTERA_LOG_INFO("mem_alloc_cnt=%u, mem_free_cnt=%u",
                       dev_cblk->mem_cblk.mem_alloc_cnt,
                       dev_cblk->mem_cblk.mem_free_cnt);
        mTERA_LOG_INFO("buf_alloc_cnt=%u, buf_free_cnt=%u",
                       dev_cblk->mem_cblk.buf_alloc_cnt,
                       dev_cblk->mem_cblk.buf_free_cnt);
    }

    return TERA_SUCCESS;
} // tera_mem_exit()

// --------------------------------------------------------------------------
//  tera_mem_alloc
///     Allocate DMA'able memory
//
//  Description:
///     This function allocates DMA'able memory from the system, which can be
///     accessed by both the software and hardware. DMA'able memory is a scarce
///     resource and should only be used when needed.
///
/// \return
///     Handle to DMA'able memory structure
///
sTERA_MEM_DMA *tera_mem_alloc(
    struct pci_dev *dev,            ///< Device object
    size_t          len             ///< Length of memory to allocate
    )
{
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(dev);

    return (cblk->mem_cblk.initialized) ? mem_alloc(dev, len) : NULL;
} // tera_mem_alloc()

// --------------------------------------------------------------------------
//  tera_mem_free
///     Free DMA'able memory
//
//  Description:
///     This function frees a DMA'able memory buffer back to the system.
///     After the memory has been freed, the user must NOT access the
///     memory again.
///
/// \return
///     None
///
void tera_mem_free(
    struct pci_dev *dev,        ///< device control block
    sTERA_MEM_DMA  *buf         ///< buffer to free
    )
{
    sTERA_MEM_CBLK *cblk;

    if ((buf != NULL) && (dev != NULL))
    {
        cblk = (sTERA_MEM_CBLK *)buf->cblk;
        cblk->mem_free_cnt++;

        pci_free_consistent(dev,
                            buf->len,
                            buf->virt_addr,
                            buf->phys_addr);

        kfree(buf);
    }
} // tera_mem_free()

// --------------------------------------------------------------------------
//  tera_mem_buf_alloc
///     Allocate a DMA'able memory buffer
//
//  Description:
///     This function allocates a DMA'able memory buffer from the memory pool.
///     If successful, this function returns a handle to a memory buffer. If
///     the pool is exhausted or if the requested length is larger than that of
///     a memory buffer, then this function returns NULL.
///
/// \return
///     Handle to DMA'able memory buffer structure
///
sTERA_MEM_BUF  *tera_mem_buf_alloc(
    struct pci_dev *dev,            ///< Device object
    size_t          len             ///< Length of memory buffer
    )
{
    sTERA_DEV_CBLK *cblk;
    sTERA_MEM_BUF  *ret = NULL;
    UINT32          i;

    while (dev != NULL)
    {
        cblk = pci_get_drvdata(dev);
        if ((cblk == NULL) || (!cblk->mem_cblk.initialized))
        {
            mTERA_LOG_ERR("invalid args (cblk NULL or cblk not initialized");
            break;
        }
        if (len > TERA_CHUNK_SIZE)
        {
            mTERA_LOG_ERR("requested length (%lu) too big", (unsigned long)len);
            break;
        }

        // Look for a free buffer from the memory pool
        for (i = 0; i < TERA_NUM_CHUNKS; i++)
        {
            if (cblk->mem_cblk.pool.chunk[i].avail)
            {
                ret = &(cblk->mem_cblk.pool.chunk[i]);

                // Mark the chunk as used
                ret->avail = 0;
                break;
            }
        }

        if (ret == NULL)
        {
            mTERA_LOG_ERR("memory pool exhausted - alloc counter %u",
                          cblk->mem_cblk.mem_alloc_cnt);
            break;
        }

        // Clear the any leftover contents in the common buffer
        memset(ret->virt_addr, 0, TERA_CHUNK_SIZE);

        cblk->mem_cblk.buf_alloc_cnt++;
        break;
    }
#if 0
    mTERA_LOG_DBG("end: virt addr 0x%lX, dma_addr: 0x%X_0x%X",
                  (unsigned long)ret->virt_addr,
                  (UINT32)(((UINT64)ret->phys_addr >> 32) & 0xFFFFFFFF),
                  (UINT32)(ret->phys_addr & 0xFFFFFFFF));
#endif

    return ret;
} // tera_mem_buf_alloc()

// --------------------------------------------------------------------------
//  tera_mem_buf_free
///     Free a DMA'able memory buffer
//
//  Description:
///     This function frees a DMA'able memory buffer. The memory buffer is
///     returned back to the memory pool. After the memory has been freed, the
///     user must NOT access the memory again.
///
/// \return
///     None
///
void tera_mem_buf_free(
    struct pci_dev *dev,        ///< Device object
    sTERA_MEM_BUF  *buf         ///< Handle to DMA'able memory buffer structure
    )
{
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(dev);

    if ((cblk != NULL) &&
        (cblk->mem_cblk.initialized) &&
        (buf != NULL))
    {
        cblk->mem_cblk.buf_free_cnt++;
        buf->avail = 1;
    }
} // tera_mem_buf_free()
