////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Configuration Data Class Implementation
///
///     This file implements the configuration data class. The purpose of
///     this class is to hide the details of storing and retrieving
///     configuration data that determines how the application is configured
///     to operate.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include "tera_errors.h"
#include "tera_log.h"
#include "tera_cfg.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_CFG::cTERA_CFG
///     Constructor
//
//  Description:
///     This function is the class contructor. This function initializes the
///     member data variables.
///
/// \return
///     None
///
cTERA_CFG::cTERA_CFG(void)
{
    mTERA_LOG_DBG("start");

    // Initialize the member data variables
    m_config = NULL;

    mTERA_LOG_DBG("end");
} // cTERA_CFG::cTERA_CFG()

// --------------------------------------------------------------------------
//  cTERA_CFG::cTERA_CFG
///     Destructor
//
//  Description:
///     This function is the class destructor. This function cleans the
///     member data variables.
///
/// \return
///     None
///
cTERA_CFG::~cTERA_CFG(void)
{
    mTERA_LOG_DBG("start");

    if (m_config != NULL)
    {
        delete m_config;
    }

    mTERA_LOG_DBG("end");
} // cTERA_CFG::~cTERA_CFG()

// --------------------------------------------------------------------------
//  cTERA_CFG::init
///     Initialize this object
//
//  Description:
///     This function initializes the config module. The caller specifies the
///     name of the local and global configuration files.
///
/// \return
///     TERA_SUCCESS
///
INT32 cTERA_CFG::init(
    const wxString *global_filename,
    const wxString *local_filename
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    if (m_config != NULL)
    {
        delete m_config;
    }

    // Set the application configuration
    m_config = new wxFileConfig(PCOIP_AGENT_REG_KEY_APPNAME,
                                PCOIP_AGENT_REG_KEY_COMPANY,
                                *local_filename,
                                *global_filename,
                                wxCONFIG_USE_LOCAL_FILE | wxCONFIG_USE_GLOBAL_FILE);

    ret = TERA_SUCCESS;

    mTERA_LOG_DBG("end: local %S, global %S",
                  local_filename->c_str(),
                  global_filename->c_str());

    return ret;
} // cTERA_CFG::init()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_setting_int
///     Read integer configuration setting
//
//  Description:
///     This function reads an integer configuration setting. It first
///     tries to read the setting from the local configuration file.
///     If the setting is not found the function reads the setting from the
///     global configuration file. If the setting is not found in the global
///     file a default value is assigned.
///
void cTERA_CFG::get_setting_int(
    const wchar_t *setting_key,
    const wxString *local_path,
    const wxChar *global_path,
    int bad_local_val,
    int default_val,
    int *cfg_value
    )
{
    wxString    global_path_str;

    // try reading the setting from the local config file
    m_config->SetPath(*local_path);
    *cfg_value = m_config->Read(setting_key, bad_local_val);
#if CFG_DBG
    mTERA_LOG_DBG("local path %S, setting = %d, bad = %d", local_path->c_str(), *cfg_value, bad_local_val);
#endif

    // if the setting is not in the local config read it from the global file
    if (*cfg_value == bad_local_val)
    {
        global_path_str.Printf(global_path);
        m_config->SetPath(global_path_str);
        *cfg_value = m_config->Read(setting_key, default_val);
#if CFG_DBG
        mTERA_LOG_DBG("global path %S, setting = %d", global_path_str.c_str(), *cfg_value);
#endif
    }

#if CFG_DBG
    mTERA_LOG_DBG("end: %S = %d", setting_key, *cfg_value);
#endif

} // cTERA_CFG::get_setting_int()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_setting_int
///     Read integer configuration setting
//
//  Description:
///     This function performs the same function as the get_setting_int()
///     function found above this. The difference between the two functions
///     is the type of parameters passed to the function.
///
void cTERA_CFG::get_setting_int(
    const wchar_t *setting_key,
    const wxChar *local_path,
    const wxChar *global_path,
    int bad_local_val,
    int default_val,
    int *cfg_value
    )
{
    wxString    local_path_str;

    // convert local path parameter to a wxString and call the other version
    // of this function
    local_path_str.Printf(local_path);
    get_setting_int(setting_key,
                                 &local_path_str,
                                 global_path,
                                 bad_local_val,
                                 default_val,
                                 cfg_value);

} // cTERA_CFG::get_setting_int()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_ws_lock
///     Write the lock workstation config setting
//
//  Description:
///     This function writes the lock workstation configuration setting to
///     the local configuration file.
///
void cTERA_CFG::set_ws_lock(
    bool lock_enable
    )
{
    mTERA_LOG_DBG("start: lock enable %d", lock_enable);

    m_config->SetPath(PCOIP_AGENT_LOCAL_CFG_FEAT_SECTION);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LOCK_WS, lock_enable);
    m_config->Flush();

} // cTERA_CFG::set_ws_lock()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_ws_lock
///     Read lock workstation config setting
//
//  Description:
///     This function reads the lock workstation configuration setting.
///
bool cTERA_CFG::get_ws_lock(void)
{
    bool        lock_en;
    int         cfg_value;

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LOCK_WS,
                                 PCOIP_AGENT_LOCAL_CFG_FEAT_SECTION,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LOCK_WS_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LOCK_WS_DEFAULT,
                                 &cfg_value);

    lock_en = (cfg_value != 0) ? 1 : 0;

    return lock_en;

} // cTERA_CFG::get_ws_lock()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_lcur_lkbd_en_setting
///     Write the local cursor and keyboard enable config setting
//
//  Description:
///     This function writes the local cursor and keyboard enable configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_lcur_lkbd_en_setting(
    const char *client_mac,
    int lcur_lkbd_en
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: lcur en %d", lcur_lkbd_en);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LCUR_LKBD_EN, lcur_lkbd_en);
    m_config->Flush();

} // cTERA_CFG::set_lcur_lkbd_en_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_lcur_lkbd_en_setting
///     Read local cursor/keyboard enable setting from the config file
//
//  Description:
///     This function reads the local cursor and keyboard enable
///     configuration setting.
///
void cTERA_CFG::get_lcur_lkbd_en_setting(
    const char *client_mac,
    int *lcur_lkbd_en
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LCUR_LKBD_EN,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_LKBD_EN_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_LKBD_EN_DEFAULT,
                                 lcur_lkbd_en);

    mTERA_LOG_DBG("end: local cursor/keyboard enable %d", *lcur_lkbd_en);

} // cTERA_CFG::get_lcur_lkbd_en_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_lcur_vis_setting
///     Write the visible cursor config setting
//
//  Description:
///     This function writes the visible cursor configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_lcur_vis_setting(
    const char *client_mac,
    int lcur_vis
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: visible cursor %d", lcur_vis);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LCUR_VIS, lcur_vis);
    m_config->Flush();

} // cTERA_CFG::set_lcur_vis_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_lcur_vis_setting
///     Read visible cursor setting from the config file
//
//  Description:
///     This function reads the visible cursor configuration setting.
///
void cTERA_CFG::get_lcur_vis_setting(
    const char *client_mac,
    int *lcur_vis
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LCUR_VIS,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_VIS_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_VIS_DEFAULT,
                                 lcur_vis);

#if CFG_DBG
    mTERA_LOG_DBG("end: visible cursor(s) %d", *lcur_vis);
#endif

} // cTERA_CFG::get_lcur_vis_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_lcur_fade_setting
///     Write the local cursor fade enable config setting
//
//  Description:
///     This function writes the local cursor fade enable configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_lcur_fade_setting(
    const char *client_mac,
    int lcur_fade
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: local cursor fade enable %d", lcur_fade);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LCUR_FADE, lcur_fade);
    m_config->Flush();

} // cTERA_CFG::set_lcur_fade_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_lcur_fade_setting
///     Read local cursor fade enable setting from the config file
//
//  Description:
///     This function reads the local cursor fade enable configuration setting.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_NOT_FOUND
///
void cTERA_CFG::get_lcur_fade_setting(
    const char *client_mac,
    int *lcur_fade
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LCUR_FADE,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_FADE_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LCUR_FADE_DEFAULT,
                                 lcur_fade);

    mTERA_LOG_DBG("end: local cursor fade enable %d", *lcur_fade);

} // cTERA_CFG::get_lcur_fade_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_repeat_en_setting
///     Write the keyboard repeat enable config setting
//
//  Description:
///     This function writes the keyboard repeat enable configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_repeat_en_setting(
    const char *client_mac,
    int lkbd_repeat_en
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: keyboard repeat enable %d", lkbd_repeat_en);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LKBD_REPEAT_EN, lkbd_repeat_en);
    m_config->Flush();

} // cTERA_CFG::set_repeat_en_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_repeat_en_setting
///     Read keyboard repeat enable setting from the config file
//
//  Description:
///     This function reads the keyboard repeat enable configuration setting.
///
void cTERA_CFG::get_repeat_en_setting(
    const char *client_mac,
    int *lkbd_repeat_en
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LKBD_REPEAT_EN,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_REPEAT_EN_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_REPEAT_EN_DEFAULT,
                                 lkbd_repeat_en);

    mTERA_LOG_DBG("end: keyboard repeat enable %d", *lkbd_repeat_en);

} // cTERA_CFG::get_repeat_en_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_repeat_delay_setting
///     Write the keyboard repeat delay config setting
//
//  Description:
///     This function writes the keyboard repeat delay configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_repeat_delay_setting(
    const char *client_mac,
    int lkbd_repeat_delay
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: keyboard repeat delay %d", lkbd_repeat_delay);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LKBD_DELAY, lkbd_repeat_delay);
    m_config->Flush();

} // cTERA_CFG::set_repeat_delay_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_repeat_delay_setting
///     Read keyboard repeat delay setting from the config file
//
//  Description:
///     This function reads the keyboard repeat delay configuration setting.
///
void cTERA_CFG::get_repeat_delay_setting(
    const char *client_mac,
    int *lkbd_repeat_delay
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LKBD_DELAY,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_DELAY_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_DELAY_DEFAULT,
                                 lkbd_repeat_delay);

    mTERA_LOG_DBG("end: keyboard delay %d", *lkbd_repeat_delay);

} // cTERA_CFG::get_repeat_delay_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_repeat_rate_setting
///     Write the keyboard repeat rate config setting
//
//  Description:
///     This function writes the keyboard repeat rate configuration
///     setting to the local configuration file.
///
void cTERA_CFG::set_repeat_rate_setting(
    const char *client_mac,
    int lkbd_repeat_rate
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: keyboard repeat rate %d", lkbd_repeat_rate);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);
    m_config->Write(PCOIP_AGENT_CFG_FEAT_LKBD_RATE, lkbd_repeat_rate);
    m_config->Flush();

} // cTERA_CFG::set_repeat_rate_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_repeat_rate_setting
///     Read keyboard repeat rate setting from the config file
//
//  Description:
///     This function reads the keyboard repeat rate configuration setting.
///
void cTERA_CFG::get_repeat_rate_setting(
    const char *client_mac,
    int *lkbd_repeat_rate
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_PORTAL_FEAT_SECTION, client_mac);

    get_setting_int(PCOIP_AGENT_CFG_FEAT_LKBD_RATE,
                                 &usr_cfg_path,
                                 PCOIP_AGENT_GLOBAL_CFG_SECTION,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_RATE_INVALID,
                                 PCOIP_AGENT_CFG_FEAT_LKBD_RATE_DEFAULT,
                                 lkbd_repeat_rate);

    mTERA_LOG_DBG("end: keyboard rate %d", *lkbd_repeat_rate);

} // cTERA_CFG::get_repeat_rate_setting()

// --------------------------------------------------------------------------
//  cTERA_CFG::delete_mon_data
///     Delete monitor settings from the config file
//
//  Description:
///     This function deletes the monitor configuration settings.
///
INT32 cTERA_CFG::delete_mon_data(
    const char *client_mac,
    int max_num_mon
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: client MAC %s", client_mac);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_MON_SECTION,
                        client_mac);
    m_config->DeleteGroup(usr_cfg_path);

    for (int i = 0; i < max_num_mon; i++)
    {
        usr_cfg_path.Printf(PCOIP_AGENT_CFG_MONX_SECTION,
                            client_mac,
                            i+1);
        m_config->DeleteGroup(usr_cfg_path);
    }
    m_config->Flush();

    mTERA_LOG_DBG("end");

    return TERA_SUCCESS;
} // cTERA_CFG::delete_mon_data()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_monitor_screen_counts
///     Write the monitor and screen counters to the config file
//
//  Description:
///     This function writes the monitor and screen count configuration
///     settings to the configuration file.
///
void cTERA_CFG::set_monitor_screen_counts(
    const char *client_mac,
    unsigned int monitor_count,
    unsigned int screen_count
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: # monitors %d # screens %d",
                  monitor_count, screen_count);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_MON_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);

    m_config->Write(PCOIP_AGENT_CFG_MON_COUNT, (int)monitor_count);
    m_config->Write(PCOIP_AGENT_CFG_SCREEN_COUNT, (int)screen_count);

    m_config->Flush();

} // cTERA_CFG::set_monitor_screen_counts()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_monitor_screen_counts
///     Read monitor and screen counters from the config file
//
//  Description:
///     This function reads the monitor and screen count configuration
///     settings.
///
void cTERA_CFG::get_monitor_screen_counts(
    const char *client_mac,
    unsigned int *monitor_count,
    unsigned int *screen_count
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_MON_SECTION, client_mac);
    m_config->SetPath(usr_cfg_path);

    *monitor_count = m_config->Read(PCOIP_AGENT_CFG_MON_COUNT,
                                    PCOIP_AGENT_CFG_MON_COUNT_DEFAULT);
    *screen_count  = m_config->Read(PCOIP_AGENT_CFG_SCREEN_COUNT,
                                    PCOIP_AGENT_CFG_SCREEN_COUNT_DEFAULT);

    mTERA_LOG_DBG("end: # monitors %d # screens %d",
                  *monitor_count, *screen_count);

} // cTERA_CFG::get_monitor_screen_counts()

// --------------------------------------------------------------------------
//  cTERA_CFG::set_monitor_info
///     Write monitor information to the config file
//
//  Description:
///     This function writes information about a monitor to the configuration
///     file.
///
void cTERA_CFG::set_monitor_info(
    const char *client_mac,
    unsigned int monitor_num,
    wxString *vid,
    wxString *pid,
    wxString *serial,
    int size_x,
    int size_y,
    int pos_x,
    int pos_y,
    unsigned int monitor
    )
{
    wxString    usr_cfg_path;

    mTERA_LOG_DBG("start: num %d sz %dx%d pos %d,%d monitor %d",
                  monitor_num, size_x, size_y, pos_x, pos_y, monitor);

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_MONX_SECTION, client_mac, monitor_num);
    m_config->SetPath(usr_cfg_path);

    m_config->Write(PCOIP_AGENT_CFG_MONX_VID, *vid);
    m_config->Write(PCOIP_AGENT_CFG_MONX_PID, *pid);
    m_config->Write(PCOIP_AGENT_CFG_MONX_SERIAL, *serial);
    m_config->Write(PCOIP_AGENT_CFG_MONX_SIZE_X, size_x);
    m_config->Write(PCOIP_AGENT_CFG_MONX_SIZE_Y, size_y);
    m_config->Write(PCOIP_AGENT_CFG_MONX_POS_X, pos_x);
    m_config->Write(PCOIP_AGENT_CFG_MONX_POS_Y, pos_y);
    m_config->Write(PCOIP_AGENT_CFG_MONX_MON, (int)monitor);

    m_config->Flush();

} // cTERA_CFG::set_monitor_info()

// --------------------------------------------------------------------------
//  cTERA_CFG::get_monitor_info
///     Read information about the specified monitor from the config file
//
//  Description:
///     This function reads information about a specified monitor from the
///     config file.
///
void cTERA_CFG::get_monitor_info(
    const char *client_mac,
    unsigned int monitor_num,
    wxString *vid,
    wxString *pid,
    wxString *serial,
    int *size_x,
    int *size_y,
    int *pos_x,
    int *pos_y,
    unsigned int *monitor
    )
{
    wxString    usr_cfg_path;

    usr_cfg_path.Printf(PCOIP_AGENT_CFG_MONX_SECTION, client_mac, monitor_num);
    m_config->SetPath(usr_cfg_path);

    *vid     = m_config->Read(PCOIP_AGENT_CFG_MONX_VID,
                              PCOIP_AGENT_CFG_MONX_VID_DEFAULT);
    *pid     = m_config->Read(PCOIP_AGENT_CFG_MONX_PID,
                              PCOIP_AGENT_CFG_MONX_PID_DEFAULT);
    *serial  = m_config->Read(PCOIP_AGENT_CFG_MONX_SERIAL,
                              PCOIP_AGENT_CFG_MONX_SERIAL_DEFAULT);
    *size_x  = m_config->Read(PCOIP_AGENT_CFG_MONX_SIZE_X,
                              PCOIP_AGENT_CFG_MONX_SIZE_X_DEFAULT);
    *size_y  = m_config->Read(PCOIP_AGENT_CFG_MONX_SIZE_Y,
                              PCOIP_AGENT_CFG_MONX_SIZE_Y_DEFAULT);
    *pos_x   = m_config->Read(PCOIP_AGENT_CFG_MONX_POS_X,
                              PCOIP_AGENT_CFG_MONX_POS_X_DEFAULT);
    *pos_y   = m_config->Read(PCOIP_AGENT_CFG_MONX_POS_Y,
                              PCOIP_AGENT_CFG_MONX_POS_Y_DEFAULT);
    *monitor = m_config->Read(PCOIP_AGENT_CFG_MONX_MON,
                              PCOIP_AGENT_CFG_MONX_MON_DEFAULT);

    mTERA_LOG_DBG("end: num %d sz %dx%d pos %d,%d monitor %d",
                  monitor_num, *size_x, *size_y, *pos_x, *pos_y, *monitor);

} // cTERA_CFG::get_monitor_info()
