////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Network Interface Class Implementation
///
///     This file implements the network interface class. This class finds all
///     the active Ethernet and wireless network interfaces.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/unistd.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <arpa/inet.h>

#include "tera_log.h"
#include "tera_mem.h"
#include "tera_errors.h"
#include "tera_netif.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// ---- Protected Functions -------------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_NETIF::if_list_entry_add
//      Add an entry to the list of network interfaces
//
//  Description:
//      This function adds an entry to the tail of the list of the network
//      interfaces. The entry MUST be valid as this funtion does not do any
//      validation.
//
void cTERA_NETIF::if_list_entry_add(
    sNETIF *entry
    )
{
    sNETIF *node = m_if_list;

    assert(entry != NULL);

    if (m_if_list == NULL)
    {
        m_if_list = entry;
    }
    else
    {
        // Traverse the list until we reach the end
        while (node->next != NULL)
        {
            node = node->next;
        }

        node->next = entry;
    }

    // Increment the list entry count
    m_if_list_count++;
} // cTERA_NETIF::if_list_entry_add()

// --------------------------------------------------------------------------
//  cTERA_NETIF::if_list_clear
//      Clear all entries from the list of network interfaces
//
//  Description:
//      This function clears all entries from the list of network interfaces.
//      The memory allocated for each entry is freed. The user must make sure
//      there are no rogue pointers pointing to those entries. This function
//      also resets the list count.
//
void cTERA_NETIF::if_list_clear(void)
{
    sNETIF *entry;
    UINT32  i = 0;

    mTERA_LOG_DBG("start - list contains '%u' entries", m_if_list_count);

    // Go through all entries and free the memory starting from the head
    while (m_if_list != NULL)
    {
        entry = m_if_list;
        m_if_list = m_if_list->next;

        mTERA_MEM_FREE(entry);
        i++;
    }

    // Reset the list pointer and entry count
    m_if_list = NULL;
    m_if_list_count = 0;

    mTERA_LOG_DBG("end - cleared '%u' entries", i);
} // cTERA_NETIF::if_list_clear()

// ---- Public Functions ----------------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_NETIF::cTERA_NETIF
///     Constructor
//
//  Description:
///     This function is the class contructor and initializes the local
///     variables.
///
/// \return
///     None
///
cTERA_NETIF::cTERA_NETIF(void)
{
    // Initialize the member data variables
    m_if_list = (sNETIF *)NULL;
    m_if_list_count = 0;

    // Look for Ethernet/wireless network interfaces
    if_query();
} // cTERA_NETIF::cTERA_NETIF()

// --------------------------------------------------------------------------
//  cTERA_NETIF::~cTERA_NETIF
///     Destructor
//
//  Description:
///     This function performs clean up of this class, such as freeing the
///     memory used for the network interface list.
///
/// \return
///     None
///
cTERA_NETIF::~cTERA_NETIF(void)
{
    // Free up any memory used for the internal interface list
    if_list_clear();

} // cTERA_NETIF::~cTERA_NETIF()

// --------------------------------------------------------------------------
//  cTERA_NETIF::if_query
///     Query Windows for active Ethernet and wireless network interfaces
//
//  Description:
///     This function queries Windows for active Ethernet and wireless network
///     interfaces and stores their friendly name, IP and MAC addresses in the
///     internal list. This function should be called whenever a change on any
///     of the network interfaces occur.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 cTERA_NETIF::if_query(void)
{
    INT32         ret = TERA_ERR_FAILURE;
    struct ifconf ifconf;
    struct ifreq  ifreq[32];
    int           sock = -1;
    sNETIF       *entry = NULL;

    mTERA_MEM_ZERO(&ifconf, sizeof(ifconf));
    ifconf.ifc_req = ifreq;
    ifconf.ifc_len = sizeof(ifreq);

    // Acquire lock
    m_list_lock.Lock();

    do
    {
        if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        {
            mTERA_LOG_ERR("failed to open socet to AF_INET");

            break;
        }

        if (ioctl(sock, SIOCGIFCONF, &ifconf) < 0)
        {
            mTERA_LOG_ERR("failed to get list of interfaces");

            break;
        }

        ret = TERA_SUCCESS;

        //
        // Go through all the detected interfaces and store the info of the
        // ones that are up and are not loopback devices
        //
        for (UINT32 i = 0; i < ifconf.ifc_len / sizeof(struct ifreq); i++)
        {
            // Get the flags for the interface
            if (ioctl(sock, SIOCGIFFLAGS, &ifreq[i]) < 0)
            {
                // Continue with next interface
                continue;
            }

            if ((ifreq[i].ifr_flags & IFF_LOOPBACK) ||
                !(ifreq[i].ifr_flags & IFF_UP))
            {
                // Ignore loopback or down interfaces
                continue;
            }

            // Allocate a new network interface entry in our internal list
            entry = (sNETIF *)mTERA_MEM_CALLOC(sizeof(sNETIF));
            if (entry == NULL)
            {
                mTERA_LOG_ERR("failed to allocate memory %u");

                ret = TERA_ERR_FAILURE;
                break;
            }

            // Get the MAC address
            if (ioctl(sock, SIOCGIFHWADDR, &ifreq[i]) < 0)
            {
                mTERA_MEM_FREE(entry);

                continue;
            }

            // Copy the MAC address into the entry
            entry->mac[0] = ifreq[i].ifr_hwaddr.sa_data[0];
            entry->mac[1] = ifreq[i].ifr_hwaddr.sa_data[1];
            entry->mac[2] = ifreq[i].ifr_hwaddr.sa_data[2];
            entry->mac[3] = ifreq[i].ifr_hwaddr.sa_data[3];
            entry->mac[4] = ifreq[i].ifr_hwaddr.sa_data[4];
            entry->mac[5] = ifreq[i].ifr_hwaddr.sa_data[5];

            // Get the IPv4 address
            if (ioctl(sock, SIOCGIFADDR, &ifreq[i]) < 0)
            {
                mTERA_MEM_FREE(entry);

                continue;
            }

            // Convert the IP address and store it into the entry
            snprintf(entry->ip, TERA_NETIF_IP_MAX_LEN, "%s",
                     inet_ntoa(((struct sockaddr_in *)(&ifreq[i].ifr_addr))->sin_addr));

            // Copy the interface name into the entry as well
            snprintf(entry->name, TERA_NETIF_NAME_MAX_LEN, "%s",
                     ifreq[i].ifr_name);

            mTERA_LOG_DBG("*******************");
            mTERA_LOG_DBG("%s", entry->name);
            mTERA_LOG_DBG("\tip address %s", entry->ip);
            mTERA_LOG_DBG("\tmac address %02X-%02X-%02X-%02X-%02X-%02X",
                          entry->mac[0], entry->mac[1], entry->mac[2],
                          entry->mac[3], entry->mac[4], entry->mac[5]);
            mTERA_LOG_DBG("*******************");

            // Add the entry to our network interface list
            if_list_entry_add(entry);
        }
    } while (0);

    if (sock >= 0)
    {
        // Close any open sockets
        close(sock);
    }

    // Release lock
    m_list_lock.Unlock();

    mTERA_LOG_DBG("found '%u' interfaces:", if_count());

    return ret;
} // cTERA_NETIF::if_query()

// --------------------------------------------------------------------------
//  cTERA_NETIF::if_count
///     Return the number of Ethernet and wireless network interfaces
//
//  Description:
///     This function returns the number of Ethernet and wireless network
///     interfaces that are currently active.
///
/// \return
///     0,
///     Number of network interfaces
///
UINT32 cTERA_NETIF::if_count(void)
{
    UINT32 count;

    // Acquire lock
    m_list_lock.Lock();

    count = m_if_list_count;

    // Release lock
    m_list_lock.Unlock();

    return count;
} // cTERA_NETIF::if_count()

// --------------------------------------------------------------------------
//  cTERA_NETIF::if_by_index
///     Get the network interface at the specified index in the list
//
//  Description:
///     This function gets the information of the network interface at the
///     specified index in the list of network interfaces.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_FOUND
///
INT32 cTERA_NETIF::if_by_index(
    UINT32 index,
    char   name[TERA_NETIF_NAME_MAX_LEN],
    char   ip[TERA_NETIF_IP_MAX_LEN],
    UINT8  mac[6]
    )
{
    sNETIF *entry = m_if_list;
    INT32   ret = TERA_ERR_NOT_FOUND;

    // Acquire lock
    m_list_lock.Lock();

    if ((ip != NULL) && (mac != NULL) && (index < m_if_list_count))
    {
        // Traverse the list to the entry at the specified index
        while ((index != 0) && (entry != NULL))
        {
            entry = entry->next;
            index--;
        }

        // Only continue if we have found the requested entry
        if ((index == 0) && (entry != NULL))
        {
            // Copy the name, IP address, and MAC address to the user buffers
            if (name != NULL)
            {
                strncpy(name, entry->name, TERA_NETIF_NAME_MAX_LEN);
            }

            strncpy(ip, entry->ip, TERA_NETIF_IP_MAX_LEN);

            mac[0] = entry->mac[0];
            mac[1] = entry->mac[1];
            mac[2] = entry->mac[2];
            mac[3] = entry->mac[3];
            mac[4] = entry->mac[4];
            mac[5] = entry->mac[5];

            ret = TERA_SUCCESS;
        }
    }

    // Release lock
    m_list_lock.Unlock();

    return ret;
} // cTERA_NETIF::if_by_index()

