////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software GUI Class
///
///     This file implements the PCoIP Host Software GUI class. This class is
///     the main GUI class. All other GUI related classes and objects hang off
///     of this class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/extensions/Xinerama.h>

#include <wx/icon.h>
#include <wx/config.h>
#include <wx/display.h>

#include "tera_log.h"
#include "tera_mem.h"
#include "tera_types.h"
#include "tera_errors.h"
#include "tera_pcoip_agent_internal.h"
#include "tera_pcoip_agent_gui.h"
#include "tera_pcoip_agent_icon_res.xpm"
#include "tera_sinst.h"
#include "tera_cfg.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// This constant determines how often the code calls the heartbeat() function.
// Currently the code calls heartbeat() every second (1000 ms).
#define PROCESS_HEARTBEAT_RATE          (1000/m_rcs_period)

// The following counters are used to determine when specific processing
// should be performed.
static int heartbeat_timer_cnt = 0;

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI, wxFrame)
    EVT_TIMER(wxID_ANY, cPCOIP_AGENT_GUI::evt_timer)
    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI::evt_init_dialog)
    EVT_CLOSE(cPCOIP_AGENT_GUI::evt_close)
    EVT_END_SESSION(cPCOIP_AGENT_GUI::evt_end_session)
    EVT_MENU(wxID_EXIT, cPCOIP_AGENT_GUI::OnExit)
END_EVENT_TABLE()

// Defines the max number of times the heartbeat thread will attempt to send
// topology messages when a session becomes connected or the display topology
// changes before giving up. Note the heartbeat function assumes this is an
// even number. It only calls the connect_proc() function every 2 seconds.
// This is done to allow the FW a chance to process new topology messages. The
// constant must be even because the agent needs to wait more than one second
// before checking the feature status after a topology message is sent. When
// the agent does not wait long enough the status returned by the FW is not
// always correct.
#define TERA_MAX_SEND_TOPO		(3*2)

// Define the minimum number of times the heartbeat thread will attempt to
// send a topology message to the firmware. When a systems screensaver becomes
// active or the displays are turned off by the Windows power manager the
// FW might report a status change. The firmware also reports a status
// change when the screensaver is turned off or the monitors are turned on.
#define TERA_MIN_SEND_TOPO      1

// Define the number of times the heartbeat thread will wait before creating
// the tray icon. Initially the host SW created the tray icon in the
// PCOIP_AGENT_GUI constructor. Sometimes the icon appeared as a vertical line
// or a copy of another icon in the tray. This problem occurred more frequently
// on SLED 11 SP 2. Delaying the creation of the icon partially fixed the
// problem. The problem still occurs but it's much less frequent.
#define TERA_SYS_TRAY_ICON_DELAY        4

// This constant determines how many times the host SW will attempt to read the
// client topology settings after a session starts before giving up. The host
// does not return valid cached topology information if imaging is not active,
// which is why the topology may need to be read multiple times
#define READ_CLIENT_TOPO_RETRY_CNT      5

// The following strings are used by the host SW to create the command string
// executed when the host SW locks the screen. This happens when the host SW
// detects a PCoIP session becomes disconnected.
#define SCREEN_LOCK_SCRIPTNAME        "pcoip_lockscreen"
#define DEFAULT_SCREEN_LOCK_SCRIPT    "/usr/bin/pcoip_lockscreen"

// ---- Local Variables -----------------------------------------------------

// agent state string table
static const char *agent_state_str_tbl[TERA_AGENT_GUI_STATE_MAX] = {
    "TERA_AGENT_GUI_DISCONNECTED_STATE",
    "TERA_AGENT_GUI_SESS_START_STATE",
    "TERA_AGENT_GUI_READ_CLIENT_TOPO_STATE",
    "TERA_AGENT_GUI_READ_HOST_TOPO_STATE",
    "TERA_AGENT_GUI_PUSH_TOPO_STATE",
    "TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE",
    "TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE"
};

// The following constants and array are used by the host software when there
// is a problem that prevents the client from displaying a specific cursor
// image. When this happens the host software sends this bitmap to the client.
// Currently this image is an arrow with the letter "L" inside of it.
//
#define TRS  (0x00000000)
#define WHT  (0xFFFFFFFF)
#define BLK  (0xFF000000)
#define PCOIP_AGENT_DEFAULT_CUR_SHAPE_WIDTH   (12)
#define PCOIP_AGENT_DEFAULT_CUR_SHAPE_HEIGHT  (20)
#define PCOIP_AGENT_DEFAULT_CUR_SHAPE_HOT_X   (0)
#define PCOIP_AGENT_DEFAULT_CUR_SHAPE_HOT_Y   (0)

static UINT32 pcoip_agent_default_cur_shape[PCOIP_AGENT_DEFAULT_CUR_SHAPE_HEIGHT][PCOIP_AGENT_DEFAULT_CUR_SHAPE_WIDTH] =
{
    BLK, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, BLK, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, WHT, BLK, TRS, TRS, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, WHT, WHT, BLK, TRS, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, WHT, WHT, BLK, TRS, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, WHT, WHT, WHT, BLK, TRS, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, WHT, WHT, WHT, WHT, BLK, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, WHT, WHT, WHT, WHT, WHT, BLK, TRS, TRS, TRS,
    BLK, WHT, BLK, WHT, WHT, WHT, WHT, WHT, WHT, BLK, TRS, TRS,
    BLK, WHT, BLK, BLK, BLK, WHT, WHT, WHT, WHT, WHT, BLK, TRS,
    BLK, WHT, WHT, WHT, WHT, WHT, WHT, BLK, BLK, BLK, BLK, BLK,
    BLK, WHT, WHT, BLK, WHT, WHT, WHT, BLK, TRS, TRS, TRS, TRS,
    BLK, WHT, BLK, TRS, BLK, WHT, WHT, WHT, BLK, TRS, TRS, TRS,
    BLK, BLK, TRS, TRS, BLK, WHT, WHT, WHT, BLK, TRS, TRS, TRS,
    BLK, TRS, TRS, TRS, TRS, BLK, WHT, WHT, WHT, BLK, TRS, TRS,
    TRS, TRS, TRS, TRS, TRS, BLK, WHT, WHT, WHT, BLK, TRS, TRS,
    TRS, TRS, TRS, TRS, TRS, TRS, BLK, WHT, WHT, WHT, BLK, TRS,
    TRS, TRS, TRS, TRS, TRS, TRS, BLK, WHT, WHT, WHT, BLK, TRS,
    TRS, TRS, TRS, TRS, TRS, TRS, TRS, BLK, BLK, BLK, TRS, TRS
};

// Contains the value of the last cursor point retrieved from the host OS. The
// host SW uses this to determine when the host cursor moved.
static UINT16 last_host_cur_abs_x = 0;
static UINT16 last_host_cur_abs_y = 0;

// PCoIP Host card device type string table
static const char *device_type_str_tbl[MAX_TERA_HOST_DEVICE_TYPE] = {
    "TERA1x00",
    "TERA22x0",
};

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::evt_init_dialog
//      Process the init event
//
//  Description:
//      This function processes the INIT_DIALOG event, which is sent when the
//      frame is about to be created. This function does any UI base
//      initialization.
//
void cPCOIP_AGENT_GUI::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    mTERA_LOG_DBG("start");

    evt.Skip();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI::evt_init_dialog()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::evt_close
//      Process the CLOSE event
//
//  Description:
//      This function processes the CLOSE event, which is sent when the frame
//      is about to be destroyed. This funciton tidies up all lose ends before
//      the frame gets destroyed.
//
void cPCOIP_AGENT_GUI::evt_close(
    wxCloseEvent& evt
    )
{
    mTERA_LOG_DBG("start - cPCOIP_AGENT_GUI");

    // Close the property sheet
    m_prop_sheet.Close();

    end_session();

    evt.Skip();

    mTERA_LOG_DBG("end - cPCOIP_AGENT_GUI");
} // cPCOIP_AGENT_GUI::evt_close()

void cPCOIP_AGENT_GUI::evt_end_session(
    wxCloseEvent& evt
    )
{
    mTERA_LOG_DBG("start");

    end_session();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI::evt_end_session()

void cPCOIP_AGENT_GUI::evt_timer(
    wxTimerEvent& evt
    )
{
    INT                                    sess_link_change;
    int                                    mon_info_ok;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

    // perform heartbeat processing if necessary
    heartbeat_timer_cnt = (heartbeat_timer_cnt + 1) % PROCESS_HEARTBEAT_RATE;
    if (heartbeat_timer_cnt == (PROCESS_HEARTBEAT_RATE - 1))
    {
        // provide a heartbeat to the PCoIP link object
        sess_link_change = 0;
        m_pcoip_link.link_heartbeat(&sess_link_change,
                                    &sess_link);

        //
        // If the number of screens or the screen resolution changes
        // erase the monitor configuration data in the config file. This
        // is done to force the user to re-configure the monitor settings.
        //
        if(screen_topology_changed())
        {
            // Only delete the monitor config settings if a session is active
            // and more than one screen is active OR the cached topology is
            // valid and the number of monitors is not equal to 1.
            if  ((sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) &&
                 ((m_screen_count != 1) ||
                  ((m_client_cache_status == TERA_SUCCESS) &&
                   (m_disp_topo.num_disp != 1))))
            {
                mTERA_LOG_DBG("Deleting monitor config data from config file");

                // delete the monitor information for the attached client
                m_prop_sheet.m_mon_page->delete_mon_usr_cfg();
                host_mon_info_invalidate();
            }

            // if a property sheet is open, close and reopen it
            if (m_prop_sheet.IsShown())
            {
                m_prop_sheet.Close();
                m_prop_sheet.Show(true);
            }
        }


        if (sess_link_change)
        {
            link_state_change_proc(sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED);
        }

        // call the heartbeat function, which performs the required state specific
        // processing
        heartbeat();
    }

    if ((m_lcur_lkbd_en) && (m_rcs_enabled))
    {
        // Check if the cursor moved. When it moves send the host cursor
        // coordinates to the driver. The driver sends an update to the client
        // if the host cursor location differs from the recently received
        // client cursor reports.
        update_client_cur_pos();
    }
} // cPCOIP_AGENT_GUI::evt_timer()

void cPCOIP_AGENT_GUI::OnExit(wxCommandEvent& WXUNUSED(event))
{
    mTERA_LOG_DBG("start");

    end_session();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI::OnExit()

void cPCOIP_AGENT_GUI::OnFatalException(void)
{
    mTERA_LOG_DBG("start");

    end_session();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI::OnFatalException()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::link_state_change_proc
//      Function called when the session link changes
//
//  Description:
//      This function is called when the session link state changes from
//      connected to disconnected or vice-a-versa.
//
void cPCOIP_AGENT_GUI::link_state_change_proc(
    UINT8 connected)
{
    // when the session state changes initialize the variable that tracks
    // the state of the local cursor feature
    set_lcur_lkbd_mode(0);

    //
    // If the session was just disconnected.
    //
    if (!(connected))
    {
        mTERA_LOG_DBG("session state change detected: disconnected");

        // Save the session link state
        m_sess_connected = 0;

        // Close the HDC
        m_pcoip_link.hdc_close();

        // set state to disconnected state
        m_agent_state = TERA_AGENT_GUI_DISCONNECTED_STATE;

        // Clear the copy of the client topology information. This forces the
        // system to check the monitor information of the next client that
        // connects to the host.
        mTERA_MEM_ZERO(&m_disp_topo, sizeof(m_disp_topo));
        m_client_cache_status = TERA_ERR_INVALID_ARG;

        // if the settings are configured to lock the workstation
        if (wxGetApp().gui().host_cfg().get_ws_lock())
        {
            // if a screen lock utility exists execute it
            if (!m_lock_utility.IsEmpty())
            {
                mTERA_LOG_DBG("locking workstation using '%S'", m_lock_utility.c_str());
                wxExecute(m_lock_utility, wxEXEC_ASYNC);
            }
            else
            {
                mTERA_LOG_CRIT("unable to lock screen because screen lock utility was not found!!!");
            }
        }

        // if the GUI is open close it
        if (m_prop_sheet.IsShown())
        {
            m_prop_sheet.Close();
#if 0 // TBD - Including this can causes problems when disconnecting a session.
      //       The statement is included in case we need to test the problem (bug
      //       7780)
            m_prop_sheet.Show(true);
#endif
        }

        // reset flag that indicates whether the topology was sent successfully
        set_topo_sent_status(0);
    }
    else
    {
        // If the previous state was connected the workstation might need to be
        // locked. This can happen if a session is established immediately
        // after a session is disconnected (the session disconnect and
        // reconnect must happen within a 1 second window).
        if (m_sess_connected)
        {
            // Lock the workstation if the agent setting is configured
            if (wxGetApp().gui().host_cfg().get_ws_lock())
            {
                // if a screen lock utility exists execute it
                if (!m_lock_utility.IsEmpty())
                {
                    mTERA_LOG_DBG("locking workstation on back-to-back connected sessions");
                    wxExecute(m_lock_utility, wxEXEC_ASYNC);
                }
                else
                {
                    mTERA_LOG_CRIT("back-to-back: unable to lock screen because screen lock utility was not found!!!");
                }
            }
        }

        // Clear the copy of the client topology information. This forces the
        // system to check the monitor information of the client connected to
        // the host.
        mTERA_MEM_ZERO(&m_disp_topo, sizeof(m_disp_topo));
        m_client_cache_status = TERA_ERR_INVALID_ARG;

        // Save the session link state
        m_sess_connected = 1;

        // Open the HDC channel
        if (m_pcoip_link.hdc_open() != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to open HDC channel");
        }
        else
        {
            // set state to session start
            m_agent_state = TERA_AGENT_GUI_SESS_START_STATE;
        }

        // re-initialize the monitor list (this actually just clears the
        // monitor list)
        m_prop_sheet.m_mon_page->mon_list_init();
    }
} // cPCOIP_AGENT_GUI::link_state_change_proc()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::heartbeat
//      Heartbeat function
//
//  Description:
//      This function carries out tasks that should be done on a regular
//      periodic interval. Such tasks include checking for any host monitor
//      configuration changes, and for any network interface changes.
//
void cPCOIP_AGENT_GUI::heartbeat(void)
{
	UINT32	                            stat;
	UINT32                              cnt;
    INT32                               ret;

    if (m_agent_state < TERA_AGENT_GUI_STATE_MAX)
    {
        mTERA_LOG_DBG("start state = %s", agent_state_str_tbl[m_agent_state]);
    }
    else
    {
        mTERA_LOG_ERR("invalid agent state = %d", m_agent_state);
    }

    // create tray icon if it isn't created yet and the countdown delay is one
    if (m_icon_create_delay)
    {
        m_icon_create_delay--;
        if (m_icon_create_delay <= 0)
        {
            // Create a system tray icon
            m_tray.SetIcon(wxICON(IDI_LOGO_16), _T("PCoIP Host Software"));
            if (!m_tray.IsOk())
            {
                mTERA_LOG_ERR("Icon initialization failed");
            }
        }
    }
    // Close host SW if the icon becomes uninstalled. This happens when a user
    // logs in or logs out.
    else if (!m_tray.IsIconInstalled())
    {
        mTERA_LOG_DBG("Icon disappeared, closing agent");
        end_session();
        Close();
    }

    switch (m_agent_state)
    {
        case TERA_AGENT_GUI_DISCONNECTED_STATE:
            break;

        case TERA_AGENT_GUI_SESS_START_STATE:
            // after a session starts send the keyboard repeat settings
            send_kbd_repeat_settings();

            //
            // Read the settings and store the results. The results are used
            // later to determine whether the client supports the cached
            // topology APDU.
            //
            m_client_cache_status = client_cache_topology_get();

            if ((m_client_cache_status == TERA_ERR_IMG_NOT_ACTIVE) ||
                (m_client_cache_status == TERA_ERR_TIMEOUT))
            {
                // set state to read the client topology again
                mTERA_LOG_DBG("Changing state to read client topo again");
                m_agent_state = TERA_AGENT_GUI_READ_CLIENT_TOPO_STATE;
                m_read_client_topo_retry = READ_CLIENT_TOPO_RETRY_CNT;
            }
            else
            {
                // Don't try to send the client topology to the host
                mTERA_LOG_DBG("Sending topo to client: topo status=%d, # client disp=%d",
                              m_client_cache_status,
                              m_disp_topo.num_disp);

                // set state to send topology to the client
                m_agent_state = TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE;
                m_client_send_topo_retry = TERA_MAX_SEND_TOPO;
                display_topology_send();
            }
            break;

        case TERA_AGENT_GUI_READ_CLIENT_TOPO_STATE:
            //
            // Read the settings and store the results. The results are used
            // later to determine whether the client supports the cached
            // topology APDU.
            //
            m_client_cache_status = client_cache_topology_get();

            if (((m_client_cache_status == TERA_ERR_IMG_NOT_ACTIVE) ||
                 (m_client_cache_status == TERA_ERR_TIMEOUT)) &&
                (m_read_client_topo_retry != 0))
            {
                m_read_client_topo_retry--;
            }
            //
            // Send topology to client if the cached topology was successfully
            // read OR the maximum attempts have been reached
            //
            else
            {
                mTERA_LOG_DBG("Not pushing client topo: client topo status=%d, read retry=%d, # client disp=%d",
                              m_client_cache_status,
                              m_read_client_topo_retry,
                              m_disp_topo.num_disp);

                // set state to send topology to the client
                m_agent_state = TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE;
                m_client_send_topo_retry = TERA_MAX_SEND_TOPO;
	            display_topology_send();
            }
            break;

        case TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE:
            mTERA_LOG_DBG("m_client_send_topo_retry=%d", m_client_send_topo_retry);

            // retry connecting every 2 seconds
	        if (m_client_send_topo_retry % 2)
	        {
		        // retry connection processing
		        connect_proc();

		        // if the property sheet is open
                if (m_prop_sheet.IsShown())
		        {
			        //
			        // refresh the feature page when the retry count is zero or after
			        // the first failed attempt to send the topology
			        //
			        if ((!m_client_send_topo_retry) || (m_client_send_topo_retry == (TERA_MAX_SEND_TOPO - 2)))
			        {
//				        m_prop_sheet.m_page_feat.PostMessage(WM_NETIF_DISP_CHANGE_NOTIFY);
                        m_prop_sheet.Close();
                        m_prop_sheet.Show(true);
			        }
		        }
	        }
	        else
	        {
		        m_client_send_topo_retry--;
                if (m_client_send_topo_retry <= 0)
                {
                    m_agent_state = TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE;
                }
	        }
            break;

        case TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE:
            mTERA_LOG_DBG("last feature status (cnt=%d stat=0x%X), client send topo retry=%d",
                          m_last_feat_stat_cnt,
                          m_last_feat_stat,
                          m_client_send_topo_retry);
	        //
	        // Read the last feature status message received by the driver. This
	        // code is in place to handle systems where the graphics driver does
	        // not notify Windows when there is a change in the display topology.
	        // In this situation the FW sends an asynchronous FEATURE STATUS APDU
	        // with new status. It also takes care of problems that arise when
	        // disconnecting a monitor.
	        //
	        if (m_pcoip_link.lcur_lkbd_last_feat_stat_get(&stat, &cnt) == TERA_SUCCESS)
	        {
		        //
		        // if new status has been received force a topology update. This is
		        // required to get the local cursor working after the monitor(s)
		        // are turned off and then back on
                //
                // The second check handles the scenario where the status has
                // changed. If this is not included the system can get into a
                // state where the FW reports good status and the local cursor
                // is disabled even though the user has enabled the feature.
		        //
		        if ((cnt > m_last_feat_stat_cnt) ||
                    (stat != m_last_feat_stat))
		        {
			        mTERA_LOG_DBG("Rcvd new status (0x%08X => 0x%08X)- start topo chg (cnt %d)",
						          m_last_feat_stat, stat, cnt);

                    //
                    // Reread the cached topology if the APDU is supported.
                    // This handles the scenario where a monitor has been
                    // connected or disconnected.
                    //
                    if (m_client_cache_status != TERA_ERR_NOT_SUPPORTED)
                    {
                        m_client_cache_status = client_cache_topology_get();
                    }

			        // this variable triggers the sending of a topology update
			        m_client_send_topo_retry = TERA_MIN_SEND_TOPO;

                    //
                    // Resend the topology to the client if the status did not
                    // change. This is necessary when both the old and new
                    // status equal 0. If the topology is not sent in this
                    // situation the client will never be sent an updated copy
                    // of the current topology. This can cause problems when
                    // two displays are connected and the host SW previously
                    // sent topology data for only one monitor. (BZ 5939)
                    //
			        if ((m_last_feat_stat == stat) && (stat == 0))
                    {
                        display_topology_send();
                    }
                    else
                    {
			            m_last_feat_stat = stat;
                    }

			        //
			        // Increment the last status count. This must be done to
			        // prevent the host SW from contantly sending topology
			        // updates. The FW sends an asynchronous feature status message
			        // after it receives a topology update. The incremented value
			        // accounts for this new message.
			        //
			        m_last_feat_stat_cnt = cnt + 1;

			        //
			        // If the status did not change and the property sheet is open
			        // close it. Earlier versions tried to refresh it but this
			        // resulted in an endless loop. When the screen is refreshed
			        // and the monitors are off an APDU is sent to the FW disabling
			        // the local cursor. This triggers the FW to send another feature
			        // status message, but for some reason the err field equals 0
			        // (success). When the agent receives this it then sends an APDU
			        // enabling the local cursor, and this repeats...
			        //
                    if (m_prop_sheet.IsShown())
			        {
			            m_prop_sheet.Close();
			            mTERA_LOG_DBG("Closing GUI");
			        }

                    m_agent_state = TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE;
                }
	        }
            break;
        default:
            mTERA_LOG_ERR("Unsupported state %d", m_agent_state);
            break;
    }
} // cPCOIP_AGENT_GUI::heartbeat

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::send_kbd_repeat_settings
//      Send the keyboard repeat settings to the client
//
//  Description:
//      This function sends the keyboard repeat settings to the client if the
//      client supports the feature. The function tries to read the settings
//      from the client. If the settings are retrieved from the client these
//      settings will be used. If no settings are available at the client then
//      the function will retrieve the settings from the registry. This
//      function also updates the feature tab to use the active settings.
//
void cPCOIP_AGENT_GUI::send_kbd_repeat_settings(void)
{
    UINT16                                 lkbd_delay;
    UINT16                                 lkbd_rate;
    INT32                                  client_rep_data_avail = TERA_ERR_TIMEOUT;
    INT32                                  ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lkbd_repeat_delay_cfg;
    int                                    lkbd_repeat_rate_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    // Configure the client repeat settings if the client supports them
    if(m_pcoip_link.lkbd_client_auto_repeat_avail() == TERA_SUCCESS)
    {
        // Set the keyboard repeat rate and delay settings on the GUI
        client_rep_data_avail = m_pcoip_link.lkbd_repeat_rates_get(&lkbd_delay, &lkbd_rate);
        if (client_rep_data_avail != TERA_SUCCESS)
        {
            // uses the registry if the settings cannot be read from the client
            wxGetApp().gui().host_cfg().get_repeat_delay_setting(sess_link.mac,
                                                                 &lkbd_repeat_delay_cfg);
            wxGetApp().gui().host_cfg().get_repeat_rate_setting(sess_link.mac,
                                                                &lkbd_repeat_rate_cfg);
            lkbd_delay = (UINT16)lkbd_repeat_delay_cfg;
            lkbd_rate  = (UINT16)lkbd_repeat_rate_cfg;
            mTERA_LOG_DBG("using registry keyboard repeat setting: delay=%d, rate=%d",
                          lkbd_delay, lkbd_rate);
        }
        else
        {
            mTERA_LOG_DBG("using client keyboard repeat setting: delay=%d, rate=%d",
                          lkbd_delay, lkbd_rate);
        }

        // update the settings on the feature page
        m_prop_sheet.m_feat_page->lkbd_delay_set(lkbd_delay);
        m_prop_sheet.m_feat_page->lkbd_rate_set(lkbd_rate);

        // write the settings to the client
        ret = m_pcoip_link.lkbd_repeat_rates_set(m_prop_sheet.m_feat_page->lkbd_delay_get(),
                                                 m_prop_sheet.m_feat_page->lkbd_rate_get());
        if (ret != TERA_SUCCESS)
        {
	        // error occurred while setting the client keyboard repeat settings
	        mTERA_LOG_ERR("failure setting client keyboard repeat settings (delay=%d, rate=%d)",
                          lkbd_delay, lkbd_rate);
        }
    }
    //
    // configure the feature sheet variable that tracks whether the
    // client supports reading kbd repeat settings
    //
    m_prop_sheet.m_feat_page->client_kbd_settings_avail_set(client_rep_data_avail);

} // cPCOIP_AGENT_GUI::send_kbd_repeat_settings

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::screen_topology_changed
//      Returns true if the screen topology changed
//
//  Description:
//      This function returns true if the number of screens changes or if
//      the resolution of the screens changes.
//
bool cPCOIP_AGENT_GUI::screen_topology_changed(void)
{
    bool                ret = false;
    int                 num_x_screens;
    XineramaScreenInfo *xsi;
    int                 num_xinerama_screens;

    if (m_sess_connected)
    {
        if (m_screen_count != get_num_screens())
        {
            ret = true;
        }
        else
        {
            num_x_screens = ScreenCount(m_xdisplay);
//            mTERA_LOG_DBG("Number of X screens = %d", num_x_screens);

            // if Xinerama is active (this is true when Xinerama is enabled AND
            // also when TwinView is enabled)
            if (XineramaIsActive(m_xdisplay))
            {
//                mTERA_LOG_DBG("Xinerama is active");
                xsi = XineramaQueryScreens(m_xdisplay, &num_xinerama_screens);
                if ((xsi) && (num_xinerama_screens > 0))
                {
                    for (UINT32 i = 0; i < m_screen_count; i++)
                    {
                        if((m_screen_cfg[i].size.x != xsi[i].width)  ||
                           (m_screen_cfg[i].size.y != xsi[i].height) ||
                           (m_screen_cfg[i].pos.x  != xsi[i].x_org)  ||
                           (m_screen_cfg[i].pos.y  != xsi[i].y_org))
                        {
                            ret = true;
                            mTERA_LOG_DBG("active disp%d: %dx%d at (%d, %d)",
                                          i, xsi[i].width, xsi[i].height,
                                          xsi[i].x_org, xsi[i].y_org);
                            mTERA_LOG_DBG("config disp%d: %dx%d at (%d, %d)",
                                          i, m_screen_cfg[i].size.x, m_screen_cfg[i].size.y,
                                          m_screen_cfg[i].pos.x, m_screen_cfg[i].pos.y);
                        }
                    }
                }
                else
                {
                    mTERA_LOG_ERR("Xinerama active but num screens = %d or XSI NULL", num_xinerama_screens);
                }

                if (xsi)
                {
                    XFree(xsi);
                }
            }
            // if only one X screen is active
            else if (num_x_screens == 1)
            {
                wxDisplay mon(0);
                wxRect    rect = mon.GetGeometry();

                if((m_screen_cfg[0].size.x != rect.width) ||
                   (m_screen_cfg[0].size.y != rect.height) ||
                   (m_screen_cfg[0].pos.x != rect.x) ||
                   (m_screen_cfg[0].pos.y != rect.y))
                {
                    ret = true;
                    mTERA_LOG_DBG("active disp%d: %dx%d at (%d, %d)",
                                  0, rect.width, rect.height,
                                  rect.x, rect.y);
                    mTERA_LOG_DBG("config disp0: %dx%d at (%d, %d)",
                                  m_screen_cfg[0].size.x, m_screen_cfg[0].size.y,
                                  m_screen_cfg[0].pos.x, m_screen_cfg[0].pos.y);
                }
            }
            // more than one X screen is active
            else
            {
                // TBD: The code needs to support reading the position of each
                // screen. Until a solution is found for this problem the code
                // will not support enabling the local cursor and keyboard
                // feature when more than one X screen is active and both
                // Xinerama and TwinView are disabled.
                mTERA_LOG_CRIT("The host SW does not support systems with multiple X screens!");
#if 0
                for (UINT32 i = 0; i < m_screen_count; i++)
                {
                    if((m_screen_cfg[i].size.x != DisplayWidth(m_xdisplay, i)) ||
                       (m_screen_cfg[i].size.y != DisplayHeight(m_xdisplay, i)))
                    {
                        ret = true;
                        mTERA_LOG_DBG("active disp%d: %dx%d",
                                      i, DisplayWidth(m_xdisplay, i), DisplayHeight(m_xdisplay, i));
                        mTERA_LOG_DBG("config disp%d: %dx%d at (%d, %d)",
                                      i, m_screen_cfg[i].size.x, m_screen_cfg[i].size.y,
                                      m_screen_cfg[i].pos.x, m_screen_cfg[i].pos.y);
                    }
                }
#endif
            }
        }

        // update the stored screen configuration data when it changes
        if(ret == true)
        {
            mTERA_LOG_DBG("screen topology change detected");
            screen_topology_init();

            //
            // If the gui is in the TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE
            // send the topology, otherwise this will happen when the
            // system eventually enters this state.
            //
            // The code also sends the topology when the state equals
            // TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE. This is done to avoid a
            // race condition observed on a Win7 64-bit system where the display
            // rotation of 2 displays were changed at the same time. When this
            // happens two display change messages are received but only the first
            // causes the host SW to send a topology message to the FW. What's
            // bad is incorrect topology is sent to the FW because it only
            // reflects the topology when one of the 2 displays has been changed.
            // This results in the local cursor not being in sync with the host
            // cursor.
            //
            if ((m_agent_state == TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE) ||
                (m_agent_state == TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE))
            {
                //
                // change the state so the host SW eventually sends the latest
                // cursor and keyboard settings to the client, this is necessary
                // for the soft client
                //
                m_agent_state = TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE;

                //
                // Set the retry counter which forces the heartbeat process to check
                // the feature status and possibly resend the topology multiple times
                //
                m_client_send_topo_retry = TERA_MAX_SEND_TOPO;

                // Send the current display topology to the firmware
                display_topology_send();
            }
        }
    }

    return ret;
} // cPCOIP_AGENT_GUI::screen_topology_changed()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::get_num_screens
//      Read the number of active screens
//
//  Description:
//      This function reads the number of screens. This is done differently
//      for different video modes (Xinerama, TwinView, multiple X Screens).
//
UINT32 cPCOIP_AGENT_GUI::get_num_screens(void)
{
    UINT32              num_screens = 0;
    int                 j;
    int                 k;
    int                 num_x_screens;
    int                 num_xinerama_screens;
    XineramaScreenInfo *xsi;

    m_xdisplay = (Display *)wxGetDisplay();
    if (NULL == m_xdisplay)
    {
        mTERA_LOG_CRIT("wxGetDisplay() failed, unable to read host topology settings");
    }
    else
    {
        num_x_screens = ScreenCount(m_xdisplay);

        // if Xinerama is active (this is true when Xinerama is enabled and
        // also when TwinView is enabled)
        if (XineramaIsActive(m_xdisplay))
        {
//            mTERA_LOG_DBG("Xinerama is active");
            xsi = XineramaQueryScreens(m_xdisplay, &num_xinerama_screens);
            if ((xsi) && (num_xinerama_screens > 0))
            {
                num_screens = num_xinerama_screens;
                //mTERA_LOG_DBG("number Xinerama Screens = %d", num_screens);
            }
            else
            {
                mTERA_LOG_ERR("Xinerama active but num screens = %d or XSI NULL", num_xinerama_screens);
            }

            if (xsi)
            {
                XFree(xsi);
            }
        }
        // if only one X screen is active
        else
        {
            num_screens = num_x_screens;
//            mTERA_LOG_DBG("Number of X screens = %d", num_screens);
        }
    }
    return num_screens;
} // cPCOIP_AGENT_GUI::get_num_screens()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::screen_topology_init
//      Store the current screen configuration settings
//
//  Description:
//      This function reads the current screen configuration settings and
//      saves them.
//
void cPCOIP_AGENT_GUI::screen_topology_init(void)
{
    int                 j=0;
    int                 k;
    int                 num_x_screens;
    int                 num_xinerama_screens;
    XineramaScreenInfo *xsi;

#if 0
    // The following code is another attempt at determining the screen
    // topology. This code currently does not provide any benefits over the
    // code in this function. It's included here in case it can be updated to
    // handle some of the multi-monitor topology problems in the code.
    GdkDisplay         *gdk_disp = gdk_display_get_default();
    GdkScreen          *gdk_screen;
    GdkRectangle        gdk_rect;
    int                 num_gdk_screens = gdk_display_get_n_screens(gdk_disp);
    int                 num_mon;
    mTERA_LOG_DBG("gdk_display_get_n_screens() = %d", num_gdk_screens);
    for (int screen=0; screen<num_gdk_screens; screen++)
    {
        gdk_screen = gdk_display_get_screen(gdk_disp, screen);
        num_mon = gdk_screen_get_n_monitors(gdk_screen);
        for(int mon=0; mon<num_mon; mon++)
        {
            gdk_screen_get_monitor_geometry(gdk_screen, mon, &gdk_rect);
            mTERA_LOG_DBG("screen %d, mon %d: size %d X %d, pos (%d, %d)",
                          screen,
                          mon,
                          gdk_rect.width,
                          gdk_rect.height,
                          gdk_rect.x,
                          gdk_rect.y);
        }
    }
#endif

    // initialize screen count to zero
    m_screen_count = 0;

    m_xdisplay = (Display *)wxGetDisplay();
    if (NULL == m_xdisplay)
    {
        mTERA_LOG_CRIT("wxGetDisplay() failed, unable to read host topology settings");
    }
    else
    {
        num_x_screens = XScreenCount(m_xdisplay);
        mTERA_LOG_DBG("Number X Screens = %d, wxDisplay::GetCount() = %d",
                      num_x_screens,
                      wxDisplay::GetCount());

        // if Xinerama is active (this is true when Xinerama is enabled and
        // also when TwinView is enabled)
        if (XineramaIsActive(m_xdisplay))
        {
            mTERA_LOG_DBG("Xinerama is active");
            xsi = XineramaQueryScreens(m_xdisplay, &num_xinerama_screens);
            if ((xsi) && (num_xinerama_screens > 0))
            {
                m_primary_width  = xsi[0].width;
                m_primary_height = xsi[0].height;
                for (k=0; k<num_xinerama_screens; k++)
                {
                    mTERA_LOG_DBG("Xscreen%d, xsi[%d]: %d x %d @ (%d, %d) screen_num %d",
                                  j,
                                  k,
                                  xsi[k].width,
                                  xsi[k].height,
                                  xsi[k].x_org,
                                  xsi[k].y_org,
                                  xsi[k].screen_number);
                    if (((TERA1x00_HOST_CARD == m_dev_type) && (m_screen_count + k < 2)) ||
                        ((TERA2240_HOST_CARD == m_dev_type) && (m_screen_count + k < 4)))
                    {
                        m_screen_cfg[m_screen_count + k].size.x = xsi[k].width;
                        m_screen_cfg[m_screen_count + k].size.y = xsi[k].height;
                        m_screen_cfg[m_screen_count + k].pos.x  = xsi[k].x_org;
                        m_screen_cfg[m_screen_count + k].pos.y  = xsi[k].y_org;
                    }
                    else
                    {
                        mTERA_LOG_ERR("Error: index exceeds max - count=%d, k=%d, num_x=%d, num_xin=%d",
                                      m_screen_count,
                                      k,
                                      num_x_screens,
                                      num_xinerama_screens);
                    }
                }
                m_screen_count = m_screen_count + num_xinerama_screens;
            }

            if (xsi)
            {
                XFree(xsi);
            }
        }
        // if one monitor is active
        else if (num_x_screens == 1)
        {
            for (unsigned i = 0; i < wxDisplay::GetCount(); i++)
            {
                wxDisplay mon(i);
                wxRect    rect = mon.GetGeometry();
                m_screen_cfg[i].size.x = rect.width;
                m_screen_cfg[i].size.y = rect.height;
                m_screen_cfg[i].pos.x  = rect.x;
                m_screen_cfg[i].pos.y  = rect.y;

                if ((rect.x == 0) && (rect.y == 0))
                {
                    m_primary_width  = rect.width;
                    m_primary_height = rect.height;
                }

                mTERA_LOG_DBG("disp%u: %d x %d @ (%d, %d) [GetGeometry() GetX() GetY]",
                              i,
                              rect.width,
                              rect.height,
                              rect.x,
                              rect.y);
            }
            m_screen_count = wxDisplay::GetCount();
        }
        // more than one X screen is active (this mode isn't supported yet)
        else
        {
            //
            // TBD - Notes:
            // 1) When two X screens are active the wxDisplay GetCount() function returns
            //    the value 1. This means I cannot use the wxDisplay GetGeometry() API to
            //    determine the position of the screens.
            // 2) The following code currently supports reading the resolution of each
            //    screen, but it does not support reading the position of each screen
            //    in relation to the other screens. This information is needed to support
            //    sending the display topology message to the firmware. The firmware needs
            //    to know this information before it can enable the local cursor feature.
            // 3) For the time being the code will generate an error message when it
            //    detects the system is configured to operate in this mode.
            // 4) If customers need to have this mode supported the GUI monitor tab could
            //    potentially be modified to allow the user to input the position of each
            //    screen.
            //

            mTERA_LOG_ERR("%d X screens detected. The pcoip_agent executable does not support this mode of operation!",
                          num_x_screens);

            for (int i=0; i<num_x_screens; i++)
            {
                Window root;
                int x, y;
                unsigned int width, height;
                unsigned int border_width;
                unsigned int depth;
                char hostname[100];
                char screen_num[10];

                // Read hostname from the DISPLAY environment variable
                sprintf(hostname, "%s", getenv("DISPLAY"));

                // Remove all characters from the end of the string that come
                // after the '.' character.
                for (int j=strlen(hostname); hostname[j] != '.'; j--)
                {
                    hostname[j] = '\0';
                }

                // Append the screen number to the hostname string
                sprintf(screen_num, "%d", i);
                strcat(hostname, screen_num);

                // open the X Windows display and read the Geometry
                Display *dsp = XOpenDisplay(hostname);
                if (XGetGeometry(dsp, RootWindow(dsp, i), &root,
                      &x, &y, &width, &height, &border_width, &depth) == False)
                {
                    mTERA_LOG_DBG("Error getting root window geometry");
                }
                else
                {
                    if (i == 0)
                    {
                        m_primary_width  = width;
                        m_primary_height = height;
                    }
                    m_screen_cfg[i].size.x = width;
                    m_screen_cfg[i].size.y = height;
                    m_screen_cfg[i].pos.x  = x;
                    m_screen_cfg[i].pos.y  = y;
                    m_screen_count = num_x_screens;

                    mTERA_LOG_DBG("hostname %s: x %d, y %d, width %d, height %d, border width %d, depth %d",
                                  hostname, x, y, width, height, border_width, depth);
                }
                XCloseDisplay(dsp);
            }
        }
    }

    mTERA_LOG_DBG("%d screens detected, primary %d x %d",
                  m_screen_count,
                  m_primary_width,
                  m_primary_height);

    sSCREEN_INFO screens[MAX_MON_COUNT];
    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        mTERA_LOG_DBG("disp%u: %d x %d @ (%d, %d)",
                      i,
                      m_screen_cfg[i].size.x,
                      m_screen_cfg[i].size.y,
                      m_screen_cfg[i].pos.x,
                      m_screen_cfg[i].pos.y);
        screens[i].width  = m_screen_cfg[i].size.x;
        screens[i].height = m_screen_cfg[i].size.y;
        screens[i].pos_x  = m_screen_cfg[i].pos.x;
        screens[i].pos_y  = m_screen_cfg[i].pos.y;
    }

    // send the screen information to the monitor page
    m_prop_sheet.m_mon_page->screen_list_init(m_screen_count, screens);
} // cPCOIP_AGENT_GUI::screen_topology_init()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::host_mon_info_invalidate
//      Invalidate the host monitor information in the PCoIP link object
//
//  Description:
//      This function invalidates host monitor information in the PCoIP link
//      object by sending down a 0-monitor configuration.
//
void cPCOIP_AGENT_GUI::host_mon_info_invalidate(void)
{
    mTERA_LOG_DBG("start");

    sTERA_PCOIP_LINK_DISP_TOPO disp_topo;
    mTERA_MEM_ZERO(&disp_topo, sizeof(disp_topo));

    // Send the 0-monitor information to the PCoIP link
    m_pcoip_link.disp_topology_send(&disp_topo,
                                    m_primary_width,
                                    m_primary_height,
                                    m_twinview_active);

    mTERA_LOG_DBG("end");

} // cPCOIP_AGENT_GUI::host_mon_info_invalidate()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::end_session
//      Do what is needed when a session ends
//
//  Description:
//      This function invalidates host monitor information, which disables the
//      local cursor feature.
//
void cPCOIP_AGENT_GUI::end_session(void)
{
    mTERA_LOG_DBG("start");

    // Invalidate the host monitor information
    host_mon_info_invalidate();

    // Blindly disable local cursor and keyboard
    m_pcoip_link.lcur_lkbd_disable();
    set_lcur_lkbd_mode(0);

    // Stop the auto-updating of the link object
    m_pcoip_link.stop();

    mTERA_LOG_DBG("end");

} // cPCOIP_AGENT_GUI::end_session()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::get_host_cursor_pos_abs
//      Retrieve current position of the host cursor in absolute coordinates
//
//  Description:
//      This function reads the position of the host cursor and returns the
//      absolute coordinates of the cursor. The range of the cursor coordinate
//      data returned by this function is 0-0xFFFF.
//
INT32 cPCOIP_AGENT_GUI::get_host_cursor_pos_abs(
    UINT16  *x_pos,
    UINT16  *y_pos)
{
    int          num_screens;
    int          i;
    Window      *root_windows;
    Window       ret_win;
    Bool         ret_query = False;
    int          root_x;
    int          root_y;
    int          win_x;
    int          win_y;
    unsigned int ret_mask;
    INT32        ret = TERA_ERR_FAILURE;

    // initialize the return data in case the correct data cannot be determined.
    *x_pos = 0;
    *y_pos = 0;

    if (m_xdisplay)
    {
        num_screens = XScreenCount(m_xdisplay);

        // TBD: currently the code only supports a single X screen
        if (num_screens == 1)
        {
            int disp_width = DisplayWidth(m_xdisplay, 0);
            int disp_height = DisplayHeight(m_xdisplay, 0);
            root_windows = (Window *)malloc(sizeof(Window) * num_screens);
            if (root_windows)
            {
                for(i=0; i<num_screens; i++)
                {
                    root_windows[i] = XRootWindow(m_xdisplay, i);
                }
                for (i=0; i<num_screens; i++)
                {
                    // TBD: the call to XQueryPointer is only needed for
                    // relative positioning. If I decide to move the cursor
                    // using XWarpPointer the code can be optimized to only
                    // call XQueryPointer while using relative positioning.
                    ret_query = XQueryPointer(m_xdisplay,
                                              root_windows[i],
                                              &ret_win,
                                              &ret_win,
                                              &root_x,
                                              &root_y,
                                              &win_x,
                                              &win_y,
                                              &ret_mask);
                    if (ret_query == True)
                    {
                        // convert host cursor coordinates from screen size
                        // to the range of the coordinates used by the client
                        *x_pos = (UINT16)((root_x * 0xFFFF) / disp_width);
                        *y_pos = (UINT16)((root_y * 0xFFFF) / disp_height);
                        ret = TERA_SUCCESS;

#if UPDATE_CLIENT_CUR_POS_DBG
                        mTERA_LOG_DBG("    # SCREENS = %d", num_screens);
                        mTERA_LOG_DBG("    SIZE      = %d X %d", disp_width, disp_height);
                        mTERA_LOG_DBG("    WINDOW    = %d", i);
                        mTERA_LOG_DBG("    ROOT(%d, %d)", root_x, root_y);
                        mTERA_LOG_DBG("    WIN (%d, %d)", win_x,  win_y);
                        mTERA_LOG_DBG("end (%d, %d) (0x%04X, 0x%04X)", *x_pos, *y_pos, *x_pos, *y_pos);
#endif
                        break;
                    }
               }
               if (ret_query == False)
               {
                   mTERA_LOG_ERR("XQueryPointer failed to determine location of host cursor");
               }
               free(root_windows);
            }
            else
            {
                mTERA_LOG_CRIT("malloc(%d) returned NULL", sizeof(Window) * num_screens);
            }
        }
        else
        {
            // TBD: Changes may be required in this function to handle multiple
            // X screens. Hopefully this is not be a serious limitation.
            mTERA_LOG_ERR("Host software does not support multiple X screens, unable to determine location of host cursor");
        }
    }
    else
    {
        mTERA_LOG_ERR("m_xdisplay is NULL, unable to determine location of host cursor");
    }

    return ret;
} // cPCOIP_AGENT_GUI::get_host_cursor_pos_abs()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::get_host_cursor_pos_pix
//      Retrieve current position of the host cursor in pixel coordinates
//
//  Description:
//      This function reads the position of the host cursor and returns the
//      pixel coordinates of the cursor. The range of the cursor coordinate
//      data returned by this function is based on the size of the desktop.
//
INT32 cPCOIP_AGENT_GUI::get_host_cursor_pos_pix(
    UINT16  *x_pos,
    UINT16  *y_pos)
{
    int          num_screens;
    int          i;
    Window      *root_windows;
    Window       ret_win;
    Bool         ret_query = False;
    int          root_x;
    int          root_y;
    int          win_x;
    int          win_y;
    unsigned int ret_mask;
    INT32        ret = TERA_ERR_FAILURE;

    // initialize the return data in case the correct data cannot be determined.
    *x_pos = 0;
    *y_pos = 0;

    if (m_xdisplay)
    {
        num_screens = XScreenCount(m_xdisplay);

        // TBD: currently the code only supports a single X screen
        if (num_screens == 1)
        {
            int disp_width = DisplayWidth(m_xdisplay, 0);
            int disp_height = DisplayHeight(m_xdisplay, 0);
            root_windows = (Window *)malloc(sizeof(Window) * num_screens);
            if (root_windows)
            {
                for(i=0; i<num_screens; i++)
                {
                    root_windows[i] = XRootWindow(m_xdisplay, i);
                }
                for (i=0; i<num_screens; i++)
                {
                    // TBD: the call to XQueryPointer is only needed for
                    // relative positioning. If I decide to move the cursor
                    // using XWarpPointer the code can be optimized to only
                    // call XQureyPointer while using relative positioning.
                    ret_query = XQueryPointer(m_xdisplay,
                                              root_windows[i],
                                              &ret_win,
                                              &ret_win,
                                              &root_x,
                                              &root_y,
                                              &win_x,
                                              &win_y,
                                              &ret_mask);
                    if (ret_query == True)
                    {
                        *x_pos = (UINT16)root_x;
                        *y_pos = (UINT16)root_y;
                        ret = TERA_SUCCESS;

#if 0
                        mTERA_LOG_DBG("    # SCREENS = %d", num_screens);
                        mTERA_LOG_DBG("    SIZE      = %d X %d", disp_width, disp_height);
                        mTERA_LOG_DBG("    WINDOW    = %d", i);
                        mTERA_LOG_DBG("    ROOT(%d, %d)", root_x, root_y);
                        mTERA_LOG_DBG("    WIN (%d, %d)", win_x,  win_y);
                        mTERA_LOG_DBG("end (%d, %d) (0x%04X, 0x%04X)", *x_pos, *y_pos, *x_pos, *y_pos);
#endif
                        break;
                    }
               }
               if (ret_query == False)
               {
                   mTERA_LOG_ERR("XQueryPointer failed to determine location of host cursor");
               }
               free(root_windows);
            }
            else
            {
                mTERA_LOG_CRIT("malloc(%d) returned NULL", sizeof(Window) * num_screens);
            }
        }
        else
        {
            // TBD: Changes may be required in this function to handle multiple
            // X screens. Hopefully this is not be a serious limitation.
            mTERA_LOG_ERR("Host software does not support multiple X screens, unable to determine location of host cursor");
        }
    }
    else
    {
        mTERA_LOG_ERR("m_xdisplay is NULL, unable to determine location of host cursor");
    }

    return ret;
} // cPCOIP_AGENT_GUI::get_host_cursor_pos_pix()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::lcur_lkbd_settings_send
//      Retrieve the local cursor settings from the config file
//
//  Description:
//      This function retrieves the session settings to determine the MAC
//      address of the client connected to the Host. It then reads the
//      client specific local cursor and keyboard settings from the config
//      file.
//
void cPCOIP_AGENT_GUI::lcur_lkbd_settings_send(
    int mon_info_ok
    )
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    UINT32                                 lcur_vis;
    INT32                                  ret;
	INT32                                  lcur_spd;
    UINT16                                 x_pos;
    UINT16                                 y_pos;
    int                                    lcur_lkbd_en_cfg;
    int                                    lcur_fade_cfg;

    mTERA_LOG_DBG("start: mon_info_ok %d, feat stat %d",
                  mon_info_ok,
                  m_last_feat_stat);

    // Get the session link status
    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    ret = m_pcoip_link.sess_link_get(&sess_link);

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("sess_link_get() failed (ret=%d)", ret);
    }
    else if ((sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) &&
             (mon_info_ok) &&
             (m_last_feat_stat == TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS) &&
             (m_prop_sheet.m_mon_page->is_screen_cfg_ok()))
    {
        // process the visible cursor option
        lcur_vis = m_prop_sheet.m_feat_page->cur_visibility_get_reg();
        mTERA_LOG_DBG("lcur_vis=%d", lcur_vis);

        // If only the host cursor is visible
        if (lcur_vis == PCOIP_AGENT_CUR_VISIBLE_HOST)
        {
            // set the visibility of the local cursor to invisible
            if (m_pcoip_link.lcur_pointer_visibility_set(0,0) != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to make local cursor invisible");
            }

            // always restore the system cursor
            show_host_cursor();
        }
        else
        {
            // send the current cursor bitmap to the client
            force_local_cursor_update();

            // if only the client cursor is visible
            if (lcur_vis == PCOIP_AGENT_CUR_VISIBLE_CLIENT)
            {
                hide_host_cursor();
            }
            else
            {
                show_host_cursor();
            }
        }

        // update the cursor fadeout
        wxGetApp().gui().host_cfg().get_lcur_fade_setting(sess_link.mac,
                                                          &lcur_fade_cfg);
        m_pcoip_link.lcur_idle_hide_set(lcur_fade_cfg);

        // If the GUI is open and the cursor speed feature is enabled read
        // the speed setting and send it to the client
        if ((m_prop_sheet.IsShown()) &&
            (m_prop_sheet.m_feat_page->lcur_spd_enable_get()))
        {
            lcur_spd = m_prop_sheet.m_feat_page->lcur_spd_get();
            ret = m_pcoip_link.lcur_spd_set(lcur_spd);
            if (ret == TERA_ERR_NOT_SUPPORTED)
            {
                //
                // the soft client does not support configuring the local
                // cursor speed
                //
                mTERA_LOG_DBG("unable to set local cursor speed (not supported by soft client)");
            }
            else if (ret != TERA_SUCCESS)
            {
                // error occurred while setting the local cursor speed
                mTERA_LOG_ERR("failure setting local cursor speed=%d", lcur_spd);
            }
        }

        //
        // If the client supports the keyboard repeat feature send the
        // settings to the client
        //
        if (m_pcoip_link.lkbd_client_auto_repeat_avail() == TERA_SUCCESS)
        {
            ret = m_pcoip_link.lkbd_repeat_rates_set(m_prop_sheet.m_feat_page->lkbd_delay_get(),
                                                     m_prop_sheet.m_feat_page->lkbd_rate_get());
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failure setting client kbd repeat settings");
            }
        }

        // Process the local cursor and keyboard settings
        wxGetApp().gui().host_cfg().get_lcur_lkbd_en_setting(sess_link.mac,
                                                             &lcur_lkbd_en_cfg);
        if (lcur_lkbd_en_cfg)
        {
            mTERA_LOG_DBG("Enabling local cursor after session connected");

            // Enable local cursor and keyboard
            if (TERA_SUCCESS == m_pcoip_link.lcur_lkbd_enable())
            {
                set_lcur_lkbd_mode(1);

                // configure the driver to submit absolute cursor reports
                m_pcoip_link.kmp_driver_cfg_set(TERA_PCOIP_LINK_LCUR_ABS_ALG);

                // if the local cursor is visible force shape update
                if (lcur_vis != PCOIP_AGENT_CUR_VISIBLE_HOST)
                {
                    force_local_cursor_update();
                }

                //
                // Update the local cursor position to match the position of
                // the host cursor.
                //
                if (TERA_SUCCESS == get_host_cursor_pos_abs(&x_pos, &y_pos))
                {
                    m_pcoip_link.lcur_pointer_location_set(0, x_pos, y_pos);
                }
            }
            else
            {
                set_lcur_lkbd_mode(0);
            }
        }
        else
        {
            mTERA_LOG_DBG("Disabling local cursor after session connected");

            // Disable local cursor and keyboard
            m_pcoip_link.lcur_lkbd_disable();
            set_lcur_lkbd_mode(0);
        }
    }
    else
    {
        // Disable local cursor and keyboard
        m_pcoip_link.lcur_lkbd_disable();
        set_lcur_lkbd_mode(0);

        // Reinitialize the local cursor/keyboard controls if there is a
        // problem with the topology or the screen settings are not configured
        if ((!mon_info_ok) ||
            (m_last_feat_stat != TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS) ||
            (!m_prop_sheet.m_mon_page->is_screen_cfg_ok()))
        {
            m_prop_sheet.m_feat_page->init_dialog_lcur_lkbd();
        }
    }
    mTERA_LOG_DBG("end - mon_info_ok=%d", mon_info_ok);
} // cPCOIP_AGENT_GUI::lcur_lkbd_settings_send()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::lcur_lkbd_enable_get
///     Return an indication of whether local cursor and kbd is enabled
//
//  Description:
///     This function returns a non-zero value if the local cursor and
///     keyboard feature is enabled. This feature must always be enabled
///     when using a soft client.
///
INT cPCOIP_AGENT_GUI::lcur_lkbd_enable_get(void)
{
    UINT32                                 stat;
    UINT32                                 cnt;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lcur_lkbd_en_cfg;

    // if the local cursor cannot be disable
    if(m_pcoip_link.lcur_disable_avail() == TERA_ERR_NOT_SUPPORTED)
    {
        return 1;
    }
    else
    {
        // if the local cursor feature status is ok return the value in the
        // registry
        if((m_pcoip_link.lcur_lkbd_last_feat_stat_get(&stat, &cnt) == TERA_SUCCESS) &&
           (stat == TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS))
        {
            mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
            wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);
            wxGetApp().gui().host_cfg().get_lcur_lkbd_en_setting(sess_link.mac,
                                                                 &lcur_lkbd_en_cfg);
            return (lcur_lkbd_en_cfg);
        }
        else
        {
            return 0;
        }
    }
} // cPCOIP_AGENT_GUI::lcur_lkbd_enable_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::lcur_fade_get
///     Return indication of whether the local cursor will be hidden when idle
//
//  Description:
///     This function returns a non-zero value if the local cursor will be
///     hidden when idle. This feature must always be disabled when the feature
///     is not supported by the client.
///
INT cPCOIP_AGENT_GUI::lcur_fade_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lcur_fade_cfg;

    // if the client does not support the feature
    if(m_pcoip_link.lcur_idle_hide_avail() == TERA_ERR_NOT_SUPPORTED)
    {
        return 0;
    }
    else
    {
        mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
        wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);
        wxGetApp().gui().host_cfg().get_lcur_fade_setting(sess_link.mac,
                                                          &lcur_fade_cfg);
        return (lcur_fade_cfg);
    }
} // cPCOIP_AGENT_GUI::lcur_fade_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::cPCOIP_AGENT_GUI
///     Constructor
//
//  Description:
///     This is the PCoIP Host Software GUI class constructor. The class
///     member data variables are initialized here.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI::cPCOIP_AGENT_GUI(
    wxWindow       *parent,
    const wxString& title
    ) :
    wxFrame(parent, wxID_ANY, title),
    m_prop_sheet(parent), // need to use "parent" instead of "this" to get taskbar icon
    m_heartbeat_timer(this)
{
    INT32       ret;

    mTERA_LOG_DBG("start");
    m_parent = parent;

    // Initialize local variables
    m_sess_connected = 0;
    m_agent_state = TERA_AGENT_GUI_DISCONNECTED_STATE;
    m_read_client_topo_retry = 0;
    m_client_send_topo_retry = 0;
    m_twinview_active = 0;
    m_screen_count = 0;
    mTERA_MEM_ZERO(&m_screen_cfg[0], sizeof(sSCREEN_CFG)*TERA_PCOIP_HCC_APDU_MAX_NUM_DISP);
    m_last_feat_stat = 0;
    m_last_feat_stat_cnt = 0;
    m_client_cache_status = TERA_ERR_INVALID_ARG;
    m_lcur_lkbd_en = 0;
    m_host_cur_hidden = 0;
    mTERA_MEM_ZERO(&m_disp_topo, sizeof(m_disp_topo));
    set_topo_sent_status(0);
    m_last_cursor_serial = 0;
    m_icon_create_delay = TERA_SYS_TRAY_ICON_DELAY;

    // Start the PCoIP link object
    m_pcoip_link.start();

    // read the host card device type
    ret = m_pcoip_link.local_info_get(&m_dev_type);
    if ((ret != TERA_SUCCESS) || (m_dev_type >= MAX_TERA_HOST_DEVICE_TYPE))
    {
        mTERA_LOG_ERR("Problem reading device type, defaulting to TERA1 (%d %d)",
                      ret, m_dev_type);
        m_dev_type = TERA1x00_HOST_CARD;
    }

    // Tell the stats page how many monitors are supported by the host card
    m_prop_sheet.m_stat_page->max_monitors_set(MAX_MONITORS(m_dev_type));

    // clear the contents of the topology structure

    // read the current screen topology settings
    screen_topology_init();

    // Invalidate the host monitor information first
    host_mon_info_invalidate();

    // initialize the screen locker
    if (!init_screen_lock_utility())
    {
        mTERA_LOG_CRIT("init_screen_lock_utility failed!!!");
    }

    // query Xfixes extension
    int major = 0, minor = 0;
    int xfixes_event_base, xfixes_error_base;
    if (XFixesQueryExtension(m_xdisplay, &xfixes_event_base, &xfixes_error_base))
    {
        XFixesQueryVersion(m_xdisplay, &major, &minor);
    }

    if (major >= 4)
    {
        // Initialize flag that determines whether the host cursor can be
        // hidden (needs Xfixes 4)
        m_host_cur_hide_supported = 1;

        // install event handler for cursor changes
        GdkDisplay *display = gdk_display_get_default();
        gdk_x11_register_standard_event_type(display,
                                             xfixes_event_base + XFixesCursorNotify,
                                             1);
        gdk_window_add_filter(NULL, xfixes_events_cback, this);

        // request events for cursor changes
        Window root = XDefaultRootWindow(m_xdisplay);
        XFixesSelectCursorInput(m_xdisplay, root, XFixesDisplayCursorNotifyMask);
    }
    else
    {
        m_host_cur_hide_supported = 0;
    }

    // enable processing of fatal exceptions
    wxHandleFatalExceptions(true);

    mTERA_LOG_DBG("device type=%s, cursor hide supported=%d",
                  device_type_str_tbl[m_dev_type],
                  m_host_cur_hide_supported);
} // cPCOIP_AGENT_GUI::cPCOIP_AGENT_GUI()

cPCOIP_AGENT_GUI::~cPCOIP_AGENT_GUI(void)
{
    if (m_heartbeat_timer.IsRunning())
    {
        m_heartbeat_timer.Stop();
    }

    if (m_tray.IsIconInstalled())
    {
        m_tray.RemoveIcon();
    }

    gdk_window_remove_filter(NULL, xfixes_events_cback, NULL);

} // cPCOIP_AGENT_GUI::~cPCOIP_AGENT_GUI()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::prop_sheet_show
///     Show the property sheet
//
//  Description:
///     This function shows the property sheet. If the property sheet is
///     already opened, then it is brought to the front.
///
/// \return
///     None
///
void cPCOIP_AGENT_GUI::prop_sheet_show(void)
{
    mTERA_LOG_DBG("start");

    // Show the property sheet (does nothing if its already shown)
    m_prop_sheet.Show(true);

    // Bring the property sheet to the top so the users can see it
    m_prop_sheet.Raise();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI::prop_sheet_show()

// TBD

int cPCOIP_AGENT_GUI::prop_sheet_is_shown(void)
{
    return m_prop_sheet.IsShown();
} // cPCOIP_AGENT_GUI::prop_sheet_is_shown()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::pcoip_link
///     Return the reference to the PCoIP link object
//
//  Description:
///     This function returns the reference to the PCoIP link object, which
///     can be used to communicate with the PCoIP device.
///
/// \return
///     cTERA_PCOIP_LINK object
///
cTERA_PCOIP_LINK& cPCOIP_AGENT_GUI::pcoip_link(void)
{
    return m_pcoip_link;
} // cPCOIP_AGENT_GUI::pcoip_link()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::host_netif
///     Return the reference to the host network interface object
//
//  Description:
///     This function returns the reference to the host network interface
///     object, which contains the details of each of the active host network
///     interfaces.
///
/// \return
///     cTERA_NETIF object
///
cTERA_NETIF& cPCOIP_AGENT_GUI::host_netif(void)
{
    return m_host_netif;
} // cPCOIP_AGENT_GUI::host_netif()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::host_cfg
///     Return the reference to the host configuration object
//
//  Description:
///     This function returns the reference to the host configuration
///     object, which can be used to read/write/delete configuration
///     settings.
///
/// \return
///     cTERA_CFG object
///
cTERA_CFG& cPCOIP_AGENT_GUI::host_cfg(void)
{
    return m_host_cfg;
} // cPCOIP_AGENT_GUI::host_cfg()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::client_cache_topology_get
//      Read the client cached display topology
//
//  Description:
//      This function gets the current client cached display topology and
//      stores it in the variables m_client_disp_topo and
//      m_client_disp_edid. If the client does not have valid cached
//      topology the structures are filled with zeros.
//
//      This function should be called when a session starts.
//
//      TBD - is the following comment accurate?
//      This function also updates the client topology settings on the
//      monitor page.
//
INT32 cPCOIP_AGENT_GUI::client_cache_topology_get(void)
{
    int                                     retry = 0;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK  sess_link;
    INT32                                   ret;

    // verify a session is active
    ret = m_pcoip_link.sess_link_get(&sess_link);
    if ((ret != TERA_SUCCESS) ||
        !(sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
    {
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        // get the client cached topology settings if they are available
        ret = m_pcoip_link.cached_topo_get_avail();
        if (ret == TERA_SUCCESS)
        {
            //
            // Read the cached topology settings from the client. The zero client
            // cached settings always contain the monitor EDID information
            // regardless of whether the client has configured the display topology
            // settings. The host SW uses this information to determine which
            // monitor is connected to each DVI port.
            //
            do
            {
                ret = wxGetApp().gui().pcoip_link().cache_disp_topology_get(&m_disp_topo);
                retry++;
            } while ((ret != TERA_SUCCESS) && (retry < 3));
        }
    }

    if ((ret != TERA_SUCCESS) && (ret != TERA_ERR_IMG_NOT_ACTIVE))
    {
        // clear the contents of the topology structure if any errors occur
        mTERA_MEM_ZERO(&m_disp_topo, sizeof(m_disp_topo));
    }

    return ret;
} // cPCOIP_AGENT_GUI::client_cache_topology_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::display_topology_send
///     Send the host display topology to the firmware
//
//  Description:
///     This function gets the current host display topology and sends it down
///     to firmware. The topology information is read from the user
///     configuration file.
///
/// \return
///     0 - monitor information not sent or an error occurrred sending it
///     1 - monitor information successfully sent
///
int cPCOIP_AGENT_GUI::display_topology_send(void)
{
    unsigned int                           mon_count;
    unsigned int                           screen_count;
    wxString                               name;
    wxString                               vid;
    wxString                               pid;
    wxString                               serial;
    INT32                                  size_x, size_y;
    INT32                                  pos_x, pos_y;
    UINT32                                 monitor;
    UINT32                                 topology_mon_index;
    wxString                               label;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    INT32                                  ret;

    // Get the session link status
    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    ret = m_pcoip_link.sess_link_get(&sess_link);

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("sess_link_get() failed (ret=%d)", ret);
    }
    else if (!(sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
    {
        mTERA_LOG_DBG("no session active - abort sending monitor info");
        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        // Get the number of valid monitors and screens in the user configuration
        wxGetApp().gui().host_cfg().get_monitor_screen_counts(sess_link.mac,
                                                              &mon_count,
                                                              &screen_count);

        sTERA_PCOIP_LINK_DISP_TOPO disp_topo;
        mTERA_MEM_ZERO(&disp_topo, sizeof(disp_topo));

        if ((mon_count == 0) || (mon_count > MAX_MONITORS(m_dev_type)) ||
            (screen_count == 0) || (screen_count > MAX_MONITORS(m_dev_type)) ||
            (m_screen_count != screen_count))
        {
            mTERA_LOG_DBG("invalid mon data in config: mon_count %d screen_count %d m_screen_count %d",
                          mon_count, screen_count, m_screen_count);
        }
        else
        {
            // try using the monitor data stored in the config file
            disp_topo.num_disp = screen_count;
            for (UINT32 i = 0; i < screen_count; i++)
            {
                wxGetApp().gui().host_cfg().get_monitor_info(sess_link.mac,
                                                             i+1,
                                                             &vid,
                                                             &pid,
                                                             &serial,
                                                             &size_x,
                                                             &size_y,
                                                             &pos_x,
                                                             &pos_y,
                                                             &monitor);

                // Break out on any incomplete entries
                if ((vid == PCOIP_AGENT_CFG_MONX_VID_DEFAULT) ||
                    (pid == PCOIP_AGENT_CFG_MONX_PID_DEFAULT) ||
                    (serial == PCOIP_AGENT_CFG_MONX_SERIAL_DEFAULT))
                {
                    mTERA_LOG_ERR("Invalid topology string data!");
                    disp_topo.num_disp = 0;
                    break;
                }
                if ((size_x == PCOIP_AGENT_CFG_MONX_POS_X_DEFAULT) ||
                    (size_y == PCOIP_AGENT_CFG_MONX_SIZE_Y_DEFAULT) ||
                    (monitor == PCOIP_AGENT_CFG_MONX_MON_DEFAULT))
                {
                    mTERA_LOG_ERR("Invalid topology data! (size %d X %d, mon %d)",
                                  size_x, size_y, monitor);
                    disp_topo.num_disp = 0;
                    break;
                }

                // get the index of this monitor's entry in the topology msg
                if (TERA_SUCCESS != m_prop_sheet.m_mon_page->get_idx_monitor(monitor, &topology_mon_index))
                {
                    mTERA_LOG_ERR("Error finding monitor %d in the list of monitors", monitor);
                    disp_topo.num_disp = 0;
                    break;
                }

                // Verify the setting read from the config file matches one of
                // the active screens. If a match is not found the topology
                // has changed, which means the user must reconfigure the
                // monitor settings.
                INT match_found = 0;
                for (UINT32 j = 0; j < m_screen_count; j++)
                {
                    if ((size_x == m_screen_cfg[j].size.x) &&
                        (size_y == m_screen_cfg[j].size.y) &&
                        (pos_x  == m_screen_cfg[j].pos.x) &&
                        (pos_y  == m_screen_cfg[j].pos.y))
                    {
                        match_found = 1;
                        break;
                    }
                }
                if (!match_found)
                {
                    mTERA_LOG_DBG("Stored topology data does not match active topology - erasing stored topology");
                    disp_topo.num_disp = 0;
                    m_prop_sheet.m_mon_page->delete_mon_usr_cfg();
                    set_topo_sent_status(0);
                    break;
                }

                mTERA_LOG_DBG("monitor %u", i);
                mTERA_LOG_DBG("\tVendorID:     %S", vid.c_str());
                mTERA_LOG_DBG("\tProductID:    %S", pid.c_str());
                mTERA_LOG_DBG("\tSerialNumber: %S", serial.c_str());
                mTERA_LOG_DBG("\tSize:         %ux%u", size_x, size_y);
                mTERA_LOG_DBG("\tPosition:     (%u, %u)", pos_x, pos_y);
                mTERA_LOG_DBG("\tMonitor       %u", topology_mon_index);

                // Add this into the list of monitors
                disp_topo.disp[topology_mon_index].width    = size_x;
                disp_topo.disp[topology_mon_index].height   = size_y;
                disp_topo.disp[topology_mon_index].x_origin = pos_x;
                disp_topo.disp[topology_mon_index].y_origin = pos_y;

                // TBD - update the code to support rotated displays
                disp_topo.disp[topology_mon_index].rotation = TERA_PCOIP_LINK_DISP_ROTATION_0;
            } // end for
        } // end if

        // Handle the situation where only one monitor is active and no
        // topology data is stored in the user config area. When this
        // happens use the cached client topology
        if ((disp_topo.num_disp == 0) &&
            (m_screen_count == 1) &&
            (m_disp_topo.num_disp == 1))
        {
            // use the active topology settings read from the OS
            mTERA_MEM_ZERO(&disp_topo, sizeof(disp_topo));
            for (UINT32 i = 0; i < TERA_PCOIP_LINK_MAX_NUM_DISP; i++)
            {

                // TBD - update the code to support rotated displays
                disp_topo.disp[i].rotation = TERA_PCOIP_LINK_DISP_ROTATION_0;

                if ((m_disp_topo.disp[i].width) && (m_disp_topo.disp[i].height))
                {
                    wxDisplay mon(0);
                    wxRect    rect = mon.GetGeometry();

                    disp_topo.disp[i].width    = rect.width;
                    disp_topo.disp[i].height   = rect.height;
                    disp_topo.disp[i].x_origin = rect.x;
                    disp_topo.disp[i].y_origin = rect.y;

                    mTERA_LOG_DBG("monitor %u", i);
                    mTERA_LOG_DBG("\tSize:         %ux%u",
                                  disp_topo.disp[i].width,
                                  disp_topo.disp[i].height);
                    mTERA_LOG_DBG("\tPosition:     (%u, %u)",
                                  disp_topo.disp[i].x_origin,
                                  disp_topo.disp[i].y_origin);
                    disp_topo.num_disp = 1;
                }
            } // end for
        }

        if (disp_topo.num_disp != 0)
        {
            // Send the host monitor information to the PCoIP link
            ret = m_pcoip_link.disp_topology_send(&disp_topo,
                                                  m_primary_width,
                                                  m_primary_height,
                                                  m_twinview_active);
        }
        else
        {
            ret = TERA_ERR_INVALID_STATE;
        }
    }
    mTERA_LOG_DBG("end (ret=%d)", ret);
    if (ret == TERA_SUCCESS)
    {
        set_topo_sent_status(1);
        return 1;
    }
    else
    {
        set_topo_sent_status(0);
        return 0;
    }
} // cPCOIP_AGENT_GUI::display_topology_send()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::display_topology_ok
///     Return non-zero value if the topology was sent to the client and the
///     firmware reports the current topology status is okay
//
//  Description:
///     This function returns an indication of whether the topology was
///     sent successfully to the client and the firmware reports the current
///     topology status is okay/good. If both conditions are not true the
///     host SW should not attempt to enable the local cursor feature.
///
/// \return
///     0 - topology not sent or firmware reports bad topology
///     1 - topology successfully sent and firmware reports topology okay
///
int cPCOIP_AGENT_GUI::display_topology_ok(void)
{
    if (m_topo_sent_ok &&
        (m_last_feat_stat == TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cPCOIP_AGENT_GUI::display_topology_ok()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::connect_proc
//      Perform necesary processing that occurs when a session is connected
//
//  Description:
//      This function does what is required when a session becomes connected
//      or a display change occurs. The function grabs the latest topology
//      and sends it to the firmware.
//
//		The function also sends the local cursor and keyboard settings. This
//      is necessary in case the local cursor was previously disabled because
//      it could not be supported (i.e. display was rotated) and now it can
//      be supported.
//
void cPCOIP_AGENT_GUI::connect_proc(void)
{
	UINT32 stat;
	UINT32 cnt;
	INT32  ret;

	mTERA_LOG_DBG("start");

	// update the state if a session is not active
	if (m_sess_connected == 0)
	{
        m_agent_state = TERA_AGENT_GUI_DISCONNECTED_STATE;
	}
	else
	{
		// retrieve the feature status and check if things are okay
		ret = m_pcoip_link.lcur_lkbd_last_feat_stat_get(&stat, &cnt);
		if ((ret == TERA_SUCCESS) &&
			(stat == TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS))
		{
			// clear the retry counter
			m_client_send_topo_retry = 0;

			// Send the local cursor and keyboard settings to the firmware
			lcur_lkbd_settings_send(m_topo_sent_ok);

            // change state since the feature status is OK
            m_agent_state = TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE;

            //
            // Change the last feature status count. This is done to
            // prevent sending the topology to the client an extra time
            // which typically happens when changing topologies. If the last
            // counter is not updated the host SW will consider the previous
            // change as a new change, and then trigger resending the topology
            // to the firmware.
            //
            m_last_feat_stat_cnt = cnt;

            //
            // update the copy of the last feature status, if this is not done
            // the host SW could be left in a state where the copy of the last
            // status is not correct. This caused problems on a ZC when turning
            // the monitors off (BZ 5979)
            //
            m_last_feat_stat = stat;
		}
		// else resend the topology if the retry count is not zero
		else if (m_client_send_topo_retry)
		{
			m_client_send_topo_retry--;

			// Send the current display topology
			display_topology_send();

			// Send the local cursor and keyboard settings to the firmware
			lcur_lkbd_settings_send(m_topo_sent_ok);
		}
	}

    mTERA_LOG_DBG("end");

} // cPCOIP_AGENT_GUI::connect_proc()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::update_client_cur_pos
//      Update client cursor position if the host moved the cursor
//
//  Description:
//      Client devices send cursor positioning data to the host in absolute
//      coordinates. This is different from standard pointing devices which
//      provide relative updates. The host and client cursor can get out of
//      sync if the host updates the cursor because the client cursor does
//      not match the host cursor location.
//
//      This function reads the current host cursor location. If the host
//      cursor position changed since the last time this function was called
//      the new point is sent to the driver. The driver compares the host
//      cursor location to the last X client cursor reports. If the host
//      coordinates do not match one of the last X client cursor reports the
//      host cursor location is sent to the client.
//
void cPCOIP_AGENT_GUI::update_client_cur_pos(void)
{
    UINT16 host_cur_abs_x;
    UINT16 host_cur_abs_y;

    // read current cursor position
    if (TERA_SUCCESS == get_host_cursor_pos_abs(&host_cur_abs_x, &host_cur_abs_y))
    {
        if ((host_cur_abs_x != last_host_cur_abs_x) ||
            (host_cur_abs_y != last_host_cur_abs_y))
        {
#if UPDATE_CLIENT_CUR_POS_DBG
            mTERA_LOG_DBG("Cursor coordinates: absolute (x%X, x%X)",
                          host_cur_abs_x,
                          host_cur_abs_y);
#endif

            m_pcoip_link.lcur_pointer_location_set_if_new(host_cur_abs_x, host_cur_abs_y);
            last_host_cur_abs_x = host_cur_abs_x;
            last_host_cur_abs_y = host_cur_abs_y;
        }
    }
} // cPCOIP_AGENT_GUI::update_client_cur_pos()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::init_screen_lock_utility
//      Initialize the utility used to lock the workstation
//
//  Description:
//      The host SW supports a feature that allows users to lock the
//      workstation when a PCoIP session disconnects. This is used to
//      ensure another user cannot access a users desktop by
//      disconnecting the users session and then connecting to the host
//      from a different client.
//
//      This function initializes a string that contains the command that will
//      be executed to lock the workstation.
//
//      The host software installation package installs a script named
//      pcoip_lockscreen in the /usr/bin directory, where the pcoip_agent
//      executable is installed. If a user moves the pcoip_agent executable
//      to another location, such as a network drive, this script can also be
//      moved. This function searches for the script in /usr/bin. If it does
//      not find it the function searches for the script in the directory
//      that contains the pcoip_agent executable.
//
//      When the host software needs to lock the screen it executes this
//      script. If the host software is installed on a system and there is a
//      problem where the host software does not lock the screen the
//      administrator should ensure this script works. It's possible the screen
//      saver utility invoked by this script does not exist on the end user's
//      Linux distribution. Using a script to lock the screen allows end users
//      to fix screen locking problems without recompiling the host software
//      executable.
//
//      This function returns true if the string is initialized properly,
//      otherwise it returns false.
//
bool cPCOIP_AGENT_GUI::init_screen_lock_utility(void)
{
    int         ret;
    int         str_len;
    char       *exe_name;
    static char script_path[512];

    // clear the utility string
    m_lock_utility.Clear();

    // check if the script is in the default location
    ret = access(DEFAULT_SCREEN_LOCK_SCRIPT, X_OK);
    if (ret == 0)
    {
        m_lock_utility.Printf(_T("%S"), DEFAULT_SCREEN_LOCK_SCRIPT);
    }
    else
    {
        mTERA_LOG_DBG("%s does not exist, checking directory containing pcoip_agent",
                      DEFAULT_SCREEN_LOCK_SCRIPT);

        //
        // Read the symbolic link /proc/self/exe, which contains the fully
        // qualified path to the pcoip_agent executable
        //
        str_len = readlink("/proc/self/exe", script_path, sizeof(script_path));
        if((str_len != -1) && (str_len < (int)sizeof(script_path)))
        {
            // terminate the string and get a pointer to the start of the executable
            script_path[str_len] = '\0';
            exe_name = strrchr(script_path, '/') + 1;
            if (exe_name && (exe_name != (script_path + 1)))
            {
                // strip the executable name and concatenate the script name
                *exe_name = '\0';
                strncat(script_path, SCREEN_LOCK_SCRIPTNAME, sizeof(script_path) - strlen(script_path));

                // check if the script is in the same directory as the
                // pcoip_agent executable
                ret = access(script_path, X_OK);
                if (ret == 0)
                {
                    m_lock_utility.Printf(_T("%S"), script_path);
                }
            }
            else
            {
                mTERA_LOG_ERR("error while determining pointer to executable name");
            }
        }
        else
        {
            mTERA_LOG_ERR("readlink() returned %d", str_len);
        }
        //
    }

    // if a screen log utility has been found
    if (!m_lock_utility.IsEmpty())
    {
        mTERA_LOG_DBG("using '%S' to lock the workstation", m_lock_utility.c_str());
        return true;
    }
    else
    {
        mTERA_LOG_ERR("Unable to find a screen lock utility!!!");
        return false;
    }
} // cPCOIP_AGENT_GUI::init_screen_lock_utility()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::get_client_topo
//      Returns a copy of the client topology
//
//  Description:
//      This function returns a copy of the client topology settings that were
//      read at the start of a session.
//
void cPCOIP_AGENT_GUI::get_client_topo(
    sTERA_PCOIP_LINK_CACHE_DISP_TOPO *disp_topo
    )
{
    // return a copy of the topology structure
    mTERA_MEM_COPY(disp_topo, &m_disp_topo, sizeof(m_disp_topo));
} // cPCOIP_AGENT_GUI::get_client_topo()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::force_local_cursor_update
//      Force the agent to send a new cursor shape to the client
//
//  Description:
//      This function forces the agent to resend the current cursor to the
//      client.
//
void cPCOIP_AGENT_GUI::force_local_cursor_update(void)
{
    // fake a change of the cursor shape
    m_last_cursor_serial = 0;

    // generate a fake cursor change event
    process_cursor_shape_change_event();

    mTERA_LOG_DBG("End: host cursor hidden %d, local cursor enabled %d",
                  m_host_cur_hidden,
                  m_lcur_lkbd_en);
} // cPCOIP_AGENT_GUI::force_local_cursor_update()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::xfixes_events_cback
//      Callback function invoked when XFIXES events occur
//
//  Description:
//      This is a callback function which is invoked when XFIXES events occur.
//      Currently XFIXES only supports one event, which occurs when the cursor
//      shape changes.
//
GdkFilterReturn cPCOIP_AGENT_GUI::xfixes_events_cback(GdkXEvent *xevent, GdkEvent *event, gpointer data)
{
    cPCOIP_AGENT_GUI* object = (cPCOIP_AGENT_GUI*) data;
    object->process_cursor_shape_change_event();
    return GDK_FILTER_CONTINUE;
} // cPCOIP_AGENT_GUI::xfixes_events_cback ()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::process_cursor_shape_change_event
//      This function processes cursor shape change events
//
//  Description:
//      This processes cursor shape change eevnts. If the local cursor feature
//      is enabled and the local cursor is visible this function calls the
//      process_cursor function which will check to see if the cursor bitmap
//      changed. If the cursor bitmap changed the new bitmap will be sent to
//      the client.
//
void cPCOIP_AGENT_GUI::process_cursor_shape_change_event (void)
{
    // wait for session to be finally established
    if ((m_agent_state != TERA_AGENT_GUI_CHECK_TOPO_CHANGE_STATE) &&
        (m_agent_state != TERA_AGENT_GUI_SEND_TOPO_TO_CLIENT_STATE))
    {
        return;
    }

    // do nothing if local cursor not enabled
    if (!m_lcur_lkbd_en)
    {
        return;
    }

    // do nothing if only the host cursor is visible
    if (PCOIP_AGENT_CUR_VISIBLE_HOST == m_prop_sheet.m_feat_page->cur_visibility_get_reg())
    {
        return;
    }

    // check for cursor change and perform update
    process_cursor();

    return;
} // cPCOIP_AGENT_GUI::process_cursor_shape_change_event ()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::process_cursor
//      Check if the cursor shape changed and send it to the client if it did
//
//  Description:
//      This function reads the cursor shape and checks if it's new. If it is
//      new the cursor bitmap is sent to the client if possible. If the cursor
//      bitmap cannot be sent to the client a cursor arrow bitmap is sent.
//
void cPCOIP_AGENT_GUI::process_cursor(void)
{
    INT32 size_valid;

    // get default root window
    Window root = XDefaultRootWindow(m_xdisplay);

    // get pointer to cursor object
    XFixesCursorImage *x_cursor = NULL;
    x_cursor = XFixesGetCursorImage (m_xdisplay);  // important: free bitmap memory to avoid massive memory leak

    if (x_cursor)
    {
        // If the cursor shape changed
        if (x_cursor->cursor_serial != m_last_cursor_serial)
        {
#ifdef HOST_CUR_BMP_DBG
            mTERA_LOG_DBG("current serial %d, last serial %d, host cursor hidden %d, local cursor enabled %d",
                          x_cursor->cursor_serial,
                          m_last_cursor_serial,
                          m_host_cur_hidden,
                          m_lcur_lkbd_en);
#endif

#if 0
            // TBD - remove?
            // Potential workaround for problem where the host cursor is not hidden
            // on multi-monitor systems with Xineram enabled and Twinview disabled.
            // When the cursor shape changes on a screen other than the primary
            // screen the host cursor becomes visible.
            //
            // When the following code is enabled the system gets into a state
            // where the cursor is stuck on one monitor. When the cursor is stuck
            // on the left monitor and the user moves the cursor beyond the right
            // edge of the monitor the ursor jumps to the left edge of the monitor.
            //
            // This change fixes the problem where the host cursor becomes
            // visible when the shape changes on a screen other than the primary
            // screen but it isn't a viable solution because the cursor cannot be
            // moved to the 2nd monitor. Until a solution is found for the problem
            // this code will remain.
            //
            UINT32 lcur_vis = m_prop_sheet.m_feat_page->cur_visibility_get_reg();
            if (lcur_vis == PCOIP_AGENT_CUR_VISIBLE_CLIENT)
            {
                Window root = XDefaultRootWindow(m_xdisplay);
                XLockDisplay(m_xdisplay);
                XFixesShowCursor(m_xdisplay, root);
                XFixesHideCursor(m_xdisplay, root);
                XFlush(m_xdisplay);
                XUnlockDisplay(m_xdisplay);
            }
#endif

            m_last_cursor_serial = x_cursor->cursor_serial;

            // allocate storage for a copy of the cursor bitmap
            long npixels = x_cursor->width * x_cursor->height;
            UINT32* cur_shape;
            cur_shape = new UINT32[npixels];

            // If the cursor cannot be displayed by the client send a pointer cursor to
            // the client.
            // - The current code does not properly handle animated cursors. The
            //   number of pixels in these cursors is 1. Maybe there is a bug in
            //   XFixes.
            // - The current code also does not properly handle hidden cursors.
            //   The number of pixels in these cursors is 1. The code cannot
            //   differentiate between hidden and animated cursors.
            // - Don't display cursors that are larger than the maximum size supported
            //   by the client.
            // - if storage for the bitmap could not be allocated
            size_valid = m_pcoip_link.lcur_shape_size_valid(x_cursor->width, x_cursor->height, 0);
            if ((npixels == 1) ||
                (size_valid != TERA_SUCCESS) ||
                (cur_shape == NULL))
            {
                mTERA_LOG_DBG("Unsupported cursor: npixels %d, size %d x %d, hot %d,%d, pos %d,%d, valid %d",
                              npixels,
                              x_cursor->width,
                              x_cursor->height,
                              x_cursor->xhot,
                              x_cursor->yhot,
                              x_cursor->x,
                              x_cursor->y,
                              size_valid);

                // send default cursor bitmap
                m_pcoip_link.lcur_shape_bmp_send(PCOIP_AGENT_DEFAULT_CUR_SHAPE_HOT_X,
                                                 PCOIP_AGENT_DEFAULT_CUR_SHAPE_HOT_Y,
                                                 PCOIP_AGENT_DEFAULT_CUR_SHAPE_WIDTH,
                                                 PCOIP_AGENT_DEFAULT_CUR_SHAPE_HEIGHT,
                                                 0,
                                                 0,
                                                 &pcoip_agent_default_cur_shape[0][0],
                                                 PCOIP_AGENT_DEFAULT_CUR_SHAPE_HEIGHT * PCOIP_AGENT_DEFAULT_CUR_SHAPE_WIDTH * sizeof(UINT32),
                                                 TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA);
            }
            else
            {
#ifdef HOST_CUR_BMP_DBG
                mTERA_LOG_DBG("cursor: npixels %d, size %d x %d, hot %d,%d, pos %d,%d, valid %d",
                              npixels,
                              x_cursor->width,
                              x_cursor->height,
                              x_cursor->xhot,
                              x_cursor->yhot,
                              x_cursor->x,
                              x_cursor->y,
                              size_valid);
#endif

                // copy cursor bitmap data
                unsigned long *pix = x_cursor->pixels;
                for ( long k = 0; k < npixels; k++)
                {
                    cur_shape[k] = (UINT32)pix[k];
                }

                // send host cursor bitmap
                m_pcoip_link.lcur_shape_bmp_send(x_cursor->xhot,
                                                 x_cursor->yhot,
                                                 x_cursor->width,
                                                 x_cursor->height,
                                                 0,
                                                 0,
                                                 &cur_shape[0],
                                                 npixels*sizeof(UINT32),
                                                 TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA);
            }
            // free memory allocated for shape bitmap
            if (cur_shape)
            {
                delete cur_shape;
            }
        }
        // free memory allocated for cursor image
        Xfree(x_cursor);
    }
    else
    {
        mTERA_LOG_ERR("XFixesGetCursorImage() returned NULL");
    }

} // cPCOIP_AGENT_GUI::process_cursor()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::hide_host_cursor
//      Show the host cursor if it is hidden
//
void cPCOIP_AGENT_GUI::hide_host_cursor(void)
{
    // switch on host cursor
    // Appears to fire X error if not already hidden -> but no way to query that
    if (m_host_cur_hide_supported && (!m_host_cur_hidden))
    {
        Window root = XDefaultRootWindow(m_xdisplay);
        mTERA_LOG_DBG("Hiding host cursor (root = %d)", root);
        XFixesHideCursor(m_xdisplay, root);
        XFlush(m_xdisplay);
        m_host_cur_hidden = 1;
    }
    else
    {
        mTERA_LOG_DBG("XFixesHideCursor() not called: supported %d, host cursor hidden %d",
                      m_host_cur_hide_supported, m_host_cur_hidden);
    }

} // cPCOIP_AGENT_GUI::hide_host_cursor()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::show_host_cursor
//      Show the host cursor if it is hidden
//
void cPCOIP_AGENT_GUI::show_host_cursor(void)
{
    // switch on host cursor
    // Appears to fire X error if not already hidden -> but no way to query that
    if (m_host_cur_hide_supported && m_host_cur_hidden)
    {
        Window root = XDefaultRootWindow(m_xdisplay);
        mTERA_LOG_DBG("Showing host cursor (root = %d)", root);
        XFixesShowCursor(m_xdisplay, root);
        XFlush(m_xdisplay);
        m_host_cur_hidden = 0;
    }
    else
    {
        mTERA_LOG_DBG("XFixesShowCursor() not called: supported %d, host cursor hidden %d",
                      m_host_cur_hide_supported, m_host_cur_hidden);
    }

} // cPCOIP_AGENT_GUI::show_host_cursor()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::set_twinview_mode
//      Enables/disables twinview mode
//
//  Description:
//      This function configures the flag that keeps track of whether TwinView
//      is enabled.
//
void cPCOIP_AGENT_GUI::set_twinview_mode(
    INT enable
    )
{
    if (enable)
    {
        m_twinview_active = 1;
    }
    else
    {
        m_twinview_active = 0;
    }
    mTERA_LOG_DBG("m_twinview_active %d, enable %d", m_twinview_active, enable);
} // cPCOIP_AGENT_GUI::set_twinview_mode()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::set_lcur_lkbd_mode
//      Enables/disables local cursor and keyboard mode
//
//  Description:
//      This function configures the flag that keeps track of whether the local
//      cursor and keyboard feature is enabled.
//
void cPCOIP_AGENT_GUI::set_lcur_lkbd_mode(
    INT enable
    )
{
    if (enable)
    {
        m_lcur_lkbd_en = 1;
    }
    else
    {
        m_lcur_lkbd_en = 0;

        // always display host cursor when local cursor feature is disabled
        show_host_cursor();
    }
} // cPCOIP_AGENT_GUI::set_lcur_lkbd_mode()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::set_disc_exit_mode
//      Enables/disables disc and exit options on the agent right-click menu
//
//  Description:
//      This function configures the flag that keeps track of whether the
//      disconnect and exit options are enabled on the agent right click menu.
//
void cPCOIP_AGENT_GUI::set_disc_exit_mode(
    INT enable
    )
{
    if (enable)
    {
        m_disc_exit_popup_enabled = 1;
    }
    else
    {
        m_disc_exit_popup_enabled = 0;
    }
} // cPCOIP_AGENT_GUI::set_disc_exit_mode()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::set_remote_cur_sync_settings
//      Set the remote cursor synchronization settings
//
//  Description:
//      This function configures the remote cursor synchronization settings.
//      The enable flag controls whether the host software monitors remote
//      cursor position changes initiated by something other than the client
//      and sends them to the client. The period setting determines how often
//      (in number of milliseconds) the host SW checks to see if the cursor
//      was moved by something otehr than the client.
//
//      Note: the procesing load of the host SW increases as the period is
//      lowered.
//
void cPCOIP_AGENT_GUI::set_remote_cur_sync_settings(
    INT enable,
    long period
    )
{
    if (enable)
    {
        m_rcs_enabled = 1;
    }
    else
    {
        m_rcs_enabled = 0;
    }

    m_rcs_period = period;

    // Create and start the heartbeat timer
    m_heartbeat_timer.Start(m_rcs_period, wxTIMER_CONTINUOUS);

} // cPCOIP_AGENT_GUI::set_remote_cur_sync_settings()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_SESS::set_topo_sent_status
//      Update state of variable that tracks whether valid topology data has
//      been sent to the client
//
//  Description:
//      This function updates the flag that keeps track of whether valid
//      topology data has been sent to the client.
//
void cPCOIP_AGENT_GUI::set_topo_sent_status(
    INT enable
    )
{
    if (enable)
    {
        m_topo_sent_ok = 1;
    }
    else
    {
        m_topo_sent_ok = 0;
    }
} // cPCOIP_AGENT_GUI::set_topo_sent_status()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::is_host_cur_hiding_supported
//      Returns non-zero value if the host cursor can be hidden
//
//  Description:
//      This function returns a non-zero value if the system supports
//      hiding the host cursor.
//
int cPCOIP_AGENT_GUI::is_host_cur_hiding_supported(void)
{
    if (m_host_cur_hide_supported)
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cPCOIP_AGENT_GUI::is_host_cur_hiding_supported()

// --------------------------------------------------------------------------
//  is_user_logged_in_locally
///     Check if a user is logged in locally
//
//  Description:
///     This function returns zero if a user is not logged in locally and 1 if
///     a user is logged in locally. If an error occurs and the function cannot
///     determine if a user is logged in locally the function returns -1.
///
///     The function uses the "who -u" command to determine if a user is logged
///     in locally. This command returns a list of the users logged in.
///     This function considers a user to be logged in locally if the output
///     contains the string "(:" or the second column starts with the colon
///     character ':' followed by a decimal number.
///
///     Below is a sample of the output when the pcoiplab user is logged in
///     locally and the apexlab user is logged in using ssh on a CentOS 6.4
///     system.
///
///       NAME     LINE         TIME             IDLE          PID COMMENT
///       pcoiplab tty7         2013-04-12 16:39  old         7076 (:0)
///       pcoiplab pts/0        2013-04-12 16:39 00:24        7315 (:0.0)
///       pcoiplab pts/1        2013-04-12 16:39 00:10        7315 (:0.0)
///       pcoiplab pts/2        2013-04-12 16:39   .          7315 (:0.0)
///       pcoiplab pts/3        2013-04-12 16:39 00:14        7315 (:0.0)
///       apexlab  pts/4        2013-04-12 16:38 00:06        6805 (10.200.2.10)
///
///     The first 5 entries are associated with the locally logged in user and
///     the last line is associated with the SSH session.
///
///     Below is a sample of the output when the pcoiplab user is logged in
///     locally and the apexlab user is logged in using ssh on a CentOS 5.9
///     system.
///
///       NAME     LINE         TIME             IDLE          PID COMMENT
///       apexlab  pts/1        2013-06-24 14:34   .         31970 (10.200.2.10)
///       pcoiplab :0           2013-06-24 14:36   ?         32123
///       pcoiplab pts/2        2013-06-24 14:50   .          1706 (:0.0)
///
///     The last 2 entries are associated with the locally logged in user and
///     the first line is associated with the SSH session.
//
int cPCOIP_AGENT_GUI::is_user_logged_in_locally(void)
{
    FILE *fp;
    int ret;
    int rv=-1;

    fp = popen("who -u", "r");
    if (fp == NULL)
    {
        mTERA_LOG_ERR("popen() failed, code=%d (%s)", errno, strerror(errno));
    }
    else
    {
        rv = 0;
        char line[500];
        char username[100];
        int  disp_num;
        while(fgets(line,sizeof line,fp)!= NULL) /* read a line from a file */
        {
            // Check for CentOS 5
            if (sscanf(line, "%s :%d", username, &disp_num) == 2)
            {
                // exit if a user is logged in locally
                mTERA_LOG_DBG("user %s logged in locally :%d", username, disp_num);
                rv = 1;
                break;
            }
            // Check for CentOS 6
            else if (strstr(line, " (:") != NULL)
            {
                // exit if a user is logged in locally
                mTERA_LOG_DBG("user logged in locally");
                rv = 1;
                break;
            }
        }

        ret = pclose(fp);

        // Log error message if pclose failed. Note: pclose returns -1
        // and sets errno = ECHILD when the child process created by the
        // popen() call above has finished, which is expected. This is why
        // the code below ignores this error.
        if ((ret == -1) && (errno != ECHILD))
        {
            mTERA_LOG_ERR("pclose() failed, code=%d (%s)", errno, strerror(errno));
        }
    }

    if (rv == 0)
    {
        mTERA_LOG_DBG("No local logins");
    }
    else if (rv == -1)
    {
        mTERA_LOG_ERR("Unable to determine whether user logged in locally!");
    }
    return rv;
} // is_user_logged_in_locally()
