////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Link Class KMP Functions Implementation
///
///     This file implements link class environment feature. All environment
///     feature related operations are dealt with in this file. This feature
///     offers:
///         - get/set OSD mouse speed
///         - set local cursor enable/disable
///         - set cursor bitmap
///         - set cursor state (visible/invisible) and position
///         - get/set OSD mouse speed
///         - open/close HDC
///         - read PRI context
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_types.h"
#include "tera_log.h"
#include "tera_errors.h"
#include "tera_mem.h"
#include "tera_pcoip_link.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_pcoip_hdc_ioctl.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

#define MAX_DIM_ALPHA_BMP   64          // Maximum height/width
#define MAX_DIM_COLOR_BMP   64          // Maximum height/width
#define MAX_DUMP_CUR_STR    ((32*3)+5)  // 32 2 character bytes plus a space
                                        // separator & 5 bytes for terminator

#define XOR_MASK_BYTES_PER_PIXEL    4   // number of bytes needed to store one
                                        // pixel of the XOR mask in a color
                                        // bitmask
#define AND_MASK_BITS_PER_BYTE      8   // number of bits contained in each
                                        // byte of the AND mask in a color
                                        // bitmap
#define ALPHA_MASK_BYTES_PER_PIXEL  4   // number of bytes in the bitmap needed
                                        // to store one pixel of an alpha bitmap

#define strcat_s(s, n, ct) strcat((s), (ct))

// ---- Local Variables -----------------------------------------------------

// driver local cursor report processing algorith type table
static const char *lcur_evt_alg_str_tbl[TERA_PCOIP_HDC_LCUR_MAX_ALG] = {
    "Absolute",
    "Relative"
};

// This lookup table converts a HEX value to its binary representation.
#ifdef HOST_CUR_BMP_DBG
static const char *hex_2_bin[16] = {
    "0000",
    "0001",
    "0010",
    "0011",
    "0100",
    "0101",
    "0110",
    "0111",
    "1000",
    "1001",
    "1010",
    "1011",
    "1100",
    "1101",
    "1110",
    "1111"
};
#endif

// This lookup table returns a value that determines which bits in a byte
// of a bitmap mask should be inspected. The index of the table is a counter
// from 0 to 7. The function that uses this table supplies an index in the
// range 1 through 7. Data in the AND bitmap mask is stored from MSB to LSB.
// When one bit of the mask is inspected the code must look at the MSB of the
// data byte in the mask.
static const UINT8 and_msk_lkup_table[8] = {
    0x00,
    0x80,
    0xC0,
    0xE0,
    0xF0,
    0xF8,
    0xFC,
    0xFE
};

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::kmp_pri_ctxt_reset
///     Reset local copy of the negotiated KMP PRI context from the firmware
//
//  Description:
///     This function resets the local copy of the negotiated KMP PRI context
///     from the firmware. This basically disables all KMP functionality on
///     the HDC.
///
/// \return
///     None
///
void cTERA_PCOIP_LINK::kmp_pri_ctxt_reset(void)
{
    mTERA_MEM_ZERO(&m_kmp_pri_ctxt, sizeof(m_kmp_pri_ctxt));
    m_kmp_pri_ctxt_valid = 0;
} // cTERA_PCOIP_LINK::kmp_pri_ctxt_reset()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::kmp_pri_ctxt_get
///     Get the negotiated KMP PRI context from the firmware
//
//  Description:
///     This function gets the negotiated KMP PRI context from the firmware. The
///     PRI context is only valid during an active PCoIP session.
///
/// \return
///     None
///
void cTERA_PCOIP_LINK::kmp_pri_ctxt_get(void)
{
    sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT pri_ctxt;
    INT32                             ret;

    mTERA_MEM_ZERO(&pri_ctxt, sizeof(pri_ctxt));

    // Fill out the APDU type and PRI number
    pri_ctxt.apdu_type = TERA_PCOIP_HCC_APDU_KMP_TYPE_PRI_CTXT;
    pri_ctxt.pri_num = 0;

    // Send a synch ioctl to the device
    ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_KMP_PRI_CTXT_GET,
                                &pri_ctxt, sizeof(pri_ctxt),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HCC_IOCTL_KMP_PRI_CTXT_GET failed (ret=%d)", ret);
    }
    else
    {
        // Update the cached version of the PRI context with the new one
        mTERA_MEM_COPY(&m_kmp_pri_ctxt, &pri_ctxt, sizeof(m_kmp_pri_ctxt));
        m_kmp_pri_ctxt_valid = 1;

        mTERA_LOG_DBG("pri_num=%u", pri_ctxt.pri_num);
        mTERA_LOG_DBG("kmp_supported=%u", pri_ctxt.kmp_supported);
        mTERA_LOG_DBG("kmp_enabled=%u", pri_ctxt.kmp_enabled);
        mTERA_LOG_DBG("encodings_supported=%u", pri_ctxt.encodings_supported);
        mTERA_LOG_DBG("pcoip_kmp_1_supported=%u",
                      pri_ctxt.pcoip_kmp_1_supported);
        mTERA_LOG_DBG("cur_shape_caches_supported=%u",
                      pri_ctxt.cur_shape_caches_supported);
        mTERA_LOG_DBG("cur_shape_cache_size_supported=%u",
                      pri_ctxt.cur_shape_cache_size_supported);
        mTERA_LOG_DBG("cur_shape_caches=%u",
                      pri_ctxt.cur_shape_caches);
        mTERA_LOG_DBG("cur_shape_cache_size=%u", pri_ctxt.cur_shape_cache_size);
        mTERA_LOG_DBG("cur_shape_bmp_supported=%u",
                      pri_ctxt.cur_shape_bmp_supported);
        mTERA_LOG_DBG("cur_shape_bmp_alpha=%u", pri_ctxt.cur_shape_bmp_alpha);
        mTERA_LOG_DBG("cur_shape_bmp_color=%u", pri_ctxt.cur_shape_bmp_color);
        mTERA_LOG_DBG("cur_shape_max_size_supported=%u",
                      pri_ctxt.cur_shape_max_size_supported);
        mTERA_LOG_DBG("cur_shape_max_size_x=%u", pri_ctxt.cur_shape_max_size_x);
        mTERA_LOG_DBG("cur_shape_max_size_y=%u", pri_ctxt.cur_shape_max_size_y);
        mTERA_LOG_DBG("kbd_client_auto_repeat_supported=%u",
                      pri_ctxt.kbd_client_auto_repeat_supported);
        mTERA_LOG_DBG("kbd_client_auto_repeat_enable=%u",
                      pri_ctxt.kbd_client_auto_repeat_enable);
        mTERA_LOG_DBG("cur_shape_fadeout_supported=%u",
                      pri_ctxt.cur_shape_fadeout_supported);
        mTERA_LOG_DBG("cur_shape_fadeout_enable=%u",
                      pri_ctxt.cur_shape_fadeout_enable);
        mTERA_LOG_DBG("cur_speed_update_supported=%u",
                      pri_ctxt.cur_speed_update_supported);
        mTERA_LOG_DBG("cur_speed_update_enable=%u",
                      pri_ctxt.cur_speed_update_enable);
        mTERA_LOG_DBG("usb_supported=%u",
                      pri_ctxt.usb_supported);
        mTERA_LOG_DBG("usb_enable=%u",
                      pri_ctxt.usb_enable);
    }
} // cTERA_PCOIP_LINK::kmp_pri_ctxt_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hdc_open
///     Open the HDC channel
//
//  Description:
///     This function opens the HDC channel by sending a HDC Comm Control APDU
///     to the HCC channel with the open channel flag equal to 1. This IOCTL
///     should complete right away. Next it sends another IOCTL to the driver
///     which will return when the HDC channel is open or when the timeout
///     expires. The function returns an indication of whether the HDC channel
///     was successfully opened.
///
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::hdc_open(void)
{
    INT32                              ret;
    sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL comm_ctrl;

    // Fill out the APDU
    comm_ctrl.apdu_type = TERA_PCOIP_HCC_APDU_KMP_TYPE_COMM_CTRL;
    comm_ctrl.open_channel = 1;

    // Send a synch ioctl to the device, this completes right away
    ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL,
                                &comm_ctrl, sizeof(comm_ctrl),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL(OPEN) failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL(OPEN) okay");

        //
        // Send ioctl to HDC which blocks until the HDC is opened or the
        // call times out.
        //
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_HDC_OPEN,
                                    NULL, 0,
                                    5000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HDC_IOCTL_KMP_HDC_OPEN(OPEN) failed (ret=%d)", ret);

            // Something went wrong so let's close the HDC
            hdc_close();
        }
        else
        {
            mTERA_LOG_DBG("TERA_PCOIP_HDC_IOCTL_KMP_HDC_OPEN(OPEN) okay");
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::hdc_open()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hdc_close
///     Close the HDC channel
//
//  Description:
///
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::hdc_close(void)
{
    INT32                              ret;
    sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL comm_ctrl;

    mTERA_MEM_ZERO(&comm_ctrl, sizeof(comm_ctrl));

    // Fill out the APDU
    comm_ctrl.apdu_type = TERA_PCOIP_HCC_APDU_KMP_TYPE_COMM_CTRL;
    comm_ctrl.open_channel = 0;

    // Send a synch ioctl to the device
    ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL,
                                &comm_ctrl, sizeof(comm_ctrl),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("TERA_PCOIP_HCC_IOCTL_KMP_HDC_COMM_CTRL okay");
    }

    //
    // Call the HDC ioctl to close the HDC regardless of the return value from
    // the HCC ioctl call
    //
    m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_HDC_CLOSE,
                          NULL, 0,
                          1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HDC_IOCTL_KMP_HDC_CLOSE failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("TERA_PCOIP_HDC_IOCTL_KMP_HDC_CLOSE okay");
    }

    return ret;
} // cTERA_PCOIP_LINK::hdc_close()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_lkbd_dev_feat_stat_get
///     Get the local cursor and keyboard feature status
//
//  Description:
///     This function gets a copy of the local cursor and keyboard feature
///     status from the device. This function queries the client for the
///     status. As such, this function is only meaningful when called during
///     a session.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_lkbd_dev_feat_stat_get(
    UINT32 *feat_stat
    )
{
    INT32                              ret;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT apdu;

    mTERA_ASSERT(feat_stat != NULL);

    mTERA_MEM_ZERO(&apdu, sizeof(apdu));

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_FEAT_STAT,
                                &apdu, sizeof(apdu),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_FEAT_STAT_GET failed (ret=%d)", ret);
    }
    else
    {
        // Copy the feature status to the user
        *feat_stat = apdu.err;
        mTERA_LOG_DBG("HDC_IOCTL_KMP_FEAT_STAT_GET ok, stat=0x%x", apdu.err);
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_lkbd_dev_feat_stat_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_lkbd_last_feat_stat_get
///     Get copy of last local cursor and keyboard feature status
//
//  Description:
///     This function gets a copy of the last local cursor and keyboard feature
///     status message sent from the firmware to the driver.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_lkbd_last_feat_stat_get(
    UINT32 *feat_stat,
    UINT32 *feat_stat_cnt
    )
{
    INT32                              ret;
    sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT apdu;

    mTERA_ASSERT(feat_stat != NULL);
    mTERA_ASSERT(feat_stat_cnt != NULL);

    mTERA_MEM_ZERO(&apdu, sizeof(apdu));

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_LAST_FEAT_STAT,
                                &apdu, sizeof(apdu),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_LAST_FEAT_STAT failed (ret=%d)", ret);
    }
    else
    {
        // Copy the feature status to the user
        *feat_stat = apdu.err;
        *feat_stat_cnt = apdu.cnt;
        mTERA_LOG_DBG("HDC_IOCTL_KMP_LAST_FEAT_STAT_GET ok, stat=0x%x, cnt=%d",
                      apdu.err, apdu.cnt);
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_lkbd_last_feat_stat_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_pointer_location_set_if_new
///     Set the local cursor pointer position if the host and client cursors
///     differ
//
//  Description:
///     The caller passes this function the current location of the host
///     cursor. This function compares the host cursor location to the location
///     of the last X client cursor reports. If all locations differ the
///     driver sends a message to the client to update the location of the
///     client cursor.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///     TERA_ERR_INVALID_ARG
///
INT32 cTERA_PCOIP_LINK::lcur_pointer_location_set_if_new(
    UINT16 abs_x,
    UINT16 abs_y
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW cur_pos;

    // Fill out the message sent to the driver
    cur_pos.abs_x = abs_x;
    cur_pos.abs_y = abs_y;

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS_IF_NEW,
                                &cur_pos, sizeof(cur_pos),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS_IF_NEW failed (ret=%d)", ret);
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_pointer_location_set_if_new()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_pointer_location_set
///     Set the local cursor pointer position
//
//  Description:
///     This function sets the local cursor pointer position.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_pointer_location_set(
    UINT32 tstamp,
    UINT16 x_pos,
    UINT16 y_pos
    )
{
    INT32                            ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS cur_pos;

    mTERA_MEM_ZERO(&cur_pos, sizeof(cur_pos));

    // Fill out the APDU
    cur_pos.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_POS;
    cur_pos.tstamp = tstamp;
    cur_pos.x = x_pos;
    cur_pos.y = y_pos;

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS,
                                &cur_pos, sizeof(cur_pos),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS set okay, tstamp=%d, x,y=%d,%d",
                      tstamp, x_pos, y_pos);
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_pointer_location_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_pointer_visibility_set
///     Set the local cursor pointer visibility
//
//  Description:
///     This function sets the local cursor visibility. The client local cursor
///     will be visiblle when the visible parameter is non-zero and not visible
///     when the parameter equals 0.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_pointer_visibility_set(
    UINT32 tstamp,
    UINT8  visible
    )
{
    INT32                            ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS cur_vis;

    mTERA_MEM_ZERO(&cur_vis, sizeof(cur_vis));

    // Fill out the APDU
    cur_vis.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_VIS;
    cur_vis.tstamp = tstamp;
    cur_vis.visible = (visible) ? 1 : 0;

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_VIS,
                                &cur_vis, sizeof(cur_vis),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HDC_IOCTL_KMP_CUR_VIS failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("TERA_PCOIP_HDC_IOCTL_KMP_CUR_VIS set okay, tstamp=%d, vis=%d",
            tstamp, cur_vis.visible);
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_pointer_visibility_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lkbd_client_auto_repeat_avail
///     Check if the local keyboard client auto-repeat is enabled
//
//  Description:
///     This function lets the caller know if the local keyboard client auto-
///     repeat is enabled.
///
///     Note: To work with soft clients who ALWAYS does client auto-repeat
///     even though it is negotiated out, the application should set the
///     repeat delay to a large number - something around the order of 35 mins -
///     to prevent the client auto repeat from kicking in.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lkbd_client_auto_repeat_avail(void)
{
    return (m_kmp_pri_ctxt.kbd_client_auto_repeat_supported && m_kmp_pri_ctxt_valid) ? TERA_SUCCESS :
           TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::lkbd_client_auto_repeat_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lkbd_repeat_rates_get
///     Get the local keyboard repeat rates from the client
//
//  Description:
///     This function gets the local keyboard repeat rate settings from the
///     zero client. If the client is responsible for performing keyboard
///     repeat detection the settings stored in the ZC will be used and the
///     settings stored in the host machine's registry will be used with
///     soft clients.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lkbd_repeat_rates_get(
    UINT16 *repeat_delay,
    UINT16 *repeat_interval
    )
{
    INT32                               ret;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT apdu;

    mTERA_LOG_DBG("start");

    if ((repeat_delay == NULL) || (repeat_interval == NULL))
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (lkbd_client_auto_repeat_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&apdu, sizeof(apdu));

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_KBD_REPEAT_GET,
                                    &apdu, sizeof(apdu),
                                    2000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_KBD_REPEAT_GET failed (ret=%d)", ret);
        }
        else
        {
            // Copy the local cursor speed to the user
            *repeat_delay    = apdu.delay;
            *repeat_interval = apdu.interval;
            mTERA_LOG_DBG("HDC_IOCTL_KMP_KBD_REPEAT_GET ok, delay=%d, interval=%d",
                          *repeat_delay,
                          *repeat_interval);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lkbd_repeat_rates_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lkbd_repeat_rates_set
///     Set the local keyboard repeat delay and rate
//
//  Description:
///     This function sets the local keyboard repeat delay and rate.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lkbd_repeat_rates_set(
    UINT16 repeat_delay,
    UINT16 repeat_interval
    )
{
    INT32                               ret;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT kbd_repeat;

    mTERA_MEM_ZERO(&kbd_repeat, sizeof(kbd_repeat));

    // Fill out the APDU
    kbd_repeat.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REPEAT;
    kbd_repeat.interval = repeat_interval;
    kbd_repeat.delay = repeat_delay;

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_KBD_REPEAT,
                                &kbd_repeat, sizeof(kbd_repeat),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_KBD_REPEAT failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("HDC_IOCTL_KMP_KBD_REPEAT set okay, delay=%d, int=%d",
                      repeat_delay, repeat_interval);
    }

    return ret;
} // cTERA_PCOIP_LINK::lkbd_repeat_rates_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lkbd_kmp_alg_type_set
///     Set the algorithm that determines how the KMP events are processed
//
//  Description:
///     This function sets the algorithm that determines how the KMP events
///     are processed by the driver.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
#if 0 // TBD _ not supported by the driver
INT32 cTERA_PCOIP_LINK::lkbd_kmp_alg_type_set(
    eTERA_PCOIP_LINK_KMP_EVT_ALG_TYPE alg_type
    )
{
    INT32                            ret;
    sTERA_PCOIP_HDC_KMP_EVT_ALG_CTRL alg_cfg;

    mTERA_MEM_ZERO(&alg_cfg, sizeof(alg_cfg));

    // Fill out the structure
    if (alg_type == TERA_PCOIP_LINK_KMP_TSTAMP_ALG)
    {
        alg_cfg.alg_type = TERA_PCOIP_HDC_KMP_TSTAMP_ALG;
    }
    else
    {
        alg_cfg.alg_type = TERA_PCOIP_HDC_KMP_MIN_DELAY_ALG;
    }

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_EVT_ALG_SET,
                                &alg_cfg, sizeof(alg_cfg),
                                500);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_EVT_ALG_SET failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("HDC_IOCTL_KMP_EVT_ALG_SET set okay, alg type=%d",
                      alg_type);
    }

    return ret;
} // cTERA_PCOIP_LINK::lkbd_kmp_alg_type_set()
#endif

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::kmp_driver_cfg_set
///     Configure the different modes supported by the driver
//
//  Description:
///     This function sets the various operating modes supported by the
///     driver. The following modes are currently supported.
///
///       LCUR_EVT_ALG
///         Determines what type of HID reports (absolute or relative) are
///         submitted to the OS.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::kmp_driver_cfg_set(
    eTERA_PCOIP_LINK_LCUR_EVT_ALG_TYPE lcur_mode
    )
{
    INT32                               ret;
    sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL settings;

    mTERA_MEM_ZERO(&settings, sizeof(settings));

    // Fill out the structure
    if (lcur_mode == TERA_PCOIP_LINK_LCUR_REL_ALG)
    {
        settings.lcur_mode = TERA_PCOIP_HDC_LCUR_REL_ALG;
    }
    else
    {
        settings.lcur_mode = TERA_PCOIP_HDC_LCUR_ABS_ALG;
    }

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_DRIVER_CFG_SET,
                                &settings, sizeof(settings),
                                500);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_DRIVER_CFG_SET failed (ret=%d)", ret);
    }
    else
    {
        mTERA_LOG_DBG("cursor event processing mode: %s",
                      lcur_evt_alg_str_tbl[lcur_mode]);
    }

    return ret;
} // cTERA_PCOIP_LINK::kmp_driver_cfg_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_shape_bmp_send_avail
///     Check if sending a local cursor shape bitmap is available
//
//  Description:
///     This function lets the caller know if sending a local cursor shape
///     bitmap is available.
///
///     Note: At this time, this function only checks to see if alpha-type
///     bitmaps are supported. To suppose color or another type of bitmap
///     format, this function will need to be modified.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lcur_shape_bmp_send_avail(void)
{
    return (m_kmp_pri_ctxt.cur_shape_bmp_alpha && m_kmp_pri_ctxt_valid) ? TERA_SUCCESS :
           TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::lcur_shape_bmp_send_avail()

#ifdef HOST_CUR_BMP_DBG
// --------------------------------------------------------------------------
//  print_alpha_bitmap
///     Print an image in the log file representing the shape of an alpha bitmap
//
//  Description:
///     This function is used for debugging purposes only.  It prints the shape
///     of the specified alpha bitmap into the log file.
//
INT32 cTERA_PCOIP_LINK::print_alpha_bitmap(
    UINT8               width,          ///< width of the bitmap
    UINT8               height,         ///< height of the bitmap
    UINT8               *bitmap         ///< buffer containing the bitmap
    )
{
    INT32   x;
    INT32   y;
    UINT8  *pPixel;
    UINT32  pixel;
    char    row_str[MAX_DIM_ALPHA_BMP+1];

    // Limit the size of the alpha bitmaps displayed by this routine
    if ( (width > MAX_DIM_ALPHA_BMP) || (height > MAX_DIM_ALPHA_BMP) )
    {
        mTERA_LOG_ERR("====> Alpha Ptr: too large to display! (width=%d height=%d)", width, height);
        return TERA_ERR_INVALID_ARG;
    }

    mTERA_LOG_DBG("====> Alpha Ptr: width=%d height=%d", width, height);

    // Loop and draw the bitmap line by line
    pPixel = (UINT8 *) bitmap;
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            // Decide on a char to represent this pixel
            pixel = pPixel[0] | (pPixel[1] << 8) | (pPixel[2] << 16) | (pPixel[3] << 24);
            if ( (pixel & 0xFF000000) == 0 )
            {
                // Fully transparent pixel (alpha = 0)
                strcat_s(row_str, sizeof(row_str), " ");
            }
            else if ( (pixel & 0xFF000000) != 0xFF000000 )
            {
                // semi-transparent pixel ( 0 < alpha < 255)
                strcat_s(row_str, sizeof(row_str), ".");
            }
            else
            {
                // Opaque pixel (alpha = 255)

                if ( (pixel & 0x00FFFFFF) == 0 )
                {
                    // White pixel (RGB = 0,0,0)
                    strcat_s(row_str, sizeof(row_str), "w");
                }
                else if ( (pixel & 0x00FFFFFF) == 0x00FFFFFF )
                {
                    // Black pixel (RGB = 255,255,255)
                    strcat_s(row_str, sizeof(row_str), "B");
                }
                else
                {
                    // Color pixel (RGB != 255,255,255 || 0,0,0)
                    strcat_s(row_str, sizeof(row_str), "c");
                }
            }

            // Advance to the next pixel
            pPixel += sizeof(UINT32);
        }

        // Draw a row
        mTERA_LOG_DBG("====>          : %s", row_str);
    }

    return TERA_SUCCESS;
}

// --------------------------------------------------------------------------
//  print_color_bitmap_1bpp
///     Print an image in the log file representing the shape of a color bitmap
//
//  Description:
///     This function is used for debugging purposes only.  It prints the shape
///     of the specified color bitmap into the log file.
//
static INT32 print_color_bitmap_1bpp(
    UINT16       width,             ///< rectangle width
    UINT16       height,            ///< rectangle height
    const UINT8  *pAndMsk,          ///< AND mask data (alpha in 1 bpp format)
    const UINT8  *pXorMsk           ///< XOR mask data (color in 1 bpp format)
    )
{
    INT32   x;
    INT32   y;
    char    row_str[MAX_DIM_COLOR_BMP+1];
    UINT8   *savedXorPtr;                  // holds initial XorMask pointer for the I-beam hack
    UINT8   *savedAndPtr;
    UINT8   andMsk=0;
    UINT8   xorMsk=0;
    UINT8   pad;
    UINT8   num_pad_bytes;

    // Figure out how many bytes of padding there are per row for the two masks
    if ((width % 32) == 0)
    {
        num_pad_bytes = 0;
    }
    else
    {
        num_pad_bytes = (32 - (width % 32)) / 8;
    }

    savedXorPtr = (UINT8 *) pXorMsk;
    savedAndPtr = (UINT8 *) pAndMsk;

    // Limit the size of the color bitmaps displayed by this routine
    if ( (width >MAX_DIM_COLOR_BMP) || (height > MAX_DIM_COLOR_BMP) )
    {
        mTERA_LOG_ERR("====> Color Ptr: too large to display! (width=%d height=%d)", width, height);
        return TERA_ERR_INVALID_ARG;
    }

    mTERA_LOG_DBG("====> 1bpp Color Ptr: width=%d height=%d", width, height);

    // Loop and print the and mask line by line
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            if( (x & 0x07) == 0 )
            {
                andMsk = *pAndMsk++;    // Read a new mask every 8 pixels
            }

            if( !(andMsk & 0x80) )
            {
                strcat_s(row_str, sizeof(row_str), "0");
            }
            else
            {
                strcat_s(row_str, sizeof(row_str), "1");
            }

            // advance the AND mask for the the next pixel
            andMsk <<= 1;

        }
        for (pad = 0; pad < num_pad_bytes; pad++)
        {
            // Pad out the rows in the masks
            andMsk = *pAndMsk++;
        }

        // Draw a row
        mTERA_LOG_DBG("AM ====>          : |%s|", row_str);
    }
    // Loop and print the XOR mask line by line
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            if( (x & 0x07) == 0 )
            {
                xorMsk = *pXorMsk++;    // Read a new mask every 8 pixels
            }
            if( !(xorMsk & 0x80) )
            {
                strcat_s(row_str, sizeof(row_str), "0");
            }
            else
            {
                strcat_s(row_str, sizeof(row_str), "1");
            }

            // advance the XOR mask for the the next pixel
            xorMsk <<= 1;
        }
        for (pad = 0; pad < num_pad_bytes; pad++)
        {
            // Pad out the rows in the masks
            xorMsk = *pXorMsk++;
        }

        // Draw a row
        mTERA_LOG_DBG("XM ====>          : |%s|", row_str);
    }

    pAndMsk = savedAndPtr;
    pXorMsk = savedXorPtr;

    // Loop and draw the bitmap line by line
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            if( (x & 0x07) == 0 )
            {
                andMsk = *pAndMsk++;    // Read a new mask every 8 pixels
            }

            if( (x & 0x07) == 0 )
            {
                xorMsk = *pXorMsk++;    // Read a new mask every 8 pixels
            }
            // check mask value (mask for the 1st pixel is at the MSB bit)
            if( !(andMsk & 0x80) )
            {
                // this pixel is not transparent (Alpha = 0xFF)
                if ( xorMsk & 0x80 )
                {
                    // this is a WHITE pixel (RGB = FF FF FF)
                    strcat_s(row_str, sizeof(row_str), "w");
                }
                else
                {
                    // this is a BLACK pixel (RGB = 00 00 00)
                    strcat_s(row_str, sizeof(row_str), "B");
                }
            }
            else if (xorMsk & 0x80)
            {
              // This is an inverted pixel (Alpha = 0xFF);
              strcat_s(row_str, sizeof(row_str), "A");
            }
            else
            {
                // this pixel is transparent (Alpha = 0x00)
                strcat_s(row_str, sizeof(row_str), " ");
            }

            // advance the AND & XOR masks for the the next pixel
            andMsk <<= 1;
            xorMsk <<= 1;
        }
        for (pad = 0; pad < num_pad_bytes; pad++)
        {
          // Pad out the rows in the masks
          xorMsk = *pXorMsk++;
          andMsk = *pAndMsk++;
        }

        // Draw a row
        mTERA_LOG_DBG("====>          : %s", row_str);
    }

    return TERA_SUCCESS;
}

// --------------------------------------------------------------------------
//  print_color_bitmap_32bpp
///     Print an image in the log file representing the shape of a color bitmap
//
//  Description:
///     This function is used for debugging purposes only.  It prints the shape
///     of the specified color bitmap into the log file.
//
static INT32 print_color_bitmap_32bpp(
    UINT16       width,             ///< rectangle width
    UINT16       height,            ///< rectangle height
    const UINT8  *pAndMsk,          ///< AND mask data (alpha in 1 bpp format)
    const UINT8  *pXorMsk           ///< XOR mask data (color in 32 bpp format)
    )
{
    INT32   x;
    INT32   y;
    char    row_str[MAX_DIM_COLOR_BMP+1];
    UINT8   *savedXorPtr;                  // holds initial XorMask pointer for the I-beam hack
    UINT8   *savedAndPtr;
    UINT8   andMsk=0;
    UINT32  xorMsk;
    UINT8   pad;
    UINT8   num_xor_pad_bytes;
    UINT8   num_and_pad_bytes;

    if ((width % 32) == 0)
    {
        num_and_pad_bytes = 0;
    }
    else
    {
        num_and_pad_bytes = (32 - (width % 32)) / 8;
    }
    // Figure out how many bytes of padding there are per row for the XOR mask
    if (((width*32)%32) == 0)
    {
        num_xor_pad_bytes = 0;
    }
    else
    {
        num_xor_pad_bytes = (32 - ((width*32)%32)) / 8;
    }

    savedXorPtr = (UINT8 *) pXorMsk;
    savedAndPtr = (UINT8 *) pAndMsk;

    // Limit the size of the color bitmaps displayed by this routine
    if ( (width >MAX_DIM_COLOR_BMP) || (height > MAX_DIM_COLOR_BMP) )
    {
        mTERA_LOG_ERR("====> Color Ptr: too large to display! (width=%d height=%d)", width, height);
        return TERA_ERR_INVALID_ARG;
    }

    mTERA_LOG_DBG("====> 32bpp Color Ptr: width=%d height=%d", width, height);

    // Loop and print the and mask line by line
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            if( (x & 0x07) == 0 )
            {
                andMsk = *pAndMsk++;    // Read a new mask every 8 pixels
            }

            if( !(andMsk & 0x80) )
            {
                strcat_s(row_str, sizeof(row_str), "0");
            }
            else
            {
                strcat_s(row_str, sizeof(row_str), "1");
            }

            // advance the AND mask for the the next pixel
            andMsk <<= 1;

        }
        for (pad = 0; pad < num_and_pad_bytes; pad++)
        {
            // Pad out the rows in the masks
            andMsk = *pAndMsk++;
        }

        // Draw a row
        mTERA_LOG_DBG("AM ====>          : |%s|", row_str);
    }
    // Loop and print the XOR mask line by line
    for ( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for ( x=0; x<width; x++ )
        {
            xorMsk = pXorMsk[3] | (pXorMsk[2] << 16) | (pXorMsk[1] << 8) | pXorMsk[0];
            if (xorMsk == 0xFFFFFFFF)
            {
                strcat_s(row_str, sizeof(row_str), "w");
            }
            else if (xorMsk == 0x00FFFFFF)
            {
                strcat_s(row_str, sizeof(row_str), "p");
            }
            else if (xorMsk > 0)
            {
                strcat_s(row_str, sizeof(row_str), "c");
            }
            else
            {
                strcat_s(row_str, sizeof(row_str), "0");
            }
            pXorMsk += 4;
        }
        for (pad = 0; pad < num_xor_pad_bytes; pad++)
        {
            // Pad out the rows in the masks
            xorMsk = *pXorMsk++;
        }

        // Draw a row
        mTERA_LOG_DBG("XM ====>          : |%s|", row_str);
    }

    pAndMsk = savedAndPtr;
    pXorMsk = savedXorPtr;

    // Loop and draw the bitmap line by line
    for( y=0; y<height; y++ )
    {
        row_str[0] = 0;
        for( x=0; x<width; x++ )
        {
            // AND mask is supposed to be padded on 32-bit boundary.
            // However, works only if I read it every 8 bits.
            if( ( x & 0x7 ) == 0 )
            {
                andMsk = *pAndMsk++;  // Read a new mask every 8 pixels
            }
            xorMsk = pXorMsk[3] | (pXorMsk[2] << 16) | (pXorMsk[1] << 8) | pXorMsk[0];

            if( !(andMsk & 0x80) )
            {
                // this pixel is not transparent (Alpha = 0xFF)
                // data format = xBGR = blank-Blue-Green-Red in little endian
                strcat_s(row_str, sizeof(row_str), "c");
                pXorMsk += 4;              // we are ingoring the 4th byte which contains no color info
            }
            else if (xorMsk == 0x00FFFFFF)
            {
                // This pixel is an Alph pixel
                strcat_s(row_str, sizeof(row_str), "A");
                pXorMsk += 4;
            }
            else
            {
                // this pixel is transparent (Alpha = 0x00)
                strcat_s(row_str, sizeof(row_str), " ");
                pXorMsk += 4;              // skip XOR mask values for transparent pixel
            }

            // advance the AND mask for the next pixel to the MSB bit
            andMsk <<= 1;
        }

        // XOR mask is padded at 2-byte boundary. Thus, skip a byte for odd widths
        if ( width & 0x01 )
        {
            pXorMsk++;                     // align XOR data to 2-byte boundary for each row
        }
        // Draw a row
        mTERA_LOG_DBG("====>          : %s", row_str);
    }

    return TERA_SUCCESS;
}

// --------------------------------------------------------------------------
//  print_color_bitmap
///     Print an image in the log file representing the shape of a color bitmap
//
//  Description:
///     This function is used for debugging purposes only.  It prints the shape
///     of the specified color bitmap into the log file.
//
//  Notes:
///     This code is similar to the dump_cursor_mask() function. The functions
///     could be consolidated but it hasn't been because it's debug code that
///     is only included in builds that define the HOST_CUR_BMP_DBG flag, which
///     should never be included in a customer build because the build outputs
///     a lot of log messages.
//
INT32 cTERA_PCOIP_LINK::print_color_bitmap(
    UINT8               width,          ///< width of the bitmap
    UINT8               height,         ///< height of the bitmap
    UINT8               and_depth,      ///< depth of the AND mask
    UINT8               xor_depth,      ///< depth of the XOR mask
    UINT8               *bitmap         ///< buffer containing the bitmap
    )
{
    INT32  rv = TERA_SUCCESS;
    UINT32 and_bytes_per_row;
    UINT8 *xor_msk;
    UINT8 *and_msk;

    mTERA_LOG_DBG("==> Color bitmap: %dx%d  and_depth=%d  xor_depth=%d",
                  width,
                  height,
                  and_depth,
                  xor_depth);

    // The color bitmap has an AND mask followed by an XOR mask.
    // The pixels for each line are aligned to start on a 32 bit boundary

    // Only 1, 16, 24 and 32 bpp are supported.
    // The display format can only handle a maximum of 8 bit Alpha mask so assume
    // that the alpha mask can only be 1, 4 or 8  bits.
    if (((and_depth * width) % 32) == 0)
    {
        and_bytes_per_row = (and_depth * width) / 8;
    }
    else
    {
        and_bytes_per_row = ((width * and_depth) + (32 - ((width *and_depth) % 32))) / 8;
    }
    and_msk = (UINT8 *) bitmap;
    xor_msk = (UINT8 *) (((UINT8 *) bitmap) + (and_bytes_per_row*height));
    switch ((and_depth << 8) | (xor_depth))
    {
        case 0x101:
            // AND_DEPTH=1, XOR_DEPTH=1
            print_color_bitmap_1bpp(width, height, and_msk, xor_msk);
            break;
        case 0x120:
            // AND_DEPTH=1, XOR_DEPTH=32
            print_color_bitmap_32bpp(width, height, and_msk, xor_msk);
            break;
        default:
            // Nothing else is currently supported.  Don't assert, raise
            // a log error so we know if we're getting unsupported bmps and need
            // to add support
            rv = TERA_ERR_INVALID_ARG;
            mTERA_LOG_ERR("AND_DEPTH=%0d, XOR_DEPTH=%0d color bmp not supported",
                          and_depth, xor_depth);
           break;
    }
    return rv;
}
#endif // HOST_CUR_BMP_DBG

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::dump_cursor_mask
///     Dump cursor AND or XOR mask
//
//  Description:
///     This function dumps a cursor bitmask to the debug log. This is a
///     debug function that can be used when debugging problems with the
///     cursor bitmaps.
///
///     Note: this function may not interpret all masks properly. It was tested
///     and verified to work with masks with xor_depth = 1 or 32 AND
///     and_depth=1.
///
#if 0
void cTERA_PCOIP_LINK::dump_cursor_mask(
    const char *mask_name,
    WORD biBitCount,
    LONG height,
    LONG width,
    UINT8 *pbits)
{
    int     i;
    int     line;
    int     column;
    int     offset;
    int     width_bytes=0;
    char    row_str[MAX_DUMP_CUR_STR];
    char    msk_data_str[4];    // 2 bytes for the hex data, 1 byte
                                // for space & 1 byte for terminator

    if (biBitCount == 1)
    {
        width_bytes = 4;
    }
    else if (biBitCount == 32)
    {
        width_bytes = width * 4;
    }
    else
    {
        mTERA_LOG_ERR("Not dumping %s bitmap because biBitCount(%d) != 1 or 32",
                      mask_name, biBitCount);
        return;
    }

    // Dump the mask as an array of binary data. This is used when displaying
    // the AND mask.
    if (width_bytes == 4)
    {
        for (i=0; i<(height*width_bytes); i+=width_bytes)
        {
            mTERA_LOG_DBG("%s %03d: %02x %02x %02x %02x %s%s%s%s%s%s%s%s",
                          mask_name,
                          i/width_bytes,
                          pbits[i],
                          pbits[i+1],
                          pbits[i+2],
                          pbits[i+3],
                          hex_2_bin[(pbits[i]>>4)&0x0F],
                          hex_2_bin[(pbits[i])&0x0F],
                          hex_2_bin[(pbits[i+1]>>4)&0x0F],
                          hex_2_bin[(pbits[i+1])&0x0F],
                          hex_2_bin[(pbits[i+2]>>4)&0x0F],
                          hex_2_bin[(pbits[i+2])&0x0F],
                          hex_2_bin[(pbits[i+3]>>4)&0x0F],
                          hex_2_bin[(pbits[i+3])&0x0F]);
        }
    }
    // Dump the mask as 4 arrays of hex data. This is used when displaying
    // the XOR mask. The bitmap consists of 4 bytes of data where each byte
    // represents a Red, Green, Blue and Alpha value. The following code
    // displays 4 different arrays (one for each of the Red, Green, Blue
    // and Alpha values).
    else if ((width_bytes == 32) || (width_bytes == 128))
    {
        // loop through each of the 4 byte masks (red, green, blue, alpha)
        for (offset=0; offset<4; offset++)
        {
            // loop through each line of the mask
            for (line=0; line<(height*width_bytes); line+=width_bytes)
            {
                row_str[0] = 0;

                // loop through each byte in the line
                for (column=0; column<width_bytes; column+=4)
                {
                    sprintf_s(msk_data_str,
                              sizeof(msk_data_str),
                              "%02x ",
                              pbits[line+column+offset]);
                    strcat_s(row_str, sizeof(row_str), msk_data_str);
                }

                mTERA_LOG_DBG("%s %03d: %s",
                              mask_name,
                              line/width_bytes,
                              row_str);
            }
        }
    }
    else
    {
        mTERA_LOG_ERR("SW does not support dumpimg mask with %d byte width",
                      width_bytes);
    }
}
#endif

// --------------------------------------------------------------------------
//  zero_transparent_bits_in_and_mask
///     Returns non-zero value if the AND mask contains no transparent bits
//
//  Description:
///     This function inspects the data in the AND mask and returns a non-zero
///     value if the mask contains no transparent bits.
///
///     The hand cursor drawn by Google maps within a Chrome browser is an
///     application specific cursor with a completely non-transparent AND
///     mask. When this cursor is drawn the hand appears on a square black
///     box.
///
///     The code treats cursors with this type of AND mask as alpha bitmaps.
///
static INT32 zero_transparent_bits_in_and_mask(
    UINT8          *and_data_start,
    unsigned int    and_line_width,
    UINT8           width,
    UINT8           height)
{
    int     i;
    int     j;
    int     bits_to_check;
    UINT8   bits_to_check_mask;

    // loop through each line
    for (i=0; i<height; i++)
    {
        bits_to_check = width;
        j = 0;
        while (bits_to_check > 0)
        {
            if (bits_to_check >= 8)
            {
                // set the mask to look at all bits
                bits_to_check_mask = 0xFF;
            }
            else
            {
                // Retrieve the mask from the lookup table
                bits_to_check_mask = and_msk_lkup_table[bits_to_check];
            }

            // exit if any bits in the AND mask are non-zero
            if (and_data_start[(i*and_line_width)+j] & bits_to_check_mask)
            {
                return 0;
            }

            if (bits_to_check >= 8)
            {
                bits_to_check -= 8;
            }
            else
            {
                bits_to_check = 0;
            }
            j++;
        }
    }
    // if the code reaches this point all bits in the mask were zero
    return 1;
}

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::flip_bitmap
///     Flip the bitmap
//
//  Description:
///     This function creates a flipped bitmap. This is necessary because the
///     color bitmaps returned by the Windows API are flipped. The client
///     devices expect them to be in the other orientation.
///
///     Notes:
///      1) The code assumes the and_depth is always 1 since this is the only
///         supported value for this mask.
///      2) I considered having the calling function flip the bitmap but I
///         decided to do it within this module because the bitmap data needed
///         to be copied to the APDU buffer sent to the driver. This function
///         flips the bitmap while copying the data to the APDU buffer.
///
INT32 cTERA_PCOIP_LINK::flip_bitmap(
    UINT8  width,
    UINT8  height,
    UINT8  xor_depth,
    UINT32 bmp_len,
    UINT8 *src,
    UINT8 *dest,
    INT32 *is_alpha)
{
    int             i;
    unsigned int    and_line_width;
    unsigned int    xor_line_width;
    int             and_src_start_off;
    int             xor_src_start_off;
    INT32           ret = TERA_SUCCESS;
    UINT8          *and_data_start;
    UINT8          *xor_data_start;

    // calculate the line width of each mask
    and_line_width = ((width+31)/32)*4;
    if (xor_depth == 1)
    {
        xor_line_width = ((width+31)/32)*4;
    }
    else
    {
        xor_line_width = width*4;
    }

    // When xor_depth is 1 the XOR mask is the first mask in the input buffer,
    // otherwise the AND mask is the first mask in the buffer. The following
    // code calculates the starting offset of the source data for each mask.
    if(xor_depth == 1)
    {
        and_src_start_off = xor_line_width*height;
        xor_src_start_off = 0;
    }
    else
    {
        and_src_start_off = 0;
        xor_src_start_off = and_line_width*height;
    }

    // Calculate the address of the last line of data for each mask. The code
    // below copies each line starting at the end of the source data and
    // writes it to the first line of the destination mask
    and_data_start = &src[and_src_start_off + and_line_width*(height-1)];
    xor_data_start = &src[xor_src_start_off + xor_line_width*(height-1)];

#if HOST_CUR_BMP_DBG
    mTERA_LOG_DBG("AND MASK: line_width=%d data_start offset=%d and_src_start_off=%d",
                  and_line_width,
                  and_src_start_off + and_line_width*(height-1),
                  and_src_start_off);
    mTERA_LOG_DBG("XOR MASK: line_width=%d data_start offset=%d xor_src_start_off=%d",
                  xor_line_width,
                  xor_src_start_off + xor_line_width*(height-1),
                  xor_src_start_off);
#endif

    // verify there's enough space in the destination buffer
    if (((and_line_width + xor_line_width) * height) > bmp_len)
    {
        mTERA_LOG_ERR("dest buffer too small, aborting copy: alw %d xlw %d h %d bl %d",
                      and_line_width,
                      xor_line_width,
                      height,
                      bmp_len);
        ret = TERA_ERR_FAILURE;
    }
    else
    {
        // if the AND mask is transparent treat the bitmap as an alpha
        // bitmap, in this situation the bitmap is stored in the XOR mask
        if (zero_transparent_bits_in_and_mask(and_data_start, and_line_width, width, height))
        {
            *is_alpha = 1;
            mTERA_LOG_DBG("Zero transparent bits in AND mask");

            // flip the XOR mask by copying one line at a time
            for(i=0; i<height; i++)
            {
                mTERA_MEM_COPY(&dest[i*xor_line_width], xor_data_start, xor_line_width);
                xor_data_start -= xor_line_width;
            }
        }
        else
        {
            // flip the AND mask by copying one line at a time
            for(i=0; i<height; i++)
            {
                mTERA_MEM_COPY(&dest[i*and_line_width], and_data_start, and_line_width);
                and_data_start -= and_line_width;
            }

            // flip the XOR mask by copying one line at a time
            for(i=0; i<height; i++)
            {
                mTERA_MEM_COPY(&dest[i*xor_line_width+(and_line_width*height)], xor_data_start, xor_line_width);
                xor_data_start -= xor_line_width;
            }
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::flip_bitmap()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::init_bitmap_cache
///     This function initializes the cache of stored bitmaps
//
//  Description:
///     This function is called at the start of a session. It allocates
///     storage for the cached bitmaps. It also initializes the variables
///     used to maintain the cached bitmaps.
///
///     The storage for the cached bitmaps must be allocated dynamically
///     because the number of cached bitmaps supported by a client is a
///     negotiated setting.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_NO_MEMORY,
///
INT32 cTERA_PCOIP_LINK::init_bitmap_cache(void)
{
    INT     i;
    INT32   ret;

    // handle the case where the client doesn't support caching bitmaps
    if (!m_kmp_pri_ctxt.cur_shape_caches_supported ||
        !m_kmp_pri_ctxt.cur_shape_cache_size_supported)
    {
        // If the bitmap cache is not empty flush it
        if (m_kmp_bmp_cache.max_num_bmps)
        {
            // this can happen when waking from hibernate
            flush_bitmap_cache();
        }
        mTERA_LOG_DBG("bitmap caching disabled because client does not support it");
        m_kmp_bmp_cache.max_num_bmps = 0;
        ret = TERA_SUCCESS;
    }
    else
    {
        // If the bitmap cache is not empty flush it
        if (m_kmp_bmp_cache.max_num_bmps)
        {
            // this can happen when waking from hibernate
            flush_bitmap_cache();
            mTERA_LOG_DBG("stored bitmap cache is already initialized");

#if 0 // TBD
            // Restore the system cursors in case they are hidden
            if (!SystemParametersInfo(SPI_SETCURSORS, 0, NULL, 0))
            {
                mTERA_LOG_ERR("failed to restore system cursor %u", GetLastError());
            }
#endif
        }
        if (m_kmp_pri_ctxt.cur_shape_caches > MAX_CACHE_BMP)
        {
            m_kmp_bmp_cache.max_num_bmps = MAX_CACHE_BMP;
        }
        else
        {
            m_kmp_bmp_cache.max_num_bmps = m_kmp_pri_ctxt.cur_shape_caches;
        }
        m_kmp_bmp_cache.idx_last_stored_bmp = 0;

        // initialize each bitmap in the cache
        for (i=0; i<m_kmp_bmp_cache.max_num_bmps; i++)
        {
            mTERA_MEM_ZERO(&m_kmp_bmp_cache.bmp[i], sizeof(sTERA_PCOIP_BMP_DATA));

            // Allocate storage for the bitmap data
            m_kmp_bmp_cache.bmp[i].bmp = (UINT8 *)mTERA_MEM_ALLOC(m_kmp_pri_ctxt.cur_shape_cache_size);
            if (m_kmp_bmp_cache.bmp[i].bmp == NULL)
            {
                mTERA_LOG_ERR("unable to allocate storage for %d cached bitmap",
                              m_kmp_bmp_cache.max_num_bmps - i);
                m_kmp_bmp_cache.max_num_bmps = i;
                break;
            }
        }
        ret = TERA_SUCCESS;

        mTERA_LOG_DBG("allocated storage for %d cached bitmaps with size %d",
                      m_kmp_bmp_cache.max_num_bmps,
                      m_kmp_pri_ctxt.cur_shape_cache_size);
    }

    return ret;
}  // cTERA_PCOIP_LINK::init_bitmap_cache()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::flush_bitmap_cache
///     This function flushes the cache of stored bitmaps
//
//  Description:
///     This function is called at the end of a session. It frees the
///     storage that was allocated at the start of the session to store the
///     cached bitmaps (if any was allocated).
///
/// \return
///     TERA_SUCCESS
///
INT32 cTERA_PCOIP_LINK::flush_bitmap_cache(void)
{
    INT     i;

    // If the cache is not empty free all allocated storage
    if (m_kmp_bmp_cache.max_num_bmps)
    {
        mTERA_LOG_DBG("freeing storage for %d bitmaps",
                      m_kmp_bmp_cache.max_num_bmps);
        for (i=0; i<m_kmp_bmp_cache.max_num_bmps; i++)
        {
            if (m_kmp_bmp_cache.bmp[i].bmp)
            {
                mTERA_MEM_FREE(m_kmp_bmp_cache.bmp[i].bmp);
            }
        }
    }

    // initialize the variables that manage the cache of bitmaps
    m_kmp_bmp_cache.max_num_bmps = 0;

    return TERA_SUCCESS;
}  // cTERA_PCOIP_LINK::flush_bitmap_cache()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::is_bitmap_cached
///     Returns non-zero value if the bitmap is cached, otherwise zero
//
//  Description:
///     This function searches the cache of bitmaps looking for a match to the
///     supplied bitmap. If a match is found the index of the bitmap stored in
///     the cache is returned in the variable cache_idx.
///
/// \return
///     zero - bitmap not in cache
///     non-zero - bitmap found in cache
///
INT32 cTERA_PCOIP_LINK::is_bitmap_cached(
    UINT8  hot_x,
    UINT8  hot_y,
    UINT8  width,
    UINT8  height,
    UINT8  and_depth,
    UINT8  xor_depth,
    void  *bmp,
    UINT32 bmp_len,
    eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type,
    UINT8 *cache_idx
    )
{
    UINT32                  i;
    INT32                   found = 0;
    UINT8                   idx;
    sTERA_PCOIP_BMP_DATA   *pcache_entry;

    // exit if the bitmap cache is empty
    if (0 == m_kmp_bmp_cache.max_num_bmps)
    {
        return 0;
    }

    // Search the cache for a bitmap that matches the caller's bitmap
    for (idx=0; idx<m_kmp_bmp_cache.max_num_bmps; idx++)
    {
        pcache_entry = &m_kmp_bmp_cache.bmp[idx];

        // compare the bitmap metadata first
        if((pcache_entry->bitmap_type != bitmap_type) ||
           (pcache_entry->hot_x != hot_x) ||
           (pcache_entry->hot_y != hot_y) ||
           (pcache_entry->width != width) ||
           (pcache_entry->height != height) ||
           (pcache_entry->and_depth != and_depth) ||
           (pcache_entry->xor_depth != xor_depth) ||
           (pcache_entry->bmp_len != bmp_len))
        {
            // something is different in the metadata, check the next entry
#if HOST_CUR_BMP_DBG
            mTERA_LOG_DBG("metadata differs for cached bitmap %d", idx);
#endif
            continue;
        }

        // compare the bitmap data byte by byte, stop comparing when a
        // difference is found
        UINT8 *pbmp = (UINT8 *)bmp;
        INT32  bmp_diff = 0;
        for (i=0; i<bmp_len; i++)
        {
            if (pcache_entry->bmp[i] != pbmp[i])
            {
#if HOST_CUR_BMP_DBG
                mTERA_LOG_DBG("bitmap data[%d] differs for cached bitmap %d", i, idx);
#endif
                bmp_diff = 1;
                break;
            }
        }

        // exit the loop if the bitmaps were identical
        if (0 == bmp_diff)
        {
            *cache_idx = idx;
            found = 1;
#if HOST_CUR_BMP_DBG
            mTERA_LOG_DBG("cached bitmap found at index %d", idx);
#endif
            break;
        }
    }
    return found;
}  // cTERA_PCOIP_LINK::is_bitmap_cached()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::add_bitmap_to_cache
///     This function stores a bitmap in the cache of bitmaps
//
//  Description:
///     The cTERA_PCOIP_LINK object maintains a cache of bitmaps that are
///     active on the client. This function determines the index of the next
///     bitmap to update and writes the bitmap data to the cache of bitmaps.
///
void cTERA_PCOIP_LINK::add_bitmap_to_cache(
    UINT8  hot_x,
    UINT8  hot_y,
    UINT8  width,
    UINT8  height,
    UINT8  and_depth,
    UINT8  xor_depth,
    void  *bmp,
    UINT32 bmp_len,
    eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type
    )
{
    sTERA_PCOIP_BMP_DATA   *pcache_entry;

    // exit if the bitmap cache is empty
    if (0 == m_kmp_bmp_cache.max_num_bmps)
    {
        mTERA_LOG_ERR("attempt to add bitmap to empty cache!!!");
        mTERA_ASSERT(0);
        return;
    }

    // calculate the index of the entry to update
    m_kmp_bmp_cache.idx_last_stored_bmp = (m_kmp_bmp_cache.idx_last_stored_bmp + 1) % m_kmp_bmp_cache.max_num_bmps;
#if HOST_CUR_BMP_DBG
    mTERA_LOG_DBG("storing new bitmap in cache at index %d",
                  m_kmp_bmp_cache.idx_last_stored_bmp);
#endif
    pcache_entry = &m_kmp_bmp_cache.bmp[m_kmp_bmp_cache.idx_last_stored_bmp];

    // write the new bitmap data to the cache
    pcache_entry->bitmap_type = bitmap_type;
    pcache_entry->hot_x = hot_x;
    pcache_entry->hot_y = hot_y;
    pcache_entry->width = width;
    pcache_entry->height = height;
    pcache_entry->and_depth = and_depth;
    pcache_entry->xor_depth = xor_depth;
    pcache_entry->bmp_len = bmp_len;
    mTERA_MEM_COPY(pcache_entry->bmp, bmp, bmp_len);

    return;
}  // cTERA_PCOIP_LINK::add_bitmap_to_cache()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_shape_bmp_send
///     Send a cursor shape bitmap
//
//  Description:
///     This function checks if the requested bitmap is in the client cache.
///     If so, the function sends the change pointer shape APDU, which supplies
///     an index to the client instead of the complete bitmap. If the bitmap
///     is not in the cache the function sends the update pointer shape APDU to
///     the client. When this occurs the bitmap is also added to the cache of
///     bitmaps stored in the cTERA_PCOIP_LINK object.
///
///     The cursor shape bitmap will be used by the firmware immediately for
///     the local cursor overlay.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_NO_MEMORY,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_shape_bmp_send(
    UINT8  hot_x,
    UINT8  hot_y,
    UINT8  width,
    UINT8  height,
    UINT8  and_depth,
    UINT8  xor_depth,
    void  *bmp,
    UINT32 bmp_len,
    eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP *buf = NULL;
    UINT32                                  buf_len;
    INT32                                   in_cache;
    INT32                                   is_alpha;
    UINT8                                   cache_index;

    mTERA_LOG_DBG("bmp_len=%d width=%d height=%d", bmp_len, width, height);

    if (bmp == NULL)
    {
        mTERA_LOG_ERR("invalid argument, bmp = NULL");

        mTERA_ASSERT(0);
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (bmp_len == 0)
    {
        mTERA_LOG_ERR("invalid argument, bmp_len = 0");

        mTERA_ASSERT(0);
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (lcur_shape_bmp_send_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else if ((bitmap_type != TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA) &&
             (bitmap_type != TERA_PCOIP_LINK_KMP_CUR_BITMAP_COLOR))
    {
        mTERA_LOG_ERR("bitmap_type %d not supported", bitmap_type);

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else if ((bitmap_type == TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA) &&
             ((and_depth != 0) || (xor_depth != 0)))
    {
        mTERA_LOG_ERR("and_depth (%d) AND xor_depth (%d) must both be zero for ALPHA bitmaps",
                      and_depth,
                      xor_depth);

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else if ((bitmap_type == TERA_PCOIP_LINK_KMP_CUR_BITMAP_COLOR) &&
             (and_depth != 1))
    {
        mTERA_LOG_ERR("and_depth (%d) must be 1 for COLOR bitmaps", and_depth);

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    // FW only supports values 1 and 32 for COLOR bitmaps.
    // Note: I haven't tested bitmaps with XOR depth = 16 or 24. Kourosh said
    // the soft host never sends them. Renee doesn't think the ZC was ever
    // tested with these types of bitmaps. The current code dumps an error
    // message if one is encountered.
    else if ((bitmap_type == TERA_PCOIP_LINK_KMP_CUR_BITMAP_COLOR) &&
             ((xor_depth != 1) && (xor_depth != 32)))
    {
        mTERA_LOG_ERR("xor_depth %d not supported for COLOR bitmaps", xor_depth);

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        // calculate the height of the stored bitmap (the height must be
        // divided by 2 if the depth of both masks is 1)
        if ((xor_depth == 1) && (and_depth == 1))
        {
            height = height/2;
        }

        // check if the bitmap is in the client cache
        in_cache = is_bitmap_cached(hot_x,
                                    hot_y,
                                    width,
                                    height,
                                    and_depth,
                                    xor_depth,
                                    bmp,
                                    bmp_len,
                                    bitmap_type,
                                    &cache_index);

        if (in_cache)
        {
            // send the bitmap cache index instead of the entire bitmap
            ret = lcur_change_shape_send(cache_index);
        }
        else
        {
            buf_len = sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP) + bmp_len;

            // Allocate memory
            buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP *)mTERA_MEM_ALLOC(buf_len);
            if (buf == NULL)
            {
                mTERA_LOG_ERR("failed to allocate memory for ioctl buffer");

                ret = TERA_ERR_NO_MEMORY;
            }
            else
            {
                // Fill out the APDU
                buf->apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SHAPE_BMP;
                buf->tstamp = 0;
                buf->bmp_type = (UINT8)bitmap_type;
                buf->apdu_len = sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP) + bmp_len;
                if (m_kmp_pri_ctxt.cur_shape_caches_supported &&
                    m_kmp_pri_ctxt.cur_shape_cache_size_supported &&
                    m_kmp_bmp_cache.max_num_bmps)
                {
                    buf->cache = ((m_kmp_bmp_cache.idx_last_stored_bmp + 1) % m_kmp_bmp_cache.max_num_bmps) |
                                  TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_CACHE_EN;
                }
                else
                {
                    buf->cache = 0;
                }
                buf->hot_x = hot_x;
                buf->hot_y = hot_y;
                buf->width = width;
                buf->height = height;
                buf->and_depth = and_depth;
                buf->xor_depth = xor_depth;

                // Copy the bitmap into the buffer right after the APDU
                if (bitmap_type == TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA)
                {
                    mTERA_MEM_COPY(&buf[1], bmp, bmp_len);
                    ret = TERA_SUCCESS;
#if HOST_CUR_BMP_DBG
                    print_alpha_bitmap(width,
                                       height,
                                       (UINT8 *)&buf[1]);
#endif
                }
                else
                {
                    // TBD: Currently this code is not used on Linux systems.
                    //      If/when it is used changes will likely be required.
                    // color bitmaps need to be flipped
                    is_alpha = 0;
                    ret = flip_bitmap(width,
                                      height,
                                      xor_depth,
                                      bmp_len,
                                      (UINT8 *)bmp,
                                      (UINT8 *)&buf[1],
                                      &is_alpha);

                    // update the bitmap type if necessary
                    if (is_alpha)
                    {
                        buf->bmp_type = TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA;
#if HOST_CUR_BMP_DBG
                        print_alpha_bitmap(width,
                                           height,
                                           (UINT8 *)&buf[1]);
#endif
                    }
#if HOST_CUR_BMP_DBG
                    else
                    {
                        print_color_bitmap(width,
                                           height,
                                           and_depth,
                                           xor_depth,
                                           (UINT8 *)&buf[1]);
                    }
#endif
                }

                if (TERA_SUCCESS == ret)
                {
                    // Send a synch ioctl to the device
                    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_SHAPE_BMP,
                                                buf, buf_len,
                                                1000);
                }
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("HDC_IOCTL_KMP_CUR_SHAPE_BMP failed (ret=%d)",
                                  ret);
                }
                else
                {
#if HOST_CUR_BMP_DBG
                    mTERA_LOG_DBG("sent local cursor shape bitmap - cache=%d",
                                  buf->cache);
#endif

                    // store the bitmap in the cache
                    add_bitmap_to_cache(hot_x,
                                        hot_y,
                                        width,
                                        height,
                                        and_depth,
                                        xor_depth,
                                        bmp,
                                        bmp_len,
                                        bitmap_type);
                }

                mTERA_MEM_FREE(buf);
            }
        }
    } // end if

    return ret;
} // cTERA_PCOIP_LINK::lcur_shape_bmp_send()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_change_shape_send
///     Send a cursor shape bitmap
//
//  Description:
///     This function sends a change pointer shape APDU message to the
///     firmware. The message contains a cache index, which the FW uses as a
///     reference into it's cache of previously sent bitmaps.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_change_shape_send(
    UINT8  index
    )
{
    INT32                               ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE  apdu;

    if (index >= m_kmp_pri_ctxt.cur_shape_caches)
    {
        mTERA_LOG_ERR("invalid argument, index>=%d",
                      m_kmp_pri_ctxt.cur_shape_caches);

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        mTERA_MEM_ZERO(&apdu, sizeof(apdu));

        // Fill out the APDU
        apdu.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SHAPE;
        apdu.index = index;

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_SHAPE,
                                    &apdu, sizeof(apdu),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_CUR_SHAPE failed, index=%d (ret=%d)",
                          apdu.index, ret);
        }
        else
        {
#if HOST_CUR_BMP_DBG
            mTERA_LOG_DBG("HDC_IOCTL_KMP_CUR_SHAPE ok, index=%d", apdu.index);
#endif
        }
    } // end if

    return ret;
} // cTERA_PCOIP_LINK::lcur_change_shape_send()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_shape_size_valid
///     Check if the size of a cursor is valid
//
//  Description:
///     This function compares the X and Y size of the callers cursor to the
///     maximum size cursor supported by the client. The PRI context data
///     read from the client at the start of a session contains information
///     on the maximum size cursor supported by the client.
///
///     A host cursor is considered valid if the X and Y dimensions are less
///     than or equal to these dimensions in the PRI context AND the amount
///     of memory needed to store the cursor bitmap is also supported by the
///     client.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lcur_shape_size_valid(
    UINT16 width,
    UINT16 height,
    INT    color_mask_present)
{
    UINT32 bitmap_size;

    if (color_mask_present)
    {
        // Color bitmaps have different sized XOR and AND masks.
        // The XOR mask is a 32-bits/pixel mask. The AND mask is
        // a 1-bit/pixel mask.
        bitmap_size = (width * height * XOR_MASK_BYTES_PER_PIXEL) +
                     ((width * height) / AND_MASK_BITS_PER_BYTE);
    }
    else
    {
        // Alpha bitmaps contain raw data sent to the client
        bitmap_size = width * height * ALPHA_MASK_BYTES_PER_PIXEL;
    }
    if ((m_kmp_pri_ctxt.cur_shape_max_size_supported) &&
        (width <= m_kmp_pri_ctxt.cur_shape_max_size_x) &&
        (height <= m_kmp_pri_ctxt.cur_shape_max_size_y) &&
        (m_kmp_pri_ctxt.cur_shape_cache_size_supported) &&
        (bitmap_size <= m_kmp_pri_ctxt.cur_shape_cache_size))
    {
        return TERA_SUCCESS;
    }
    else
    {
#ifdef HOST_CUR_BMP_DBG
        mTERA_LOG_DBG("invalid: w %d, h %d, bmp sz %d, max bmp sz %d",
                      width,
                      height,
                      bitmap_size,
                      m_kmp_pri_ctxt.cur_shape_cache_size);
#endif
        return TERA_ERR_NOT_SUPPORTED;
    }
} // cTERA_PCOIP_LINK::lcur_shape_size_valid()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_spd_avail
///     Check if the local cursor speed is available for getting/setting
//
//  Description:
///     This function lets the caller know if the local cursor speed is
///     available for setting/getting.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lcur_spd_avail(void)
{
    return (m_kmp_pri_ctxt.cur_speed_update_enable && m_kmp_pri_ctxt_valid) ? TERA_SUCCESS :
           TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::lcur_spd_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_spd_get
///     Get the local cursor speed
//
//  Description:
///     This function gets the local cursor speed. The local cursor speed is
///     determined by the client. As such, this function is only meaningful
///     when called during in a session.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_spd_get(
    UINT32 *lcur_spd
    )
{
    INT32                                 ret;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD apdu;

    if (lcur_spd == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (lcur_spd_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&apdu, sizeof(apdu));

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_SPD_GET,
                                    &apdu, sizeof(apdu),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_CUR_SPD_GET failed (ret=%d)", ret);
        }
        else
        {
            // Copy the local cursor speed to the user
            *lcur_spd = apdu.spd;
            mTERA_LOG_DBG("HDC_IOCTL_KMP_CUR_SPD_GET ok, speed=%d", apdu.spd);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_spd_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_spd_set
///     Set the local cursor speed
//
//  Description:
///     This function gets the local cursor speed. The local cursor speed is
///     determined by the client. As such, this function is only meaningful
///     when called during in a session.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_spd_set(
    UINT32 lcur_spd
    )
{
    INT32                                 ret;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD apdu;


    if (lcur_spd_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else if ((lcur_spd < TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MIN) ||
             (lcur_spd > TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MAX))
    {
        mTERA_LOG_ERR("error setting speed equal to %d, must be >= %d and <= %d",
                      lcur_spd,
                      TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MIN,
                      TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MAX);

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        mTERA_MEM_ZERO(&apdu, sizeof(apdu));

        // Fill out the APDU
        apdu.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD;
        apdu.spd = lcur_spd;

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_SPD_SET,
                                    &apdu, sizeof(apdu),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_CUR_SPD_SET failed, speed=%d (ret=%d)",
                          apdu.spd, ret);
        }
        else
        {
            mTERA_LOG_DBG("HDC_IOCTL_KMP_CUR_SPD_SET ok, speed=%d", apdu.spd);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_spd_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_lkbd_avail
///     Check if local cursor and keyboard are available
//
//  Description:
///     This function lets the caller know if local cursor and keyboard are
///     available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lcur_lkbd_avail(void)
{
    return (m_kmp_pri_ctxt.pcoip_kmp_1_supported && m_kmp_pri_ctxt_valid) ? TERA_SUCCESS :
           TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::lcur_lkbd_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_lkbd_enable
///     Enable the local cursor and keyboard
//
//  Description:
///     This function enables the local cursor and keyboard, which translates
///     to enabling the KMP features. This function creates an APDU to enable
///     the KMP features, and sends it to the driver via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_lkbd_enable(void)
{
    INT32                            ret;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN feat_en;

    if (lcur_lkbd_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&feat_en, sizeof(feat_en));

        // Fill out the APDU
        feat_en.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_EN;
        feat_en.en = 1;

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_FEAT_EN,
                                    &feat_en, sizeof(feat_en),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_FEAT_EN failed (ret=%d)", ret);
        }
        else
        {
            mTERA_LOG_DBG("enabled local cursor and keyboard");
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_lkbd_enable()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_lkbd_disable
///     Disable the local cursor and keyboard
//
//  Description:
///     This function disables the local cursor and keyboard, which translates
///     to disabling the KMP features. This function creates an APDU to disable
///     the KMP features, and sends it to the driver via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_lkbd_disable(void)
{
    INT32                            ret;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN feat_en;

    if (lcur_lkbd_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&feat_en, sizeof(feat_en));

        // Fill out the APDU
        feat_en.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_EN;
        feat_en.en = 0;

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_FEAT_EN,
                                    &feat_en, sizeof(feat_en),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_FEAT_EN failed (ret=%d)", ret);
        }
        else
        {
            mTERA_LOG_DBG("disabled local cursor and keyboard");
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_lkbd_disable()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_idle_hide_avail
///     Check if the local cursor idle hide feature is available
//
//  Description:
///     This function lets the caller know if the local cursor idle hide feature
///     is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::lcur_idle_hide_avail(void)
{
    return (m_kmp_pri_ctxt.cur_shape_fadeout_enable && m_kmp_pri_ctxt_valid) ? TERA_SUCCESS :
           TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::lcur_idle_hide_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_idle_hide_set
///     Set the local cursor idle hide
//
//  Description:
///     This function sets the local cursor idle hide feature. The caller can
///     specify the amount of time (in seconds) that the local cursor is idle
///     before the overlay is hidden. If a time value of zero is specified,
///     then the overlay is not hidden.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::lcur_idle_hide_set(
    UINT32 tstamp
    )
{
    INT32                                  ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE apdu;

    mTERA_LOG_DBG("start");

    if (lcur_idle_hide_avail() != TERA_SUCCESS)
    {
        // the soft client does not support hiding the local cursor
        mTERA_LOG_DBG("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&apdu, sizeof(apdu));

        // Fill out the APDU
        apdu.apdu_type = TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_IDLE_HIDE;
        apdu.time = tstamp;

        // Send a synch ioctl to the device
        ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_CUR_IDLE_HIDE,
                                    &apdu, sizeof(apdu),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HDC_IOCTL_KMP_CUR_IDLE_HIDE failed (ret=%d)", ret);
        }
        else
        {
            mTERA_LOG_DBG("local cursor idle hide set successfully");
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::lcur_idle_hide_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::is_client_sc
///     Returns non-zero value if the client is a soft client, otherwise zero
//
//  Description:
///     A "client type" field is not defined in any message exchanged between
///     the FW and host SW. A field will be added in the future. Until this
///     is implemented the host SW will infer this by looking at the KMP PRI
///     cursor hide/fade feature. This feature is only supported by zero
///     clients.
///
/// \return
///     zero - client is NOT soft client (must be zero client)
///     non-zero - client is soft client
///
INT32 cTERA_PCOIP_LINK::is_client_sc(void)
{
    // if the context data is not valid assume the client is not a soft client
    if (!m_kmp_pri_ctxt_valid)
    {
        return 0;
    }
    else
    {
        return (m_kmp_pri_ctxt.cur_shape_fadeout_enable) ? 0 : 1;
    }
} // cTERA_PCOIP_LINK::is_client_sc()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::lcur_disable_avail
///     Check if the local cursor feature can be disabled
//
//  Description:
///     This function lets the caller know if the local cursor feature can
///     be disabled.
///
/// \return
///     TERA_ERR_NOT_SUPPORTED - feature cannot be disabled,
///     TERA_SUCCESS - featu re can be disabled
///
INT32 cTERA_PCOIP_LINK::lcur_disable_avail(void)
{
    if ((!m_kmp_pri_ctxt.usb_supported) || (!m_kmp_pri_ctxt.usb_enable))
    {
        return TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        return TERA_SUCCESS;
    }
} // cTERA_PCOIP_LINK::lcur_disable_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::kbd_data_get
///     Get the available keyboard data
//
//  Description:
///     This function reads the available keyboard event data, if any is
///     available from the driver.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
#if 0
INT32 cTERA_PCOIP_LINK::kbd_data_get(
    sTERA_PCOIP_HDC_KMP_KBD_DATA_EVT *kbd_data
    )
{
    INT32   ret;

    mTERA_ASSERT(kbd_data != NULL);

    mTERA_MEM_ZERO(kbd_data, sizeof(sTERA_PCOIP_HDC_KMP_KBD_DATA_EVT));

    // Send a synch ioctl to the device
    ret = m_hdc_dev.synch_ioctl(TERA_PCOIP_HDC_IOCTL_KMP_KBD_DATA_GET,
                                kbd_data,
                                sizeof(sTERA_PCOIP_HDC_KMP_KBD_DATA_EVT),
                                1000);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HDC_IOCTL_KMP_KBD_DATA_GET failed (ret=%d)", ret);
    }
#if 0
    else
    {
        // dump the event data if present
        if (0 != kbd_data->num_evts)
        {
            UINT  i;
            for (i=0; i<kbd_data->num_evts; i++)
            {
                mTERA_LOG_DBG("APDU %X tstamp %d code %X lang %X lay %X state %d flg %X lck %x mod %X",
                              kbd_data->evt[i].apdu_type,    ///< APDU type
                              kbd_data->evt[i].tstamp,       ///< Timestamp
                              kbd_data->evt[i].scancode,     ///< Scancode (PS/2 or USB)
                              kbd_data->evt[i].lang,         ///< Language
                              kbd_data->evt[i].layout,       ///< Layout
                              kbd_data->evt[i].state,        ///< Key state {0: up; 1: down}
                              kbd_data->evt[i].flags,        ///< Flags (bitmapped)
                              kbd_data->evt[i].lock,         ///< Lock status (bitmapped)
                              kbd_data->evt[i].mod);         ///< Modifier keys (bitmapped)
            }
        }
    }
#endif

    return ret;
} // cTERA_PCOIP_LINK::kbd_data_get()
#endif
